/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

declare name "OneTrick SIMIAN";
declare author "Oren Kurtz";
declare copyright "Punk Labs LLC 2022";
declare version "1.00";
declare license "GPLv3-or-later";

//declare options "[midi:on][nvoices:1]";
declare options "[midi:on]";
import("stdfaust.lib");
import("../OneTrick/DSP/Envelopes.lib");
import("../OneTrick/DSP/Reverbs.lib");
import("../OneTrick/DSP/Sampler.lib");

simianAttackSeconds = 1/1000; // 0 could result in popping
simianClickSeconds = 19/1000; // Original Was 12ms? Seems short
simianChokeSeconds = 20/1000;

// Samples:
// Original Recording (not licensed, can't use):
//cymbals = environment { import("Samples/SDSV_Cym_and_HH_Waveform-22khz.lib"); };
// Public Domain:
//cymbals = environment { import("Samples/75342__neotone__looped-cymbol-scrape-22khz.lib"); };
cymbals = environment { import("Samples/428614__hallkev__cymbal-swells-kevinsticks--22khz.lib"); };

simianNoise =
	  no.noise
	: ba.downSample(min(ma.SR, 48000)) // Cut above sparkle, prevents dulling when downsampling from 96kHz+
	: fi.highpass(2, 200); // Cut mud


wakeUpGate = checkbox("WakeUp");

smoothParam(gate) = si.smooth(0.9995 * (wakeUpGate<=wakeUpGate')); // Skip smoothing when waking up

//log_interpolate(n, a, b) = 2^it.interpolate_linear(n, ma.log2(a), ma.log2(b));

// Inputs:
// Doesn't work in FaustIDE, but does work in builds... Useful for combining a kit.

sign(t) = 1.0 - ((t<0)*2);
falloff(t, halfwayValue) = t / ((1.0 / min(max(halfwayValue, 0.001), 0.999) - 2.0) * (1.0 - t) + 1.0);
waveshape(halfwayValue, t) = falloff(abs(t), halfwayValue) * sign(t);
applySaturation(s) = result with {
	saturationInverseGain = it.interpolate_linear(s, 1, 0.3);
    result = waveshape(it.interpolate_linear(s, 0.5, 0.9)) * saturationInverseGain;
};

ms2s(ms) = ms/1000;
ns2s(ns) = ns/1000000;

// 1176-style Limiting using a fast compressor
//1176
//applyLimit = stereo(*(2.0) : co.limiter_lad_bw : *(0.5));
//applyLimit = co.compressor_stereo(8, -3, ns2s(20), ms2s(50));
applyLimit = co.compressor_stereo(20, -1, ns2s(20), ms2s(50));
finalizeOutput = stereo(*(globalDrive)) : applyLimit : stereo(applySaturation(globalSaturation) : *(globalGain));



globalGain = ba.db2linear(max(-100, hslider("t:Globals/v:Globals/[0]Gain_Global", 0, -100, 6, 0.1))) : smoothParam(1);
globalDrive = ba.db2linear(max(-100, hslider("t:Globals/v:Globals/[1]Drive_Global", -6, -100, 6, 0.1))) : smoothParam(1);
globalReverbDecay = it.interpolate_linear(hslider("t:Globals/v:Globals/[2]ReverbDecay_Global", 50, 0, 100, 0.01) * 0.01, 0.25, 1.0);
globalReverbGate = hslider("t:Globals/v:Globals/[3]ReverbGate_Global", -100, -100, 0, 0.01);
globalSensitivity = hslider("t:Globals/v:Globals/[4]Sensitivity_Global", 100, 0, 100, 0.01) * 0.01;
globalSaturation = hslider("t:Globals/v:Globals/[5]Saturation_Global", 0, 0, 100, 0.01) * 0.01 : smoothParam(1);

simianModule(osc, oscWaveform, gain, lowpassFreq, oscPitch, oscEnvBend, envDecay, envPunch, fadeToneNoise, fadeClick, gate, chokeGate, velocity, gainDynamics, oscEnvBendDynamics, lowpassQ, saturation, pan, lowpassEnvBend, lowpassEnvBendDynamics, sendReverb) = result with {
	gainVelocity = it.interpolate_linear(gainDynamics, 1, velocity);
	toneVelocity = it.interpolate_linear(velocity, 1-oscEnvBendDynamics, 1);
	filterVelocity = it.interpolate_linear(velocity, 1-lowpassEnvBendDynamics, 1);
	
	hold = ba.sAndH(gate > gate');
	
	EnvDecayDynamics = 0.30; // 30% decay dynamics hardcoded... Could be configurable for 2.0?
	heldEnvDecay = envDecay/1000*it.interpolate_linear(1-EnvDecayDynamics, 1, velocity) : hold;
	heldEnvPunch = envPunch : hold;
	
	// Envelopes
	envSlopeMid = it.interpolate_linear(heldEnvPunch, 0.5, 0.9);
    env = ar_running_slope(simianAttackSeconds, heldEnvDecay, gate, 1, envSlopeMid) * chokeEnv;
    gainEnv = ba.lin2LogGain(ar_running_slope(simianAttackSeconds, heldEnvDecay, gate, ba.log2LinGain(gainVelocity), envSlopeMid) * chokeEnv);
    toneEnv = ar_running_slope(simianAttackSeconds, heldEnvDecay, gate, toneVelocity, envSlopeMid) * chokeEnv;
    filterEnv = ar_running_slope(simianAttackSeconds, heldEnvDecay, gate, filterVelocity, envSlopeMid) * chokeEnv;

    chokeEnv = en.ar(0, simianChokeSeconds, chokeGate) * chokeGate + (1-chokeGate);
    clickGainEnv = ba.lin2LogGain(ar_running(simianAttackSeconds, simianClickSeconds, gate, ba.log2LinGain(gainVelocity)));
    
    punchBend = 0;//12;
    punchEnv = 0;//ar_running(1/1000, 19/1000, gate, velocity)^2;
    
    toneSweep = ba.semi2ratio(oscEnvBend * toneEnv + punchBend*punchEnv);
    //filterSweep = it.interpolate_linear(filterSweep, 1, toneSweep);
    filterSweep = ba.semi2ratio(lowpassEnvBend * filterEnv);
    
    // Can be reused by osc, if wanted
	noiseFilter = fi.resonlp(min(lowpassFreq*filterSweep, 19000), max(0.01, lowpassQ), 1);
    noise = simianNoise : noiseFilter;
    pitch = oscPitch * toneSweep;
    tone = osc(pitch, oscWaveform, noiseFilter);
    drum = it.interpolate_linear(fadeToneNoise, tone, noise);
    
    // Crossfade Click with Drum (keeps volume in check)
    preResult = it.interpolate_linear(fadeClick*clickGainEnv, drum, noise) : *(gainEnv) : applySaturation(saturation) : *(gain);
    // Add Click (better for noise-only voices)
    //preResult = drum+(noise*clickGainEnv*fadeClick) : *(gainEnv) : applySaturation(saturation) : *(gain);
   
    // Mono:
    //result = preResult;
    // Stereo:
    //result = preResult <: (_*max(0,1-pan), _*max(0,pan+1));
    // Stereo+Sidechain
    result = (preResult <: (_*min(1,1-pan), _*min(1,pan+1)), preResult * sendReverb);
    // Reverb Test:
    //result = it.interpolate_linear(pan * 0.5 + 0.5, preResult, (preResult : re.satrev : select2(0))) <: si.bus(2);
};

simianModuleGenericUI(osc) = result with {
    //[style:knob]
    // Hack:	Combine faust midiGate with "Trigger_N" parameter so either work
    //			Also, multiply by midiVelocity for implicit Velocity
    //			This will allow "Trigger_N" to include velocity!
    gateAndVelocity = nentry("Trigger", 0, 0, 1, 0.01);
    gate = (gateAndVelocity>0);
    velocity = gateAndVelocity : ba.sAndH(gate);
    //it.interpolate_linear(globalSensitivity, 1, gateAndVelocity : ba.sAndH(gate));
    
    //"_%voiceIndex"
    
    chokeGate = nentry("Choke", 0, 0, 1, 0.01);

	// Tone
   	oscWaveform = hslider("[1]OscWaveform", 0, 0, 100, 0.001) * 0.01 : smoothParam(gate);
   	oscPitch = hslider("[1]OscPitch[midi:ctrl 71][scale:log]", 220, 30, 4400, 0.001) : smoothParam(gate);
    oscEnvBend = hslider("[2]OscEnvBend[midi:ctrl 72]", 0, -36, 36, 0.01);
    oscEnvBendDynamics = hslider("[3]OscEnvBendDynamics", 100, 0, 100, 0.01) * 0.01;

	// Noise
    lowpassFreq = hslider("[10]LowpassFreq[scale:log]", 10000, 440, 15000, 1);
    lowpassQ = hslider("[11]LowpassQ[scale:log]", 0, 0, 100, 0.01) * 0.01 + 1;
    lowpassEnvBend = hslider("[12]LowpassEnvBend[midi:ctrl 72]", 0, -36, 36, 0.01);
    lowpassEnvBendDynamics = hslider("[13]LowpassEnvBendDynamics", 100, 0, 100, 0.01) * 0.01;

	// Envelope
    envDecay = hslider("[20]EnvDecay[midi:ctrl 74]", 200, 10, 2000, 1);
    envPunch = hslider("[21]EnvPunch[midi:ctrl 74]", 0, 0, 100, 0.01) * 0.01;

	// Mix
    fadeToneNoise = hslider("[30]FadeToneNoise[midi:ctrl 73]", 25, 0, 100, 0.01) * 0.01;
    fadeClick = hslider("[31]FadeClick", 50, 0, 100, 0.01) * 0.01;
    sendReverb = hslider("[32]SendReverb", 0, 0, 100, 0.01) * 0.01;

	// Output
	gain = ba.db2linear(max(-100, hslider("[40]Gain", 0, -100, 6, 0.1))) : smoothParam(gate);
	pan = hslider("[41]Pan", 0, -100, 100, 0.01) * 0.01 : smoothParam(gate);
    gainDynamics = hslider("[42]GainDynamics", 100, 0, 100, 0.01) * 0.01;
	saturation = hslider("[43]Saturation", 0, 0, 100, 0.01) * 0.01;

    result = simianModule(osc, oscWaveform, gain, lowpassFreq, oscPitch, oscEnvBend, envDecay, envPunch, fadeToneNoise, fadeClick, gate, chokeGate, velocity, gainDynamics, oscEnvBendDynamics, lowpassQ, saturation, pan, lowpassEnvBend, lowpassEnvBendDynamics, sendReverb);
};

simianModuleUI = result with {
   	osc(freq, wave, noiseFilter) = it.interpolate_linear(wave,
   		os.polyblep_triangle(freq) : fi.lowpass(1, 338.6), // Softer filtered Triangle
   		sampleOsc(freq/220, cymbals) : noiseFilter // Sampled Cymbal sound
   		); 
	result = simianModuleGenericUI(osc);
};


addStereo(b) = (_,b,_) : par(i, 2, +);
// Stereo Addition
/*
addModule = addStereo;
moduleToStereo(l,r) = (l, r);
//*/
// Stereo + Sidechain Addition
//*
moduleToStereo(l,r,fx) = (l, r);
getModuleStereo(m) = m:moduleToStereo;
moduleToSidechain(l,r,fx) = fx;
getModuleSidechain(m) = m:moduleToSidechain;
modulesCombine(l1,r1,fx1,l2,r2,fx2) = (l1+l2,r1+r2,fx1+fx2);
addModule(m) = m,(_,_,_):modulesCombine;
addModules(m1,m2) = modulesCombine;
stereo(m) = (m,m);
//*/


// SatRev
//simianReverb = re.satrev; //Nice, but can't control decay?
simianReverb = satrev_scaled(globalReverbDecay);

// Freeverb
//(lpCoef, apCoef, lpDamp, spreadSamples)
//simianReverb = _ <: re.stereo_freeverb(0.75, 0.75, 0.99, ma.SR/1000);
/* //Can't figure this one out, hangs the compiler
simianReverb = _ <: si.bus(2) : re.fdnrev0(MAXDELAY, delays, BBS0, freqs, durs, loopgainmax, nonl) with {
	MAXDELAY=ma.np2(ma.SR/2);
	pathmin = 5; //shortest room size in meters
	pathmax = 10; //longest room size in meters
	delays = de.prime_power_delays(2,pathmin,pathmax);
	BBS0 = 3; //Odd (order of bandsplits at freqs)
	freqs = (440, 220);
	durs = (0.5, 0.5, 0.5);
	loopgainmax = 0.99;
	nonl = 0.1;
};
//*/

/* // Zita
(DelayMS, CrossFreq1, CrossFreq2, DecayLowSec, DecayMidSec, MaxSampleRateHz
blendZita(echo, reverb) = it.interpolate_linear(1.0, echo, reverb) <: si.bus(2);
simianReverb = _ <: re.zita_rev1_stereo(20, 300, 7000, 0.1, 1.5, ma.SR) : blendZita;
//*/

/* // Datto
simianReverb = _ <: re.dattorro_rev(pre_delay, bw, i_diff1, i_diff2, decay, d_diff1, d_diff2, damping) with {
    pre_delay = 0;		// 0: pre-delay in samples (fixed at compile time)
    bw = 0.9995;		// 0.9995: band-width filter (pre filtering); (0 - 1)
    i_diff1 = 0.750;	// 0.750: input diffusion factor 1; (0 - 1)
    i_diff2 = 0.625;	// 0.625: input diffusion factor 2;
    decay = 0.4;		// 0.5: decay rate; (0 - 1); infinite decay = 1.0
    d_diff1 = 0.70;		// 0.70: decay diffusion factor 1; (0 - 1)
    d_diff2 = 0.50;		// 0.50: decay diffusion factor 2;
    damping = 0.0005;	// 0.0005: high-frequency damping; no damping = 0.0
};
//*/

/* // JPverb
simianReverb = _ <: re.jpverb(t60, damp, size, early_diff, mod_depth, modFreq, low, mid, high, lowcut, highcut) with {
    t60=1.0;			//: approximate reverberation time in seconds ([0.1..60] sec) (T60 - the time for the reverb to decay by 60db when damp == 0 ). Does not effect early reflections
    damp=0.0005;		//: controls damping of high-frequencies as the reverb decays. 0 is no damping, 1 is very strong damping. Values should be between ([0..1])
    size=1.5;			//: scales size of delay-lines within the reverberator, producing the impression of a larger or smaller space. Values below 1 can sound metallic. Values should be between `[0.5..5]
    early_diff=0.707;	//: controls shape of early reflections. Values of 0.707 or more produce smooth exponential decay. Lower values produce a slower build-up of echoes. Values should be between ([0..1])
    mod_depth=0.05;		//: depth ([0..1]) of delay-line modulation. Use in combination with mod_freq to set amount of chorusing within the structure
    modFreq=3;			//: frequency ([0..10] Hz) of delay-line modulation. Use in combination with modDepth to set amount of chorusing within the structure
    low=0.25;			//: multiplier ([0..1]) for the reverberation time within the low band
    mid=0.5;			//: multiplier ([0..1]) for the reverberation time within the mid band
    high=1.0;			//: multiplier ([0..1]) for the reverberation time within the high band
    lowcut=250;			//: frequency (100..6000 Hz) at which the crossover between the low and mid bands of the reverb occurs
    highcut=5000;		//: frequency (1000..10000 Hz) at which the crossover between the mid and high bands of the reverb occurs
};
//*/

/* // Greyhole
simianReverb = _ <: re.greyhole(dt, damp, size, early_diff, feedback, mod_depth, mod_freq) with {

    dt=0.5;				//: approximate reverberation time in seconds ([0.1..60 sec])
    damp=0.0005;		//: controls damping of high-frequencies as the reverb decays. 0 is no damping, 1 is very strong damping. Values should be between ([0..1])
    size=2.5;			//: control of relative "room size" roughly between ([0.5..3])
    early_diff=0.707;	//: controls pattern of echoes produced by the diffuser. At very low values, the diffuser acts like a delay-line whose length is controlled by the 'size' parameter. Medium values produce a slow build-up of echoes, giving the sound a reversed-like quality. Values of 0.707 or greater than produce smooth exponentially decaying echoes. Values should be in the range ([0..1])
    feedback=0.0;		//: amount of feedback through the system. Sets the number of repeating echoes. A setting of 1.0 produces infinite sustain. Values should be in the range ([0..1])
    mod_depth=0.1;		//: depth ([0..1]) of delay-line modulation. Use in combination with mod_freq to produce chorus and pitch-variations in the echoes
    mod_freq=3;			//: frequency ([0..10] Hz) of delay-line modulation. Use in combination with mod_depth to produce chorus and pitch-variations in the echoes
};
//*/


processDrum = simianModuleUI;
processOutput(l,r,fx) =
	getModuleStereo(modules) :
	addStereo(getModuleSidechain(modules) : simianReverb : stereo(fi.highpass(2, 300)) : ef.gate_stereo(globalReverbGate, 0.1/1000, 100/1000, 20/1000)) :
 	finalizeOutput /* <: si.bus(2) */
 	with { 
		modules = (l,r,fx);
	} ;

