/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

module gui;

import std.algorithm.searching;
import std.process;

import dplug.gui;
import dplug.flatwidgets;
import dplug.client;
import dplug.canvas;
import dplug.graphics.font;
//import dplug.wren;

import core.stdc.stdio;
import core.stdc.string;

import main;
import theme; // Wren replacement
import onetrick;

class PluginGUI : CustomGUI, IButtonListener //Don't fill background, we'll use a UIColorPanel instead that's scriptable
{
public:
nothrow:
@nogc:
    PluginClient _client;
    this(PluginClient client)
    {
        _client = client;

        static immutable float[7] ratios = [0.75f, 1.0f, 1.25f, 1.5f, 2.0f];
        super( makeSizeConstraintsDiscrete(1000, 640, ratios), flagRaw);
        
        //super( makeSizeConstraintsContinuous(1280, 720, 0.5, 2.0));
        
        setUpdateMargin(0); // adds additional padding around update regions
        
        UIHelper.initialize(); // Avoid potential race condition with lazy-loading during render
        
        const string tekoLicense = import("Fonts/Teko/OFL.txt");
        _labelFont = mallocNew!Font(cast(ubyte[])(import("Fonts/Teko/Teko-Regular.ttf")));
        const string rajdhaniLicense = import("Fonts/Rajdhani/OFL.txt");
        _logoFont = mallocNew!Font(cast(ubyte[])(import("Fonts/Rajdhani/Rajdhani-Bold.ttf")));

        addChild(background = withId(mallocNew!UIColorPanel(context()), "background"));
        
		background.addChild(withId(mallocNew!UIIcon(context(), Icon.OT_OneTrickIcon), "logoIcon"));
		background.addChild(withId(mallocNew!UILabel(context(), _logoFont, "SIMIAN"), "logoLabel"));
		debug
		{
			char[1024] versionString;
			getEditionText(versionString.ptr, versionString.length, "Debug Build %d.%d.%d");
			background.addChild(withId(mallocNew!UILabel(context(), _labelFont, versionString), "versionLabel"));
		}
		else
		{
			version(Prerelease)
			{
				char[1024] versionString;
				getEditionText(versionString.ptr, versionString.length, "Prerelease %d.%d.%d");
				background.addChild(withId(mallocNew!UILabel(context(), _labelFont, versionString), "versionLabel"));
			}
			else version(Lite)
			{
				char[1024] versionString;
				getEditionText(versionString.ptr, versionString.length, "Lite Version %d.%d.%d");
				background.addChild(withId(mallocNew!UILabel(context(), _labelFont, versionString), "versionLabel"));
			}
		}
		background.addChild(withId(mallocNew!UIPresetSelector(client, context(), _labelFont), "presetSelector"));
		background.addChild(creditsButton = withId(mallocNew!UIButton(context(), Icon.Info), "creditsButton"));
		creditsButton.addListener(this);

		background.addChild(globalPanel = withId(mallocNew!UIColorPanel(context()), "globalPanel"));
		globalPanel.addChild(simianIcon = withId(mallocNew!UIIcon(context(), Icon.Plugin_Icon1), "simianIcon"));
		globalPanel.addChild(globalKnobGain = withId(mallocNew!UIKnob(context(), client.param(paramGlobalGain), "Volume", _labelFont), "globalKnobGain"));
		globalPanel.addChild(globalKnobDrive = withId(mallocNew!UIKnob(context(), client.param(paramGlobalDrive), "Drive", _labelFont), "globalKnobDrive"));
		globalPanel.addChild(globalKnobSensitivity = withId(mallocNew!UIKnob(context(), client.param(paramGlobalSensitivity), "Velocity", _labelFont), "globalKnobSensitivity"));
		globalPanel.addChild(globalKnobSaturation = withId(mallocNew!UIKnob(context(), client.param(paramGlobalSaturation), "Saturation", _labelFont), "globalKnobSaturation"));
		globalPanel.addChild(globalKnobReverbDecay = withId(mallocNew!UIKnob(context(), client.param(paramGlobalReverbDecay), "Decay", _labelFont), "globalKnobReverbDecay"));
		globalPanel.addChild(globalKnobReverbGate = withId(mallocNew!UIKnob(context(), client.param(paramGlobalReverbGate), "Gate", _labelFont), "globalKnobReverbGate"));
		globalPanel.addChild(withId(mallocNew!UISeparator(context()), "globalSeparator1"));
		globalPanel.addChild(withId(mallocNew!UISeparator(context()), "globalSeparator2"));
		globalPanel.addChild(withId(mallocNew!UILabel(context(), _labelFont, "Input"), "globalLabel1"));
		globalPanel.addChild(withId(mallocNew!UILabel(context(), _labelFont, "Room Reverb"), "globalLabel2"));
		globalPanel.addChild(withId(mallocNew!UILabel(context(), _labelFont, "Output"), "globalLabel3"));

		globalPanel.addChild(upgradeButton = withId(mallocNew!UIButton(context(), Icon.OT_Upgrade), "upgradeButton"));
		upgradeButton.addListener(this);
		version(Lite){} else { upgradeButton.visibility(false); }

		
		
		background.addChild(tabPager = withId(mallocNew!UIPager(context()), "tabPager"));
		background.addChild(tabContainer = withId(mallocNew!UITabContainer(context(), tabPager), "tabContainer"));
		
		tabContainer.addChild(module1Tab = withId(mallocNew!UITab(context(), "Kick", _labelFont), "module1Tab"));
		tabContainer.addChild(module2Tab = withId(mallocNew!UITab(context(), "Snare", _labelFont), "module2Tab"));
		tabContainer.addChild(module3Tab = withId(mallocNew!UITab(context(), "Rimshot", _labelFont), "module3Tab"));
		tabContainer.addChild(module4Tab = withId(mallocNew!UITab(context(), "Clap", _labelFont), "module4Tab"));
		tabContainer.addChild(module5Tab = withId(mallocNew!UITab(context(), "Tom1", _labelFont), "module5Tab"));
		tabContainer.addChild(module6Tab = withId(mallocNew!UITab(context(), "Tom2", _labelFont), "module6Tab"));
		tabContainer.addChild(module7Tab = withId(mallocNew!UITab(context(), "Tom3", _labelFont), "module7Tab"));
		tabContainer.addChild(module8Tab = withId(mallocNew!UITab(context(), "HH Closed", _labelFont), "module8Tab"));
		tabContainer.addChild(module9Tab = withId(mallocNew!UITab(context(), "HH Open", _labelFont), "module9Tab"));
		tabContainer.addChild(module10Tab = withId(mallocNew!UITab(context(), "Cymbal", _labelFont), "module10Tab"));
		const immutable(Icon)[9] icons = [Icon.OT_KickDrum, Icon.OT_SnareDrum, Icon.OT_Claves, Icon.Clap, Icon.OT_TomDrumLarge, Icon.OT_TomDrumMedium, Icon.OT_TomDrumSmall, Icon.OT_Cymbal, Icon.OT_Hihat];
		int moduleIndex = 0;
		module1Tab.addChild(module1Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.OT_KickDrum).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module1Indicator"));
		module2Tab.addChild(module2Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.OT_SnareDrum).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module2Indicator"));
		module3Tab.addChild(module3Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.OT_Claves).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module3Indicator"));
		module4Tab.addChild(module4Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.Clap).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module4Indicator"));
		module5Tab.addChild(module5Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.OT_TomDrumSmall).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module5Indicator"));
		module6Tab.addChild(module6Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.OT_TomDrumMedium).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module6Indicator"));
		module7Tab.addChild(module7Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.OT_TomDrumLarge).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module7Indicator"));
		module8Tab.addChild(module8Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.OT_Cymbal).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module8Indicator"));
		module9Tab.addChild(module9Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.OT_Hihat).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module9Indicator"));
		module10Tab.addChild(module10Indicator = withId(mallocNew!UIPeakIndicator(context(), Icon.OT_Cymbal).withSelectableIcon(icons, cast(IntegerParameter)_client.moduleParam(paramModuleIcon, moduleIndex++)), "module10Indicator"));

		version(Lite)
		{		
			module3Tab.enabled(false);
			module4Tab.enabled(false);
			module5Tab.enabled(false);
			module6Tab.enabled(false);
			module10Tab.enabled(false);
		}
		
		UIElement e = null;

		for(int i=0; i<Module.KitPieceCount; i++)
			addModule(i, (i>=7));		


        globalPanel.addChild(withId(mallocNew!UIWindowResizer(context()), "resizerHint"));


		// Somehow allows clicking through to knobs!?
		const(char)[] creditsTextPattern = import("credits.txt");
		auto creditsTextWithVersion = StringHelper.duplicateString(creditsTextPattern);
		getEditionText(creditsTextWithVersion.ptr, creditsTextWithVersion.length, creditsTextPattern);
		
		background.addChild(creditsPanel = withId(mallocNew!UICreditsPanel(context(), _labelFont, creditsTextWithVersion), "creditsPanel"));
		creditsPanel.visibility(false);
		creditsPanel.setZOrder(100);
        
        /* Enable Wren
        mixin(fieldIdentifiersAreIDs!PluginGUI); // Each UIElement in this object receives its identifier as runtime ID, ie. _inputSlider receives ID "_inputSlider"
        context.enableWrenSupport();
        
        // Import Classes to ensure Script has access:
        context.wrenSupport.registerUIElementClass!UIButton();
        context.wrenSupport.registerUIElementClass!UIColorPanel();
        context.wrenSupport.registerUIElementClass!UIConnector();
        context.wrenSupport.registerUIElementClass!UICreditsPanel();
        context.wrenSupport.registerUIElementClass!UIKnob();
        context.wrenSupport.registerUIElementClass!UILabel();
        context.wrenSupport.registerUIElementClass!UIIcon();
        context.wrenSupport.registerUIElementClass!UIPager();
        context.wrenSupport.registerUIElementClass!UIPanel();
        context.wrenSupport.registerUIElementClass!UIPeakIndicator();
        context.wrenSupport.registerUIElementClass!UIPresetSelector();
        context.wrenSupport.registerUIElementClass!UISeparator();
        context.wrenSupport.registerUIElementClass!UITab();
        context.wrenSupport.registerUIElementClass!UITabContainer();
        context.wrenSupport.registerUIElementClass!UIParameter();
        context.wrenSupport.registerUIElementClass!UIValue();
        
        debug
        {
        	import std.path;
        	import punklabs.string.helper;
			enum codeDir = dirName(__FILE_FULL_PATH__);
			char[100000] scriptPath;
			StringHelper.formatString(scriptPath, "%s/GUI.wren", CString(codeDir).storage);
			context.wrenSupport.addModuleFileWatch("plugin", scriptPath);
		}
        else
        {
			context.wrenSupport.addModuleSource("plugin", import("GUI.wren"));
		}
		
		
        context.wrenSupport.registerScriptExports!OneTrickSimianGUI; // Note: for now, only UIElement should be @ScriptExport
        context.wrenSupport.callCreateUI();
        //*/
        
        tabContainer.setTab(0);
        tabPager.setPage(0);
        
        theme = mallocNew!Theme(this);
    }
	
	~this()
	{
		_labelFont.destroyFree();
		_logoFont.destroyFree();
		theme.destroyFree();

        //context.disableWrenSupport();
	}
	
	void getEditionText(char* result, ulong length, const(char)[] pattern)
	{
		auto ver = _client.getPublicVersion();
		snprintf(result, length, pattern.ptr, ver.major, ver.minor, ver.patch);
	}

	auto withId(T:UIElement)(T element, string id)	
	{
		element.setId(StringHelper.duplicateString(cast(char[])id));
		return element;
	}
	UIElement withFormattedId(UIElement element, const(char)[] pattern, ...)
	{
		char[65] newId;
		
		//StringHelper.formatString(newId, pattern, index);

		// More generic, but probably not necessary:
		vsnprintf(newId.ptr, newId.length, pattern.ptr, _argptr);

		element.setId(StringHelper.duplicateString(newId));

		return element;
	}

	bool isModuleEnabled(int index)
	{
		version(Lite)
		{
			switch(index)
			{
				case 2:
				case 3:
				case 4:
				case 5:
				case 9:
					return false;
				default:
			}
		}
		return true;
	}
	void addModule(int index, bool isCymbal)
	{
		int index1 = index+1;
		UIColorPanel panel = null;
		tabPager.addChild(panel = cast(UIColorPanel)withFormattedId(mallocNew!UIColorPanel(context()), "module%dPanel", index1));
		//setIndexedId(panel, "module%dPanel", index1);
		
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleGain, index), "Gain", _labelFont), "module%dKnobGain", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModulePan, index), "Pan", _labelFont), "module%dKnobPan", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleLowpassFreq, index), "Cutoff", _labelFont), "module%dKnobNoise", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleTone, index), "Pitch", _labelFont), "module%dKnobTone", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleBend, index), "Bend", _labelFont), "module%dKnobBend", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleDecay, index), "Decay", _labelFont), "module%dKnobDecay", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModulePunch, index), "Punch", _labelFont), "module%dKnobPunch", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleToneNoise, index), "Tone    Noise", _labelFont), "module%dKnobToneNoise", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleClick, index), "Transient", _labelFont), "module%dKnobClick", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleGainDynamics, index), "Velocity", _labelFont), "module%dKnobGainDynamics", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleBendDynamics, index), "Velocity", _labelFont), "module%dKnobBendDynamics", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleLowpassQ, index), "Resonance", _labelFont), "module%dKnobLowpassQ", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleSaturation, index), "Saturation", _labelFont), "module%dKnobSaturation", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleLowpassEnvBend, index), "Bend", _labelFont), "module%dKnobFilterEnvBend", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleLowpassEnvBendDynamics, index), "Velocity", _labelFont), "module%dKnobFilterEnvBendDynamics", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleOscWaveform, index), "Waveform", _labelFont), "module%dKnobOscWaveform", index1));
		panel.addChild(withFormattedId(mallocNew!UIKnob(context(), _client.moduleParam(paramModuleSendReverb, index), "Room", _labelFont), "module%dKnobSendReverb", index1));
		panel.addChild(withFormattedId(mallocNew!UISeparator(context()), "module%dSeparator1", index1));
		panel.addChild(withFormattedId(mallocNew!UISeparator(context()), "module%dSeparator2", index1));
		panel.addChild(withFormattedId(mallocNew!UISeparator(context()), "module%dSeparator3", index1));
		panel.addChild(withFormattedId(mallocNew!UISeparator(context()), "module%dSeparator4", index1));
		panel.addChild(withFormattedId(mallocNew!UILabel(context(), _labelFont, "Tone"), "module%dLabel1", index1));
		panel.addChild(withFormattedId(mallocNew!UILabel(context(), _labelFont, "Filter"), "module%dLabel2", index1));
		panel.addChild(withFormattedId(mallocNew!UILabel(context(), _labelFont, "Envelope"), "module%dLabel3", index1));
		panel.addChild(withFormattedId(mallocNew!UILabel(context(), _labelFont, "Mix"), "module%dLabel4", index1));
		panel.addChild(withFormattedId(mallocNew!UILabel(context(), _labelFont, "Amp"), "module%dLabel5", index1));
		panel.addChild(withFormattedId(mallocNew!UIConnector(context(), cast(immutable(Parameter)[2])[_client.moduleParam(paramModuleBendDynamics, index), _client.moduleParam(paramModuleBend, index)]), "module%dConnector1", index1));
		panel.addChild(withFormattedId(mallocNew!UIConnector(context(), cast(immutable(Parameter)[2])[_client.moduleParam(paramModuleLowpassEnvBendDynamics, index), _client.moduleParam(paramModuleLowpassEnvBend, index)]), "module%dConnector2", index1));
		panel.addChild(withFormattedId(mallocNew!UIConnector(context(), cast(immutable(Parameter)[2])[_client.moduleParam(paramModuleGainDynamics, index), _client.moduleParam(paramModuleGain, index)]), "module%dConnector3", index1));
		{
			
			auto param = cast(IntegerParameter)_client.moduleParam(paramModuleIcon, index);
			panel.addChild(withFormattedId(mallocNew!UIButton(context(), Icon.Stamp).cyclesParameter(param), "module%dButtonIcon", index1));
			
			param = cast(IntegerParameter)_client.moduleParam(paramModuleColor, index);
			panel.addChild(withFormattedId(mallocNew!UIButton(context(), Icon.Palette).cyclesParameter(param), "module%dButtonColor", index1));
			addChild(withFormattedId(mallocNew!UIParameterReader(context(), param, context().getRootElement()), "module%dParamColor", index1));
		}
		
		{
			addChild(withFormattedId(mallocNew!UIValue(context(), isModuleEnabled(index)), "module%dValueEnabled", index1));
		}
		//panel.addChild(withFormattedId(mallocNew!UIConnector(context()), "module%dConnector4", index1));
		//panel.addChild(withFormattedId(mallocNew!UIConnector(context()), "module%dConnector5", index1));
	}
	
    void onPressed(UIButton button)
    {
   		if (button is creditsButton)
   		{
			creditsPanel.visibility(true);
		}
		else if(button is upgradeButton)
		{
			upgrade();
		}
    }
    
    void upgrade()
    {
	    browse("https://punklabs.com/ot-simian");
	    //browseNoGC("https://punklabs.com/ot-simian"); //Untested
    }
	
	/*
    debug
    {
		override void onAnimate(double dt, double time)
		{
			//context.wrenSupport.callReflowWhenScriptsChange(dt); // Automatic but we need to notify our UI of wren reflow
		    if (context.wrenSupport.reloadScriptsThatChanged())
		    {
				notifyPreScript();
		        context.wrenSupport.callReflow();
				notifyPostScript();
		    }
		}
	}
	*/
	
	/* // Stress Test
	void* testPtr = null;
	override void onAnimate(double dt, double time)
	{
		import core.stdc.stdlib;
		foreach(z; 0..100000)
		{
			if (testPtr !is null)
				free(testPtr);
			testPtr = malloc(z%24+8);
		}
	}
	//*/
	override void parameterChanged()
	{
		if (theme !is null)
			theme.parameterChanged();
	}
	
    override void reflow()
    {
        super.reflow();
   		creditsPanel.position = position;


        int W = position.width;
        int H = position.height;

        float S = W / cast(float)(context.getDefaultUIWidth());
		
        notifyPreScript();

        //context.wrenSupport.callReflow();
        if (theme !is null)
	        theme.reflow();

        notifyPostScript();
    }


	// Publics for Main.d to manipulate
    @ScriptExport
    {
		UIPeakIndicator module1Indicator;
		UIPeakIndicator module2Indicator;
		UIPeakIndicator module3Indicator;
		UIPeakIndicator module4Indicator;
		UIPeakIndicator module5Indicator;
		UIPeakIndicator module6Indicator;
		UIPeakIndicator module7Indicator;
		UIPeakIndicator module8Indicator;
		UIPeakIndicator module9Indicator;
		UIPeakIndicator module10Indicator;
	}
private:
	Theme theme = null;
	Font _labelFont;
	Font _logoFont;
    
	UIPager tabPager;
	
	
    @ScriptExport
    {
    	UIColorPanel background;
    	UIIcon simianIcon;
		UIButton creditsButton;
		UIButton upgradeButton;
		UICreditsPanel creditsPanel;

		UITabContainer tabContainer;

		UITab module1Tab;
		UITab module2Tab;
		UITab module3Tab;
		UITab module4Tab;
		UITab module5Tab;
		UITab module6Tab;
		UITab module7Tab;
		UITab module8Tab;
		UITab module9Tab;
		UITab module10Tab;

		UIColorPanel globalPanel;
		UIKnob globalKnobGain;
		UIKnob globalKnobDrive;
		UIKnob globalKnobSensitivity;
		UIKnob globalKnobSaturation;
		UIKnob globalKnobReverbDecay;
		UIKnob globalKnobReverbGate;
		
	}
}
