/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

module main;

import std.math;
import std.algorithm;
import std.meta;

import dplug.core;
import dplug.core.thread;
import dplug.client;

import gui;

import onetrick;

import dsp_drum;
import dsp_output;

mixin(pluginEntryPoints!PluginClient);

enum Module
{
    Output,
    
    /* //Multi Channel Out
    OutputKick,
    OutputSnare,
	OutputRimshot,
	OutputClap,
	OutputTom1,
	OutputTom2,
	OutputTom3,
	OutputHihatClosed,
	OutputHihatOpen,
	OutputCymbal,
	*/
	
    Kick, FirstKitPiece=Kick,
    Snare,
    Rimshot,
    Clap,
    Tom1,
    Tom2,
    Tom3,
    HihatClosed,
    HihatOpen,
    Cymbal,
    
    Count,
    KitPieceCount=Count-1,
}

enum : int	//################## MUST BE IN ORDER OF DEFINITIONS BELOW!!! ##################
{
	// Waveform CC Bug Workaround:
	paramPlaceholder,
	
	// Globals:
	paramGlobalGain,
	paramGlobalDrive,
	paramGlobalSensitivity,
	paramGlobalSaturation,
	paramGlobalReverbDecay,
	paramGlobalReverbGate,

	// Module:
	paramModuleSTART, // = = = = = = = = = = =
   	paramModuleGain = paramModuleSTART,
   	paramModulePan,
   	paramModuleLowpassFreq,
   	paramModuleTone,
   	paramModuleBend,
   	paramModuleDecay,
   	paramModulePunch,
   	paramModuleToneNoise,
   	paramModuleClick,
   	paramModuleGainDynamics,
   	paramModuleBendDynamics,
   	paramModuleLowpassQ,
   	paramModuleSaturation,
   	paramModuleLowpassEnvBend,
   	paramModuleLowpassEnvBendDynamics,
   	paramModuleOscWaveform,
   	paramModuleSendReverb,
   	paramModuleIcon,
   	paramModuleColor,
	paramModuleEND, // = = = = = = = = = = =

	// Metadata
	paramModuleSTRIDE = paramModuleEND-paramModuleSTART,
	paramModuleCOUNT = Module.KitPieceCount,
   	paramMAX = paramModuleSTART + (paramModuleSTRIDE * paramModuleCOUNT),
}

enum Note
{
	Kick = 36,
	Snare = 38,
	Rimshot = 37,
	Clap = 39,
	Tom1 = 41, Tom1Alt = 40,
	Tom2 = 45, Tom2Alt = 43,
	Tom3 = 48, Tom3Alt = 47,
	HihatClosed = 42,
	HihatPedal = 44,
	HihatOpen = 46,
	Cymbal = 49,
	CymbalChoke = 51,
}

final class PluginClient : OneTrickClient
{
public:
nothrow:
@nogc:

    override void onInitialize()
    {
		int m = 0;
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Output)());

		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Kick
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Snare
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Rimshot
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Clap
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Tom1
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Tom2
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Tom3
		
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Hihat Closed
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Hihat Open
		dspModules[m++] = mallocNew!(OneTrickDSP)(mallocNew!(DSP_Drum)()); //Cymbal

       	allocateAccumulationBuffers(3);
    }
    
    override void onExit()
    {
		foreach(int i; 0..cast(int)dspModules.length)
		{
			dspModules[i].destroyFree();
			dspModules[i] = null;
		}
    }
    
    override float tailSizeInSeconds()
    {
        return 3.0f; // 2s + 1s reverb...?
    }

	Parameter moduleParam(int moduleParamIndex, int moduleIndex)
	{
		return param(paramModuleSTART + (moduleIndex * paramModuleSTRIDE) + (moduleParamIndex - paramModuleSTART));
	}
	
	override void onBuildParameterDescriptions()
	{
		// Bulid Parameter Descriptions. ################## MUST BE IN ORDER OF ENUM ABOVE!!! ##################
		{
			int dspIndex = 0;
			
			// Globals:
			{
				auto dsp = dspModules[dspIndex++];
				// Waveform CC Bug Workaround:
				addParamDescription(ParamDescription("reserved", "", dsp.findParam("")));
				addParamDescription(ParamDescription("global_gain", "dB", dsp.findParam("Gain_Global")));
				addParamDescription(ParamDescription("global_drive", "dB", dsp.findParam("Drive_Global")));
				addParamDescription(ParamDescription("global_sensitivity", "%", FaustParam("", null, 1, 0, 1, 0.001f))); //dsp.findParam("Sensitivity_Global")
				addParamDescription(ParamDescription("global_saturation", "%", dsp.findParam("Saturation_Global")));
				addParamDescription(ParamDescription("global_reverbdecay", "%", dsp.findParam("ReverbDecay_Global")));
				addParamDescription(ParamDescription("global_reverbgate", "dB", dsp.findParam("ReverbGate_Global")));
		    }
			
			// Modules:
			int index=0;
			foreach(prefix; AliasSeq!("kick", "snare","rim","clap","tom1","tom2","tom3","hhclosed","hhopen","cymbal"))
			{
				bool isCymbal = (dspIndex > 7);
				auto dsp = dspModules[dspIndex++];
				addParamDescription(ParamDescription(prefix~"_gain", "dB", dsp.findParam("Gain")));
				addParamDescription(ParamDescription(prefix~"_pan", "%", dsp.findParam("Pan")));
				addParamDescription(ParamDescription(prefix~"_lowpassfreq", "Hz", dsp.findParam("LowpassFreq")));
				addParamDescription(ParamDescription(prefix~"_oscpitch", "Hz", dsp.findParam("OscPitch")));
				addParamDescription(ParamDescription(prefix~"_oscbend", "st", dsp.findParam("OscEnvBend")));
				addParamDescription(ParamDescription(prefix~"_envdecay", "ms", dsp.findParam("EnvDecay")));
				addParamDescription(ParamDescription(prefix~"_envpunch", "%", dsp.findParam("EnvPunch")));
				addParamDescription(ParamDescription(prefix~"_tonenoise", "%", dsp.findParam("FadeToneNoise")));
				addParamDescription(ParamDescription(prefix~"_click", "%", dsp.findParam("FadeClick")));
				addParamDescription(ParamDescription(prefix~"_gaindynamics", "%", dsp.findParam("GainDynamics")));
				addParamDescription(ParamDescription(prefix~"_oscbenddynamics", "%", dsp.findParam("OscEnvBendDynamics")));
				addParamDescription(ParamDescription(prefix~"_lowpassq", "%", dsp.findParam("LowpassQ")));
				addParamDescription(ParamDescription(prefix~"_saturation", "%", dsp.findParam("Saturation")));
				addParamDescription(ParamDescription(prefix~"_lowpassbend", "st", dsp.findParam("LowpassEnvBend")));
				addParamDescription(ParamDescription(prefix~"_lowpassbenddynamics", "%", dsp.findParam("LowpassEnvBendDynamics")));
				auto waveformParam = dsp.findParam("OscWaveform");
				if(isCymbal)
					waveformParam.initial = 100;
				addParamDescription(ParamDescription(prefix~"_oscwaveform", "%", waveformParam));
				addParamDescription(ParamDescription(prefix~"_reverb", "%", dsp.findParam("SendReverb")));
				const immutable(int)[10] defaultIcons = [0,1,2,3,4,5,6,7,8,7];
				addParamDescription(ParamDescription(prefix~"_icon", "", FaustParam("", null, defaultIcons[index], 0, 8, 1))); //Default to GenericDrum,/Cymbal
				addParamDescription(ParamDescription(prefix~"_color", "", FaustParam("", null, index%8, 0, 7, 1))); //Default to Rainbow
				index++;
			}
		}
    }

    override LegalIO[] buildLegalIO()
    {
        auto io = makeVec!LegalIO();
        
        io ~= LegalIO(0, 1); // Mono
        io ~= LegalIO(0, 2); // Stereo
        
        return io.releaseData();
    }

    // This override is optional, the default implementation will
    // have one default preset.
    override Preset[] buildPresets()
    {
    	auto internalPresets = makeVec!Preset();
    	
    	// Internal Presets
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Basic.preset")));
		    preset.setName("Basic");
		    PresetHelper.applyPreset(this, preset, false);
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Factorial Fiction.preset")));
		    preset.setName("Factorial Fiction");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Passive Aggressive.preset")));
		    preset.setName("Passive Aggressive");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Sequence.preset")));
		    preset.setName("Sequence");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/No Problem.preset")));
		    preset.setName("No Problem");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Bar Fight.preset")));
		    preset.setName("Bar Fight");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Fuchsia Sunrise.preset")));
		    preset.setName("Fuchsia Sunrise");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Fuchsia Sunset.preset")));
		    preset.setName("Fuchsia Sunset");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Sandy Fuego.preset")));
		    preset.setName("Sandy Fuego");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Undeadhead.preset")));
		    preset.setName("Undeadhead");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Italo Disco.preset")));
		    preset.setName("Italo Disco");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Disco Sol.preset")));
		    preset.setName("Disco Sol");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Discopsychology.preset")));
		    preset.setName("Discopsychology");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Clique Beetle.preset")));
		    preset.setName("Clique Beetle");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Suburbanight.preset")));
		    preset.setName("Suburbanight");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Eticroquette.preset")));
		    preset.setName("Eticroquette");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Whatever.preset")));
		    preset.setName("Whatever");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Dustbin.preset")));
		    preset.setName("Dustbin");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Thrashcan.preset")));
		    preset.setName("Thrashcan");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Shhhhh.preset")));
		    preset.setName("Shhhhh");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Aquatic Traffic.preset")));
		    preset.setName("Aquatic Traffic");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Boomsticks.preset")));
		    preset.setName("Boomsticks");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Alpacalypse.preset")));
		    preset.setName("Alpacalypse");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Sasquelch.preset")));
		    preset.setName("Sasquelch");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Life Support.preset")));
		    preset.setName("Life Support");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Hootenanny.preset")));
		    preset.setName("Hootenanny");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Raptor Hatchling.preset")));
		    preset.setName("Raptor Hatchling");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Oh Shoot.preset")));
		    preset.setName("Oh Shoot");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Bubble Factory.preset")));
		    preset.setName("Bubble Factory");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Macrochip.preset")));
		    preset.setName("Macrochip");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Power Up.preset")));
		    preset.setName("Power Up");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Power Down.preset")));
		    preset.setName("Power Down");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Duh Duh Duh.preset")));
		    preset.setName("Duh Duh Duh");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Tommy Tom Tom.preset")));
		    preset.setName("Tommy Tom Tom");
		    internalPresets ~= preset;
		}
		{ // Default
			Preset preset = makeDefaultPreset();
			preset.setName("Init");
	        internalPresets ~= preset;
	    }

		_internalPresetBank = mallocNew!PresetBank(this, internalPresets.releaseData());

    
    	// Public Presets
        auto presets = makeVec!Preset();

		// REQUIRED: Dplug:VST3 can overwrite bank[_current] without safety checks.
		{ // Default
			Preset preset = makeDefaultPreset();
			preset.setName("Init");
	        presets ~= preset;
	    }

        return presets.releaseData();
    }
	override PresetBank internalPresetBank()
	{
		return _internalPresetBank;
	}


    override int maxFramesInProcess()
    {
	    //Dplug automatically splits buffers on MidiEvent bounds, so the main concern is automation resolution.
	    //256 is approx 172hz
        return 256;
    }

    override void reset(double sampleRate, int maxFrames, int numInputs, int numOutputs) 
    {
    	foreach(dsp; dspModules)
    		dsp.initialize(cast(int)sampleRate, maxFrames).trackSilence(100);
    }

	override void onMidiInput(MidiMessage msg)
	{
		/* // We could affect global decay with the Modwheel...?
    	if (msg.isControlChange() && (msg.controlChangeControl() == MidiControlChange.modWheel)) {
    		modWheel = msg.controlChangeValue0to1();
    	} else
    	*/
    	if (msg.isNoteOn()) {
        	auto velocity = cast(float)(msg.noteVelocity()+1.0f)/128.0f;
        	velocity = 1.0f + (velocity-1.0f) * readParam!float(paramGlobalSensitivity);
        	switch(msg.noteNumber())
        	{
        		case Note.Kick:
        			dspModules[Module.Output].wakeUp();
        			dspModules[Module.Kick].trigger(velocity);
            		break;
        		case Note.Snare:
        			dspModules[Module.Output].wakeUp();
        			dspModules[Module.Snare].trigger(velocity);
            		break;
        		case Note.Tom3: case Note.Tom3Alt:
        			dspModules[Module.Output].wakeUp();
					dspModules[Module.Tom3].trigger(velocity);
            		break;
        		case Note.HihatClosed:
        			dspModules[Module.Output].wakeUp();
        			dspModules[Module.HihatClosed].trigger(velocity);
        			dspModules[Module.HihatOpen].choke();
            		break;
        		case Note.HihatPedal:
        			dspModules[Module.HihatClosed].choke();
        			dspModules[Module.HihatOpen].choke();
            		break;
        		case Note.HihatOpen:
        			dspModules[Module.Output].wakeUp();
        			dspModules[Module.HihatOpen].trigger(velocity);
        			dspModules[Module.HihatClosed].choke();
            		break;
       			version(Lite) { }
       			else {
        		case Note.Rimshot:
        			dspModules[Module.Output].wakeUp();
        			dspModules[Module.Rimshot].trigger(velocity);
            		break;
        		case Note.Clap:
        			dspModules[Module.Output].wakeUp();
        			dspModules[Module.Clap].trigger(velocity);
            		break;
        		case Note.Tom1: case Note.Tom1Alt:
        			dspModules[Module.Output].wakeUp();
        			dspModules[Module.Tom1].trigger(velocity);
            		break;
        		case Note.Tom2: case Note.Tom2Alt:
        			dspModules[Module.Output].wakeUp();
        			dspModules[Module.Tom2].trigger(velocity);
            		break;
            	case Note.Cymbal:
        			dspModules[Module.Output].wakeUp();
        			dspModules[Module.Cymbal].trigger(velocity);
            		break;
        		case Note.CymbalChoke:
        			dspModules[Module.Cymbal].choke();
            		break;
            	}
            	default:
        	}
        }
        /* // Not needed...
        else if (msg.isAllNotesOff() || msg.isAllSoundsOff()) {
        	dspKick.choke();
        	dspSnare.choke();
        	dspTom1.choke();
        	dspTom2.choke();
        	dspTom3.choke();
        	dspHihatClosed.choke();
        	dspHihatOpen.choke();
        	dspCymbal.choke();
        }
        */
	}

	override void onUpdateDspParameters()
	{
		writeDspParameters();
	}
	
    override void onProcessAudio(const(float*)[] inputs, float*[]outputs, int frames, TimeInfo info)
    {
		if (dspModules[Module.Output].isSilent())
		{
			foreach (int chan; 0..cast(int)outputs.length)
				outputs[chan][0..frames] = 0;
		}
		else
		{
			clearAccumulation(frames);
			
			int from = Module.FirstKitPiece;
			int count = Module.KitPieceCount;
	        computeDspModules(dspModules[from .. from+count], frames);
	
			foreach(int m; 0..Module.KitPieceCount)
			{
				auto dsp = dspModules[Module.FirstKitPiece + m];
				if(!dsp.wasComputeSkipped()) {
					addBuffers(frames, dspBufferAccum.opSlice(), dsp.TempBuffer);
				}
			}


		    dspModules[Module.Output].compute(frames, cast(float*[])dspBufferAccum.opSlice(), outputs);

		}
	}
	
    override void onAccessGuiUnsafe(CustomGUI gui)
    {
    	auto pluginGUI = cast(PluginGUI)gui;
        pluginGUI.module1Indicator.setValueFromDSP(dspModules[Module.Kick].triggered()?1:0);
        pluginGUI.module2Indicator.setValueFromDSP(dspModules[Module.Snare].triggered()?1:0);
        pluginGUI.module3Indicator.setValueFromDSP(dspModules[Module.Rimshot].triggered()?1:0);
        pluginGUI.module4Indicator.setValueFromDSP(dspModules[Module.Clap].triggered()?1:0);
        pluginGUI.module5Indicator.setValueFromDSP(dspModules[Module.Tom1].triggered()?1:0);
        pluginGUI.module6Indicator.setValueFromDSP(dspModules[Module.Tom2].triggered()?1:0);
        pluginGUI.module7Indicator.setValueFromDSP(dspModules[Module.Tom3].triggered()?1:0);
        pluginGUI.module8Indicator.setValueFromDSP(dspModules[Module.HihatClosed].triggered()?1:0);
        pluginGUI.module9Indicator.setValueFromDSP(dspModules[Module.HihatOpen].triggered()?1:0);
        pluginGUI.module10Indicator.setValueFromDSP(dspModules[Module.Cymbal].triggered()?1:0);
    }

    override IGraphics createGraphics()
    {
        return mallocNew!PluginGUI(this);
    }
    
    ParamDescription getParamDesc(int index)
    {
	    return paramDesc[index];
    }

private:
	PresetBank _internalPresetBank;
	//float modWheel = 0;
	

    OneTrickDSP[Module.Count] dspModules;
    
}

