/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.client.midi;

import dplug.client;

import onetrick.client.faust;
import onetrick.string;

class MidiKeySplit
{
public:
nothrow:
@nogc:
	this(int keyStart, int keyCount, FaustParam keyParam, FaustParam velocityParam, FaustParam gateParam)
	{
		this.keyStart = keyStart;
		this.keyCount = keyCount;
		this.keyParam = keyParam;
		this.velocityParam = velocityParam;
		this.gateParam = gateParam;
	}
	MidiKeySplit withKeyAsRatio()
	{
		sendKeyAsRatio = true;
		return this;
	}
	
	void processMidiMessage(MidiMessage msg)
	{
		if (msg.isNoteOn() || msg.isNoteOff())
		{
			int noteNumber = msg.noteNumber();
			if (noteNumber >= keyStart && noteNumber < (keyStart + keyCount))
			{
				if (msg.isNoteOn())
				{
					tableTimestamp[noteNumber] = keyCounter++;
					tableVelocity[noteNumber] = msg.noteVelocity();
					tableGate[noteNumber] = true;
				}
				else if(msg.isNoteOff())
				{
					//tableVelocity[noteNumber] = msg.noteVelocity(); // Release Velocity, not using now
					tableTimestamp[noteNumber] = -1;
					tableGate[noteNumber] = false;
				}
				onUpdateDspParameters();
			}
		}
	}
	
	void onUpdateDspParameters()
	{
		activeKey = -1;
		long activeTimestamp = -1;
		for(int i=keyStart; i<(keyStart+keyCount); i++)
		{
			if(tableGate[i] && tableTimestamp[i] >= activeTimestamp)
			{
				activeKey = i;
				activeTimestamp = tableTimestamp[i];
				activeVelocity = tableVelocity[i];
			}
		}
		if (activeKey >= 0 && keyParam.val !is null)
		{
			if (sendKeyAsRatio)
			{
				*(keyParam.val) = (keyCount>1) ? cast(float)(activeKey-keyStart)/cast(float)(keyCount-1) : 1;
			}
			else
			{
				*(keyParam.val) = activeKey;
			}
		}
		if (velocityParam.val !is null)
			*(velocityParam.val) = activeVelocity / 127.0f;
		if (gateParam.val !is null)
			*(gateParam.val) = (activeKey >= 0);
		
			
	}
	
private:
	long keyCounter = 0;
	const int MAX_KEYS = 128;
	int keyStart = 0;
	int keyCount = MAX_KEYS;
	long[MAX_KEYS] tableTimestamp;
	int[MAX_KEYS] tableVelocity;
	bool[MAX_KEYS] tableGate;
	
	FaustParam keyParam;
	FaustParam velocityParam;
	FaustParam gateParam;

	int activeKey = -1;
	float activeVelocity = 0;
	bool sendKeyAsRatio = false;
}

