/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.preset.helper;

import gui;
import core.atomic;
import dplug.core;
import dplug.client;

import core.stdc.stdio;
import core.stdc.string;
import dplug.core.nogc;
import core.stdc.stdlib;
import std.bitmanip;

import onetrick.client;

/// Helper methods for Preset Management
class PresetHelper
{
public:
nothrow:
@nogc:

	static int countFactoryPresets(OneTrickClient client)
	{
		return client.internalPresetBank.numPresets();
	}
	static Preset getFactoryPreset(OneTrickClient client, int index)
	{
		auto bank = client.internalPresetBank;
		if (bank.isValidPresetIndex(index))
			return bank.preset(index);
		return null;
	}
	/*
	static const(char)[] getFactoryPresetName(OneTrickClient client, int index)
	{
		Preset preset = getFactoryPreset(client, index);
		return (preset !is null) ? CString(preset.name()).storage : "User";
	}
	*/
	static void applyPresetFromGUI(OneTrickClient client, Preset preset)
	{
		applyPreset(client, preset, true);
	}
	static void applyPresetFromDSP(OneTrickClient client, Preset preset)
	{
		applyPreset(client, preset, false);
	}
	static void applyPreset(OneTrickClient client, Preset preset, bool fromGUI)
	{
		if (preset !is null)
		{
			float[] values = preset.getNormalizedParamValues();
			Parameter[] params = client.params();
			for(int i=0; i<params.length; i++)
			{
				Parameter param = params[i];
				if(param)
				{
					float value = (i<values.length) ? values[i] : param.getNormalizedDefault();
					
					// Values outside this will trigger an Assert in Dplug
					if (value < 0)
						value = 0;
					else if (value > 1)
						value = 1;

					setParamNormalized(param, value, fromGUI);
				}
			}
		}
	}
	
	static void setParamNormalized(Parameter param, double value, bool fromGUI)
	{
		version(LV2)
		{
		/*
			Problem: Ardour/Qtractor crash when changing presets with automation.
			Cause: Threading issues with the SUIL library, or integration of the library.
			Hack 1: Don't send automation updates... Doesn't work, plugin won't save state
			Hack 2: Thread synchronize on our end... Doesn't work, still crashes
			Giving up on LV2 support for the time being...
		*/
		}
	
		if (fromGUI)
			param.beginParamEdit();
			
		if (auto p = cast(FloatParameter)param)
		{
			if (fromGUI)
				p.setFromGUINormalized(value);
			else
				p.setNormalized(value);
		}
		else if (auto p = cast(IntegerParameter)param)
		{
			if (fromGUI)
				p.setFromGUINormalized(value);
			else
				p.setNormalized(value);
		}
		else if (auto p = cast(BoolParameter)param)
		{
			if (fromGUI)
				p.setFromGUI(value > 0.5);
			else
				p.setNormalized(value);
		}
			
		if (fromGUI)
			param.endParamEdit();
	}

	static Preset saveBinaryPresetToPath(OneTrickClient client, const(char)[] path)
	{
        Preset preset = client.makeDefaultPreset();
        preset.setName("User Preset"); //TODO: Use Filename
		return saveBinaryPresetToPath(client, preset, path);
	}	
	static Preset saveBinaryPresetToPath(OneTrickClient client, Preset preset, const(char)[] path)
	{
		FILE* file = fopen(CString(path).storage, "w".ptr); //"wb" for binary
		if (file)
		{
			scope(exit) fclose(file);
	        int index=0;
			foreach(p; client.params())
			{
				float value = p.getNormalized();
				preset.setNormalized(index, value);
				ubyte[4] value4 = nativeToBigEndian!float(value);
				fwrite(&value4, 4, 1, file);
				index++;
			}
			return preset;
		}
		return null;
	}

	static Preset loadBinaryPresetFromPath(OneTrickClient client, const(char)[] path)
	{
		auto data = readFile(path);
		scope(exit) free(data.ptr);
		return loadBinaryPresetFromMemory(client, data);
	}
	static Preset loadBinaryPresetFromMemory(OneTrickClient client, ubyte[] data)
	{
		if (data !is null)
		{
	        Preset preset = client.makeDefaultPreset();
	        preset.setName("User Preset"); //TODO: Use Filename
			
			Parameter[] params = client.params();
			for (int i=0; i <= data.length-4; i+=4)
			{
				int p = i/4;
				if (p < params.length)
				{
					//float value = bigEndianToNative!float(data[i..i+4]);
					// Couldn't quite figure out Dlang casting... *SIGH*
					float value = bigEndianToNative!float([data[i],data[i+1],data[i+2],data[i+3]]);
					//setParamNormalized(params[p], value);
					preset.setNormalized(p, value);
				}
			}
			return preset;
		}
		return null;
	}
	
	
private:

}
