/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.string.helper;

import gui;
import core.atomic;
import dplug.core;
import dplug.client;

import core.stdc.stdio;
import core.stdc.string;
import dplug.core.nogc;
import core.stdc.stdlib;

/// Helper methods for String management
class StringHelper
{
public:
nothrow:
@nogc:

	static void formatString(char[] result, const(char)[] pattern, ...)
	{
		vsnprintf(result.ptr, result.length, CString(pattern).storage, _argptr); // use the v_ form of snprintf to allow D variable args
	}
	
	static char[] getStringSlice(char* value)
	{
		if (value !is null)
			return value[0..strlen(value)];
		else
			return null;
	}
	static char[] getStringSlice(char[] value)
	{
		return value[0..strlen(value.ptr)];
	}
	
	static char[] duplicateString(char* value)
	{
		if (value !is null)
			return duplicateString(getStringSlice(value));
		else
		{
			return duplicateString(cast(char[])"");
		}
	}
	static char[] duplicateString(const(char)[] value)
	{
		return duplicateString(cast(char[])value);
	}
	static string duplicateString(const(string) value)
	{
		return cast(string)duplicateString(cast(char[])value);
	}
	static char[] duplicateString(char[] value)
	{
		if (value is null)
		{
			auto blank = cast(char*)malloc(char.sizeof);
			blank[0] = '\0';
			return getStringSlice(blank);
		}
		value = getStringSlice(value); //Trim to null terminator
		auto newCopy = cast(char*)malloc(char.sizeof * (value.length+1));
		strcpy(newCopy, value.ptr);
		return newCopy[0..value.length];
	}
	static void freeString(char[] value)
	{
		if (value !is null)
			free(value.ptr);
	}
	static void freeString(char* value)
	{
		if (value !is null)
			free(value);
	}
	
	static bool equals(string a, string b)
	{
		return strcmp(a.ptr, b.ptr) == 0;
	}
	static bool equals(char[] a, char[] b)
	{
		return strcmp(CString(a).storage, CString(b).storage) == 0;
	}
	static bool equals(char* a, char* b)
	{
		//return strcmp(CString(cast(char[])*a).storage, CString(cast(char[])*b).storage) == 0;
		return strcmp(a, b) == 0;
	}
}
