/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.activatable;

interface IActivatable
{
nothrow:
@nogc:
	void setActive(bool active);
	bool isActive();
	bool isActiveSelf();
}

mixin template Activatable()
{
	private bool _isActive = false;
	nothrow @nogc public void setActive(bool active)
	{
		if (_isActive != active)
		{
			_isActive = active;
			setDirtyWhole();
		}
	}
	nothrow @nogc public bool isActiveSelf()
	{
		return _isActive;
	}
	nothrow @nogc public bool isActive()
	{
		if (isActiveSelf())
			return true;
		UIElement p = this;
		while ((p = p.parent) !is null)
		{
			auto ia = (cast(IActivatable)p);
			if (ia !is null && ia.isActiveSelf())
				return true;
		}
		return false;
	}
}
