/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.button;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;
import dplug.graphics.font;

import onetrick.ui.helper;

interface IButtonListener
{
nothrow:
@nogc:

    void onPressed(UIButton button);
}

class UIButton : UIElement
{
public:
nothrow:
@nogc:

	// TODO: Move these to regular @properties that can setDirty
	@ScriptProperty float iconSize = 0.8f;

	@ScriptProperty RGBA backgroundColor = RGBA(0, 0, 0, 0);
	@ScriptProperty RGBA backgroundColorHover = RGBA(255, 255, 255, 64);
	@ScriptProperty RGBA backgroundColorPressed = RGBA(0, 0, 0, 32);
	
	@ScriptProperty RGBA color = RGBA(200, 200, 200, 255);
	@ScriptProperty RGBA colorHover = RGBA(255, 255, 255, 255);
	@ScriptProperty RGBA colorPressed = RGBA(255, 255, 255, 255);

	@ScriptProperty RGBA shadowColor = RGBA(0, 0, 0, 0);
	@ScriptProperty float shadowOffset = 0.1f;

	this(UIContext context, Icon icon)
	{
		super(context, flagRaw);
		this.icon = icon;
        _listeners = makeVec!IButtonListener();
	}
	~this()
	{
		_listeners.clearContents();
		//_listeners is struct, so it should automatically free...
	}
	
	UIButton cyclesParameter(IntegerParameter param, int step = 1)
	{
		this.param = param;
		this.paramStep = step;

		//param.beginParamEdit();
		//param.setFromGUI(2);//val
		//param.endParamEdit();

		return this;
	}
	
    /// Adds a listener.
    void addListener(IButtonListener listener)
    {
        _listeners.pushBack(listener);
    }

    /// Removes a listener.
    void removeListener(IButtonListener listener)
    {
        int index = _listeners.indexOf(listener);
        if (index != -1)
            _listeners.removeAndReplaceByLastElement(index);
    }


	void setIcon(Icon newIcon)
	{
		icon = newIcon;
		setDirtyWhole();
	}
	
    void emitPressed()
    {
        foreach(listener; _listeners)
            listener.onPressed(this);
    }
    void pressed()
    {
    	if(paramStep != 0)
    	{
			auto intParam = cast(IntegerParameter)param;
			if(intParam !is null)
			{
				intParam.beginParamEdit();
				if(shouldResetParam)
				{
					intParam.setFromGUI(intParam.defaultValue());
				}
				else
				{
					int min = intParam.minValue();
					int max = intParam.maxValue();
					int val = intParam.value();
					val += paramStep;
					if (val > max)
						val = min;
					else if (val < min)
						val = max;
					intParam.setFromGUI(val);
				}
				intParam.endParamEdit();
			}
		}
    	emitPressed();
    }
    
	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		bool isPressed = isDragged;//(isMouseOver && isDragged);
		RGBA currentBackgroundColor = isPressed ? backgroundColorPressed : (isMouseOver ? backgroundColorHover : backgroundColor);
		RGBA currentColor = isPressed ? colorPressed : (isMouseOver ? colorHover : color);
		
		UIHelper.fillColorRaw(rawMap, dirtyRects, currentBackgroundColor);

		if (shadowColor.a > 0)
		{
			int y = cast(int)(position.size.y*shadowOffset);
			if (y != 0)
				UIHelper.drawIconRaw(rawMap, dirtyRects, box2f.rectangle(0, y, position.size.x, position.size.y), icon, shadowColor, iconSize);
		}
		
		UIHelper.drawIconRaw(rawMap, dirtyRects, box2f.rectangle(0, 0, position.size.x, position.size.y), icon, currentColor, iconSize);
	}

	override void onMouseEnter()
	{
		setDirtyWhole();
	}

	override void onMouseExit()
	{
		setDirtyWhole();
	}


	override Click onMouseClick(int x, int y, int button, bool isDoubleClick, MouseState mstate)
	{
		shouldRegisterPress = true;
		shouldResetParam = mstate.altPressed;
		/*
		{
			//Simple Version, doesn't feel as nice (no click color)
			setDirtyWhole();
			pressed();
			return Click.handled;
		}
		*/
		return Click.startDrag;
	}

	override void onBeginDrag()
	{
		setDirtyWhole();
	}
	override void onMouseDrag(int x, int y, int dx, int dy, MouseState mstate)
	{
		shouldRegisterPress = contains(x,y);
		setDirtyWhole();
	}
	override void onStopDrag()
	{
		if (shouldRegisterPress) // Not actually 
			pressed();
		setDirtyWhole();
	}

private:
	bool shouldRegisterPress = false;
	bool shouldResetParam = false;
	IntegerParameter param;
	int paramStep = 0;
    Vec!IButtonListener _listeners;
	Icon icon;

}
