/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.connector;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import onetrick.ui.helper;

/// A simple visual connector
final class UIConnector : UIElement, IParameterListener
{
public:
nothrow:
@nogc:

	enum Style {
		VerticalLeft2,
		VerticalRight2,
		HorizontalUp2,
		HorizontalDown2,
	}
	// TODO: Move these to regular @properties that can setDirty
	@ScriptProperty int style = Style.VerticalLeft2;
	@ScriptProperty RGBA color = RGBA(255, 255, 255, 255);
	@ScriptProperty RGBA colorEditing = RGBA(255, 255, 0, 255);
	@ScriptProperty float lineWidth = 2;

	this(UIContext context)
	{
		super(context, flagRaw);
	}
	this(UIContext context, Parameter param)
	{
		super(context, flagRaw);// | flagAnimated);
		
		this.params ~= param;
		param.addListener(this);
	}
	this(UIContext context, immutable Parameter[] params)
	{
		super(context, flagRaw);// | flagAnimated);
		
		foreach(param; params)
			this.params ~= cast(Parameter)param;
			
		foreach(param; this.params)
		{
			param.addListener(this);
		}
	}
	~this()
	{
		foreach(param; params)
		{
			param.removeListener(this);
		}
	}
	
	void addParam(Parameter param)
	{
		params ~= param;
		param.addListener(this);
	}
	
	override void onBeginParameterHover(Parameter sender) { }
	override void onEndParameterHover(Parameter sender) { } 
	override void onBeginParameterEdit(Parameter sender)
	{
		atomicStore(__parameterEditing, true);
		setDirtyWhole();
	}
	override void onParameterChanged(Parameter sender)
	{
	}
	override void onEndParameterEdit(Parameter sender)
	{
		atomicStore(__parameterEditing, false);
		setDirtyWhole();
	}	
	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		auto r = position;
		auto size = position.size;
		bool isEditing = atomicLoad(__parameterEditing);
		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			canvas.initialize(croppedRaw);
			canvas.translate(-dirtyRect.min.x, -dirtyRect.min.y); // Offset so we're using a consistent coordinate system
			
			canvas.fillStyle = isEditing ? colorEditing : color;
			switch (style)
			{
			case Style.VerticalLeft2:
			case Style.VerticalRight2:
				if(style == Style.VerticalRight2)
					flipX();
				canvas.beginPath();
				canvas.moveTo(0, 0);
				canvas.lineTo(vec2f(size.x, 0));
				canvas.lineTo(vec2f(size.x, size.y));
				canvas.lineTo(vec2f(0, size.y));
				canvas.lineTo(vec2f(0, size.y-lineWidth));
				canvas.lineTo(vec2f(size.x-lineWidth, size.y-lineWidth));
				canvas.lineTo(vec2f(size.x-lineWidth, lineWidth));
				canvas.lineTo(vec2f(0, lineWidth));
				canvas.closePath();
				canvas.fill();
				break;
			case Style.HorizontalUp2:
			case Style.HorizontalDown2:
				if(style == Style.HorizontalDown2)
					flipY();
				canvas.beginPath();
				canvas.moveTo(0, 0);
				canvas.lineTo(vec2f(0, size.y));
				canvas.lineTo(vec2f(size.x, size.y));
				canvas.lineTo(vec2f(size.x, 0));
				canvas.lineTo(vec2f(size.x-lineWidth, 0));
				canvas.lineTo(vec2f(size.x-lineWidth, size.y-lineWidth));
				canvas.lineTo(vec2f(lineWidth, size.y-lineWidth));
				canvas.lineTo(vec2f(lineWidth, 0));
				canvas.closePath();
				canvas.fill();
				break;
			default:
			}
		}
	}
private:
    shared(bool) __parameterEditing = false;  // access to this is through atomic ops
	Canvas canvas;
	Vec!Parameter params = makeVec!Parameter();

	void flipX()
	{
		int halfSize = position.size.x/2;
		canvas.translate(halfSize, 0);
		canvas.scale(-1, 1);
		canvas.translate(-halfSize, 0);
	}
	void flipY()
	{
		int halfSize = position.size.y/2;
		canvas.translate(0, halfSize);
		canvas.scale(1, -1);
		canvas.translate(0, -halfSize);
	}
}
