/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.helper;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;
import dplug.graphics.font;

import dplug.graphics.draw;
import std.math; //PI
import std.algorithm;

import core.stdc.stdlib: malloc, realloc, free;
import core.stdc.stdio;
import core.stdc.string;

import onetrick.string.helper;



enum Icon
{
	//PluginIcons:
	Plugin_Circle, // Circle to calibrate font ascent
	Plugin_Icon1,
	Plugin_Icon2,
	Plugin_Icon3,
	Plugin_Icon4,
	Plugin_Icon5,
	Plugin_Icon6,
	Plugin_Icon7,
	Plugin_Icon8,
	Plugin_Icon9,
	Plugin_Icon10,
	Plugin_Icon11,
	Plugin_Icon12,
	Plugin_Icon13,
	Plugin_Icon14,
	Plugin_Icon15,
	Plugin_Icon16,
	
	//OneTrickIcons:
	OT_Circle, Plugin_MAX=OT_Circle, // Circle to calibrate font ascent
	OT_KickDrum,
	OT_SnareDrum,
	OT_Claves,
	OT_DrumSticks,
	OT_TomDrumSmall,
	OT_TomDrumMedium, OT_TomDrum=OT_TomDrumMedium,
	OT_TomDrumLarge,
	OT_Cymbal,
	OT_Hihat,
	OT_OneTrickIcon,
	OT_Upgrade,
	
	//FontAwesome:
	None, OT_MAX=None,
	New,
	New2,
	Duplicate,
	Open,
	Save,
	Edit,
	Delete,
	Delete2,
	Refresh,
	Add,
	Remove,
	Accept,
	Cancel,
	Help,
	Info,
	Settings,
	Settings2,
	Settings3,
	Settings4,
	Settings5,
	ZoomIn,
	ZoomOut,
	Up,
	Down,
	Left,
	Right,
	Up2,
	Down2,
	Left2,
	Right2,
	Up3,
	Down3,
	Left3,
	Right3,
	Up4,
	Down4,
	Left4,
	Right4,
	Up5,
	Down5,
	Left5,
	Right5,
	Up6,
	Down6,
	Left6,
	Right6,
	KickDrum,
	SnareDrum,
	SnareRim,
	Wand,
	MagicWand,
	SteelDrum,
	GenericDrum,
	Rim,
	Cymbal,
	Hand,
	Clap,
	Clap2,
	Cow,
	Bell,
	Guitar,
	Disc,
	Favorite,
	Flag,
	Circle,
	Copyright,
	Trademark,
	Registered,
	Home,
	Cut,
	Share,
	Freeze,
	SquareWave,
	Volume1,
	Volume3,
	VolumeMute,
	Mic,
	MicMute,
	PaintBrush,
	PaintBrush2,
	PaintRoller,
	Spraypaint,
	Stamp,
	Palette,
	Background,
	Image,
	Fire,
	Screwdriver,
	Icons,
	MAX,
}

const string[Icon.MAX] IconChar = [
	//PluginIcons: (16+1)
	"A", // Circle to calibrate font ascent
	"B",
	"C",
	"D",
	"E",
	"F",
	"G",
	"H",
	"I",
	"J",
	"K",
	"L",
	"M",
	"N",
	"O",
	"P",
	"Q",
	//OneTrickIcons: (13)
	"A", // Circle to calibrate font ascent
	"B",
	"C",
	"D",
	"E",
	"F",
	"G",
	"H",
	"I",
	"J",
	"K",
	"L",
	//FontAwesome:
	"",
	"📄",//📄🗋
	"",
	"",
	"📂",//📂🗁
	"",//💾
	"📝",
	"",//
	"⌫",
	"🗘",//🗘
	"",
	"",
	"",//
	"",//
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"🥁",//🥁
	"🔘",
	"",
	"",
	"",
	"",
	"",
	"",
	"✋",
	"",
	"",
	"",
	"🔔",
	"",
	"💿",
	"♥",//♥❤🤎💙💚💛🧡💜🖤🤍
	"",
	"",//🟠🟡🟢🟣🟤
	"",
	"",
	"",
	"",
	"✀",//✀✂✄
	"",
	"",
	"",
	"🔈",
	"🔊 ",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"🔥",
	"🪛",
	"",
];

/// Helper methods for drawing...
class UIHelper
{
public:
nothrow:
@nogc:

	static void initialize() // Ideally should be called early to avoid race conditions on first draw
	{
		ensureResources();
	}
	private static void ensureResources()
	{
		if (fontPluginIcons is null)
			fontPluginIcons = mallocNew!Font(cast(ubyte[])(import("Fonts/PluginIcons/PluginIcons.ttf")));
		if (fontOneTrickIcons is null)
			fontOneTrickIcons = mallocNew!Font(cast(ubyte[])(import("Fonts/OneTrickIcons/OneTrickIcons.ttf")));
		if (fontAwesomeIcons is null)
			//fontIcons = mallocNew!Font(cast(ubyte[])(import("Fonts/icon-works_v1/icon-works-webfont.ttf")));
			fontAwesomeIcons = mallocNew!Font(cast(ubyte[])(import("Fonts/Font-Awesome-6.x/webfonts/fa-solid-900.ttf")));
	}
	
	static void fillColorRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, RGBA color)
	{
		if (color.a >= 255) //Overwrite, faster
		{
			foreach(dirtyRect; dirtyRects)
		    {
		        for (int y = dirtyRect.min.y; y < dirtyRect.max.y; ++y)
		        {
		            RGBA[] rawScan = rawMap.scanline(y);
		            rawScan[dirtyRect.min.x .. dirtyRect.max.x] = color;
		            /*
		            for (int x = dirtyRect.min.x; x < dirtyRect.max.x; ++x)
		            {
		                rawScan.ptr[x] = color;
		            }
		            */
		        }
		    }
		}
		else if(color.a > 0) //Blend, slower
		{
			foreach(dirtyRect; dirtyRects)
		    {
		        for (int y = dirtyRect.min.y; y < dirtyRect.max.y; ++y)
		        {
		            RGBA[] rawScan = rawMap.scanline(y);
		            for (int x = dirtyRect.min.x; x < dirtyRect.max.x; ++x)
		            {
		                rawScan.ptr[x] = blendColor(color, rawScan.ptr[x], color.a);
		            }
		        }
		    }
		}
	}

	static void drawCircleRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, vec2f origin, float radius, RGBA color)
	{
		if (color.a <= 0)
			return;
		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			Canvas canvas;
			canvas.initialize(croppedRaw);
			canvas.translate(-dirtyRect.min.x, -dirtyRect.min.y); // Offset so we're using a consistent coordinate system
			canvas.fillStyle = color;
			canvas.fillCircle(origin, radius);
		}
	}
	
	static void drawRoundedRectangleRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, RGBA color, float radius, int resolution = 10)
	{
		if (color.a <= 0)
			return;
		resolution += 2;
		float W = position.width;
		float H = position.height;
		float minSide = min(W, H);
		float r = min(radius, minSide*0.5);
		if (resolution <= 2)
			r = 0;
		const float stepAngle = PI * 0.5 / cast(float)(resolution-1);
		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			Canvas canvas;
			canvas.initialize(croppedRaw);
			canvas.translate(position.min.x-dirtyRect.min.x, position.min.y-dirtyRect.min.y); // Offset so we're using a consistent coordinate system
			canvas.fillStyle = color;
			canvas.beginPath();
			canvas.moveTo(r,0);
			canvas.lineTo(W-r,0);
			if (r > 0)
			{
				for(int i=1; i<resolution-1; i++)
				{
					float a = cast(float)i*stepAngle - PI*0.5f;
					canvas.lineTo(W-r+cos(a)*r,r+sin(a)*r);
				}
			}
			canvas.lineTo(W,H-r);
			if (r > 0)
			{
				for(int i=1; i<resolution-1; i++)
				{
					float a = cast(float)i*stepAngle;
					canvas.lineTo(W-r+cos(a)*r,H-r+sin(a)*r);
				}
			}
			canvas.lineTo(r,H);
			if (r > 0)
			{
				for(int i=1; i<resolution-1; i++)
				{
					float a = cast(float)i*stepAngle + PI * 0.5;
					canvas.lineTo(r+cos(a)*r,H-r+sin(a)*r);
				}
			}
			canvas.lineTo(0,r);
			if (r > 0)
			{
				for(int i=1; i<resolution-1; i++)
				{
					float a = cast(float)i*stepAngle + PI;
					canvas.lineTo(r+cos(a)*r,r+sin(a)*r);
				}
			}
			canvas.closePath();
			canvas.fill();
		}
	}
	
	static void drawPillHorizontalRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, RGBA color, bool leftCap = true, bool rightCap = true, int resolution = 10)
	{
		drawPillRaw(rawMap, dirtyRects, position, color, false, leftCap, rightCap, resolution);
	}
	static void drawPillVerticalRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, RGBA color, bool topCap = true, bool bottomCap = true, int resolution = 10)
	{
		drawPillRaw(rawMap, dirtyRects, position, color, true, topCap, bottomCap, resolution);
	}
	static void drawPillRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, RGBA color, bool vertical, bool startCap, bool endCap, int resolution)
	{
		if (color.a <= 0)
			return;
		resolution += 2;
		float W = position.width;
		float H = position.height;
		float r = (vertical?W:H) * 0.5f;
		if (resolution <= 2)
			r = 0;
		const float stepAngle = PI / cast(float)(resolution-1);
		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			Canvas canvas;
			canvas.initialize(croppedRaw);
			canvas.translate(position.min.x-dirtyRect.min.x, position.min.y-dirtyRect.min.y); // Offset so we're using a consistent coordinate system
			canvas.fillStyle = color;
			canvas.beginPath();
			if(vertical)
			{
				canvas.moveTo(0,r);
				if (resolution > 2 && startCap)
				{
					for(int i=1; i<resolution-1; i++)
					{
						float a = cast(float)i*stepAngle - PI;
						canvas.lineTo(r+cos(a)*r,r+sin(a)*r);
					}
				}
				canvas.lineTo(W,r);
				canvas.lineTo(W,H-r);
				if (resolution > 2 && endCap)
				{
					for(int i=1; i<resolution-1; i++)
					{
						float a = cast(float)i*stepAngle;
						canvas.lineTo(r+cos(a)*r,H-r+sin(a)*r);
					}
				}
				canvas.lineTo(0,H-r);
				//canvas.lineTo(0,0);
			}
			else
			{
				canvas.moveTo(r,0);
				canvas.lineTo(W-r,0);
				if (resolution > 2 && endCap)
				{
					for(int i=1; i<resolution-1; i++)
					{
						float a = cast(float)i*stepAngle - PI*0.5f;
						canvas.lineTo(W-r+cos(a)*r,r+sin(a)*r);
					}
				}
				canvas.lineTo(W-r,H);
				canvas.lineTo(r,H);
				if (resolution > 2 && startCap)
				{
					for(int i=1; i<resolution-1; i++)
					{
						float a = cast(float)i*stepAngle + PI*0.5f;
						canvas.lineTo(r+cos(a)*r,r+sin(a)*r);
					}
				}
				//canvas.lineTo(0,0);
			}
			canvas.closePath();
			canvas.fill();
		}
	}

	static void drawLabelRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2i position, const(char)[] text, Font font, RGBA color, bool stableCenterY = true)
	{
		drawLabelRaw(rawMap, dirtyRects, cast(box2f)position, text, font, color, stableCenterY);
	}
	static void drawLabelRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, const(char)[] text, Font font, RGBA color, bool stableCenterY = true)
	{
		if (color.a <= 0)
			return;
		float W = position.width;
		float H = position.height;
		vec2f textPosition = position.min+vec2f(W * 0.5f, H * 0.5f);
		float letterSpacingPx = 0;
		
		//float fontHeight = cast(float)font.measureText(fontTestFullString, H, 0).height;
		float fontAscent = font.getAscent(H);
		float fontDescent = font.getDescent(H);
		float fontHeight = fontAscent;//+fontDescent;
		
		float fontBaselineRatio = fontDescent/fontHeight;
		float fontSizeCorrection = H / fontHeight;
		float fontSizePx = H * fontSizeCorrection;
		
		if (stableCenterY)
			textPosition.y += H/2 / fontSizeCorrection;

		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			
			/* //Debug Boundaries
			Canvas canvas;
			canvas.initialize(croppedRaw);
			canvas.translate(position.min.x-dirtyRect.min.x, position.min.y-dirtyRect.min.y); // Offset so we're using a consistent coordinate system
			canvas.fillStyle = RGBA(255,255,255,16);
			canvas.beginPath();
			canvas.moveTo(0,0);
			canvas.lineTo(W,0);
			canvas.lineTo(W,H);
			canvas.lineTo(0,H);
			//canvas.lineTo(0,0);
			canvas.closePath();
			canvas.fill();
			//*/
			
		    vec2f positionInDirty = textPosition - dirtyRect.min;
		    auto alignmentY = stableCenterY ? VerticalAlignment.baseline : VerticalAlignment.center;
			croppedRaw.fillText(font, text, fontSizePx, 0, color, positionInDirty.x, positionInDirty.y, HorizontalAlignment.center, alignmentY);
		}
	}

	static int countLines(const(char)[] text)
	{
		int lines = 1;
		for(int i=0; i < cast(int)text.length; i++)
		{
			if (text[i] == '\n')
				lines++;
		}
		return lines;
	}

	static void drawTextRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2i position, const(char)[] text, Font font, float fontSizePx, RGBA color, bool stableCenterY = true)
	{
		drawTextRaw(rawMap, dirtyRects, cast(box2f)position, text, font,  fontSizePx, color, stableCenterY);
	}
	static void drawTextRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, const(char)[] text, Font font, float fontSizePx, RGBA color, bool stableCenterY = true)
	{
		//ulong init_size = strlen(text.ptr);
		char[] delim = cast(char[])"\n";
		char[] textcopy = StringHelper.duplicateString(text);
		char *ptr = strtok(textcopy.ptr, delim.ptr);
		box2f rect = position;
		rect.max.y = rect.min.y + fontSizePx;
		
		// Vertically center...
		int lineCount = countLines(text);
		float totalSize = fontSizePx*lineCount;
		float extraSize = position.height - totalSize;
		rect.min.y += extraSize/2;
		rect.max.y += extraSize/2;
		
		while(ptr != null)
		{
			//printf("'%s'\n", ptr);
			drawLabelRaw(rawMap, dirtyRects, rect, ptr[0..strlen(ptr)], font, color, stableCenterY);
			rect.min.y += fontSizePx;
			rect.max.y += fontSizePx;
			ptr = strtok(null, delim.ptr);
		}
		free(textcopy.ptr);
		
		/*
		int lineCount = text.length;
		box2i linePosition = position;
		linePosition.height = fontSizePx;
		for(int i=0; i<text.length; i++)
		{
			drawLabelRaw(rawMap, dirtyRects, linePosition, text[i], font, fontSizePx, color, stableCenterY);
			position.min.y += position.height;
		}
		*/
	}
	
	static void drawIconRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2i position, Icon icon, RGBA color, float scale = 1.0f)
	{
		drawIconRaw(rawMap, dirtyRects, cast(box2f)position, icon, color, scale);
	}
	static void drawIconRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, Icon icon, RGBA color, float scale = 1.0f)
	{
		if (color.a <= 0)
			return;
		
		ensureResources();
			
		float W = position.width;
		float H = position.height;
		vec2f iconPosition = position.min+vec2f(W * 0.5f, H * 0.5f);
		float fontSizePx = H * scale;
		const(char)[] iconText = IconChar[icon];

		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			
		    vec2f positionInDirty = iconPosition - dirtyRect.min;
		    Font f = null;
		    if(icon<Icon.Plugin_MAX)
		    	f = fontPluginIcons;
			else if (icon<Icon.OT_MAX)
		    	f = fontOneTrickIcons;
		    else
				f = fontAwesomeIcons;
			croppedRaw.fillText(f, iconText, fontSizePx, 0, color, positionInDirty.x, positionInDirty.y, HorizontalAlignment.center, VerticalAlignment.center);
		}
	}
	
	// TODO: Save/Load the layout in some format to replace Wren in production
	static void saveLayout(UIElement rootElement, const(char)[] path)
	{
	}
	static void loadLayout(UIElement rootElement, const(char)[] path)
	{
	}
	
	private:
		static Font fontPluginIcons = null;
		static Font fontOneTrickIcons = null;
		static Font fontAwesomeIcons = null;
		
		//static const string fontTestQuickString = "AHXYQjpgqp01(){}[]";
		//static const string fontTestFullString = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890(){}[]";

}
