/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.icon;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import core.stdc.stdlib: malloc, realloc, free;
import core.stdc.string;

import onetrick.ui.helper;

/// A simple icon
class UIIcon : UIElement
{
public:
nothrow:
@nogc:

	@ScriptProperty RGBA color = RGBA(128, 128, 128, 255);
	@ScriptProperty RGBA colorShadow = RGBA(0, 0, 0, 0);
	@ScriptProperty RGBA colorBackground = RGBA(0, 0, 0, 0);
	@ScriptProperty float shadowOffset = 0.1f;

	this(UIContext context, const Icon icon, RGBA color = RGBA(255,255,255,255))
	{
		super(context, flagRaw);
		setIcon(icon);
		this.color = color;
	}

	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		if (colorBackground.a > 0)
		{
			UIHelper.fillColorRaw(rawMap, dirtyRects, colorBackground);
		}
		if (icon != Icon.None)
		{
			if (colorShadow.a > 0)
			{
				float y = position.size.y*shadowOffset;
				if (y != 0)
					UIHelper.drawIconRaw(rawMap, dirtyRects, box2f.rectangle(0, y, position.size.x, position.size.y), icon, colorShadow);
			}
			UIHelper.drawIconRaw(rawMap, dirtyRects, box2f.rectangle(0, 0, position.size.x, position.size.y), icon, color);
		}
	}

	void setColor(RGBA color)
	{
		if (this.color != color)
		{
			this.color = color;
			setDirtyWhole();
		}
	}
	void setBackgroundColor(RGBA color)
	{
		if(this.colorBackground != color)
		{
			this.colorBackground = color;
			setDirtyWhole();
		}
	}

	void setIcon(Icon newIcon)
	{
		if (newIcon != icon)
		{
			icon = newIcon;
			setDirtyWhole();
		}
	}

private:
	Icon icon = Icon.None;
	
}
