/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.label;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;
import dplug.graphics.font;

import core.stdc.string;

import onetrick.ui.helper;
import onetrick.string.helper;

/// A simple label
class UILabel : UIElement
{
public:
nothrow:
@nogc:
	this(UIContext context, Font font, const(char)[] text, RGBA color = RGBA(255,255,255,255))
	{
		super(context, flagRaw);
		_font = font;
		setText(text);
		this.color = color;
	}
	~this()
	{
		if (_text !is null)
			StringHelper.freeString(_text.ptr);
	}

	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		if (_backgroundColor.a > 0)
		{
			UIHelper.drawPillHorizontalRaw(rawMap, dirtyRects, box2f(0,0,position.size.x,position.size.y), _backgroundColor);
			//UIHelper.drawCircleRaw(rawMap, dirtyRects, vec2f(position.size.x/2, position.size.y/2), position.size.y/2, RGBA(255,255,255,255));
		}
		if (_text !is null)
		{
			if (_shadowColor.a > 0)
			{
				float y = position.size.y*_shadowOffset;
				if (y != 0)
					UIHelper.drawLabelRaw(rawMap, dirtyRects, box2f.rectangle(0, y, position.size.x, position.size.y), StringHelper.getStringSlice(_text), _font, _shadowColor);
			}
			UIHelper.drawLabelRaw(rawMap, dirtyRects, box2f.rectangle(0, 0, position.size.x, position.size.y), StringHelper.getStringSlice(_text), _font, _color);
		}
	}

	@property RGBA color() { return _color; }
	@property void color(RGBA value)
	{
		if (value != _color)
		{
			_color = value;
			setDirtyWhole();
		}
	}

	@property RGBA shadowColor() { return _shadowColor; }
	@property void shadowColor(RGBA value)
	{
		if (value != _shadowColor)
		{
			_shadowColor = value;
			setDirtyWhole();
		}
	}

	@property RGBA backgroundColor() { return _backgroundColor; }
	@property void backgroundColor(RGBA value)
	{
		if (value != _backgroundColor)
		{
			_backgroundColor = value;
			setDirtyWhole();
		}
	}

	@property float shadowOffset() { return _shadowOffset; }
	@property void shadowOffset(float value)
	{
		if (value != _shadowOffset)
		{
			_shadowOffset = value;
			setDirtyWhole();
		}
	}

	void setText(char* newText)
	{
		if (newText !is null)
			setText(newText[0..strlen(newText)]);
		else
			setText("");
	}
	void setText(const(char)[] newText)
	{
		/* // Allocate as normal
		if (_text !is null)
			StringHelper.freeString(_text.ptr);
		_text = StringHelper.duplicateString(newText);	
		//*/

		/* // Stress Test
		foreach(i; 0..1000)
		{
			if (_text !is null)
				StringHelper.freeString(_text.ptr);
			_text = StringHelper.duplicateString(newText);	
		}
		//*/
		
		//* // Avoid unnecessary allocations:
		if(_text is null)
		{
			_text = StringHelper.duplicateString(newText);
		}
		else
		{
			if(newText.length > strlen(_text.ptr))
			{
				StringHelper.freeString(_text.ptr);
				_text = StringHelper.duplicateString(newText);
			}
			else
			{
				strcpy(_text.ptr, newText.ptr);
			}
		}		
		//*/
		
		setDirtyWhole();
	}

private:
	Font _font;
    char[] _text = null;

	RGBA _color = RGBA(128, 128, 128, 255);
	RGBA _shadowColor = RGBA(0, 0, 0, 0);
	RGBA _backgroundColor = RGBA(0, 0, 0, 0);
	float _shadowOffset = 0.1f;
	
}
