/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.pager;

// Actually remove/add panels as they are toggled?
// This is no longer needed as invisible elements don't recieve mouse events
const bool DISCONNECT_HIDDEN = false;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import onetrick.ui.customgui;

/// A logical grouping of child elements that can have their visibility toggled
final class UIPager : UIElement, ICustomGUIListener
{
public:
nothrow:
@nogc:

	this(UIContext context)
	{
		super(context, flagRaw);
		childPanels = makeVec!UIElement();
	}

	~this()
	{
		
	}

	void lazyInit()
	{
		if (childPanels.length <= 0)
		{
			foreach(panel; _children[])		
			{
				childPanels ~= panel;
			}
			auto root = cast(CustomGUI)topLevelParent();
			if (root !is null)
				root.addListener(this);
		}

	}
	void onPreScript()
	{
		updatePanels(DISCONNECT_HIDDEN);
	}
	void onPostScript()
	{
		updatePanels();
	}


	int getPage() { return _activePanel; }

	void updatePanels(bool forceShowAll=false)
	{
		int i = 0;
		foreach(panel; childPanels)//_children[])
		{
			bool newVisible = (i==_activePanel || forceShowAll);
			bool currentVisible = DISCONNECT_HIDDEN ? (_children.indexOf(panel) >= 0) : panel.isVisible();
			if (newVisible != currentVisible)
			{
				if (newVisible) {
					if (DISCONNECT_HIDDEN)
						addChild(panel);
					else
						panel.visibility(true);
				} else {
					if (DISCONNECT_HIDDEN)
						removeChild(panel);
					else
						panel.visibility(false);
				}
			}
			i++;
		}
	}

	void setPage(int index)
	{
		if (index==_activePanel)
			return;
			
		lazyInit();
		
		_activePanel = index;
		
		updatePanels(index==-1);
	}


private:
	Vec!UIElement childPanels;
	int _activePanel = -1;
}
