/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.presetselector;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;
import dplug.graphics.font;
import dplug.client.params;

import onetrick.client;
import onetrick.preset;
import onetrick.ui.helper;
import onetrick.ui.customgui;
import onetrick.ui.button;
import onetrick.string.helper;

/// A preset selector
final class UIPresetSelector : UIElement, IParameterListener, IButtonListener
{
public:
nothrow:
@nogc:

	@ScriptProperty RGBA color = RGBA(128, 128, 128, 255);
	@ScriptProperty RGBA colorLabel = RGBA(255, 255, 255, 255);
	@ScriptProperty RGBA colorLabelDisabled = RGBA(200, 200, 200, 255);
	@ScriptProperty float labelSize = 0.8;

	version(StressTest)
	{
		int timeout = 0;
		override void onAnimate(double dt, double time)
		{
			timeout++;
			if (timeout > 10)
				selectFactoryNext();
		}
	}
	this(OneTrickClient client, UIContext context, Font font)
	{
		version(StressTest)
		{
			super(context, flagRaw | flagAnimated);
		}
		else
		{
			super(context, flagRaw);
		}
		
		this.client = client;
		this.font = font;
		
		addChild(buttonPrevious = mallocNew!UIButton(context, Icon.Left));
		addChild(buttonNext = mallocNew!UIButton(context, Icon.Right));
		
		debug
		{
			//addChild(buttonNew = mallocNew!UIButton(context, Icon.New));
			addChild(buttonSave = mallocNew!UIButton(context, Icon.Save));
			addChild(buttonOpen = mallocNew!UIButton(context, Icon.Open));
		}
		
		buttonPrevious.addListener(this);
		buttonNext.addListener(this);
		debug
		{
			//buttonNew.addListener(this);
			buttonSave.addListener(this);
			buttonOpen.addListener(this);
		}
        foreach(Parameter param; client.params())
        {
        	param.addListener(this);
        }
	}
	~this()
	{
		buttonPrevious.removeListener(this);
		buttonNext.removeListener(this);
		debug
		{
			//buttonNew.removeListener(this);
			buttonSave.removeListener(this);
			buttonOpen.removeListener(this);
		}
        foreach(Parameter param; client.params())
        {
			param.removeListener(this);
		}
	}

	override void reflow()
	{
		auto p = position;
		buttonPrevious.position = rectangle(p.min.x, p.min.y, p.size.y, p.size.y);
		buttonNext.position = rectangle(p.min.x+p.size.x-p.size.y, p.min.y, p.size.y, p.size.y);
		debug
		{
			buttonSave.position = rectangle(p.min.x+p.size.x-p.size.y*2, p.min.y, p.size.y, p.size.y);
			buttonOpen.position = rectangle(p.min.x+p.size.x-p.size.y*3, p.min.y, p.size.y, p.size.y);
		}
	}
	
	void onPressed(UIButton button)
	{
		if (button is buttonPrevious)
			selectFactoryPrevious();
		else if (button is buttonNext)
			selectFactoryNext();
		else {
			debug
			{
				char[4096] presetPath;
				
				// WARNING: Do not remove presetNameCStr, it's needed to keep string in scope!!!
				auto presetNameCStr = selectedPreset !is null ? CString(selectedPreset.name()) : CString("");
				
				auto presetName = (selectedPreset !is null ? presetNameCStr.storage : "Preset".ptr);
				
				version (Posix)
				{
					import std.path;
					auto path = expandTilde("~/Downloads/%s.preset"); // This is just for development at the moment
				}
				else
				{
					auto path = "%s.preset";
				}
				StringHelper.formatString(presetPath, path, presetName);
				
				//else if (button is buttonNew)
				//	{}
				if (button is buttonSave)
				{
					Preset newPreset = PresetHelper.saveBinaryPresetToPath(client, StringHelper.getStringSlice(presetPath));
					//selectedPreset = newPreset;
					//selectedIndex = -1;
					presetDirty = false;
					setDirtyWhole();
				}
				else if (button is buttonOpen)
				{
					Preset newPreset = PresetHelper.loadBinaryPresetFromPath(client, StringHelper.getStringSlice(presetPath));
					//selectedPreset = newPreset;
					if (newPreset !is null)
					{
						PresetHelper.applyPresetFromGUI(client, newPreset);
						//selectedIndex = -1;
						presetDirty = false;
						setDirtyWhole();
					}
				}
			}
		}
	}
	override void onBeginParameterHover(Parameter sender) { }
	override void onEndParameterHover(Parameter sender) { } 
	override void onBeginParameterEdit(Parameter sender)
	{
		setPresetDirty();
	}
	override void onParameterChanged(Parameter sender)
	{
		//setPresetDirty(); //Triggered when we change presets, but late for some reason...
	}
	override void onEndParameterEdit(Parameter sender)
	{
		setPresetDirty();
	}
	
	void setPresetDirty()
	{
		if (!presetDirty)
		{
			presetDirty = true;
			setDirtyWhole();
		}
	}

	
    //inout(Preset) preset(int i) inout nothrow @nogc
    //int numPresets() nothrow @nogc
    //int currentPresetIndex() nothrow @nogc
    //Preset currentPreset() nothrow @nogc
    //bool isValidPresetIndex(int index) nothrow @nogc
	
	void applyFactoryPreset(int index)
	{
		Preset preset = PresetHelper.getFactoryPreset(client, index);
		if (preset !is null)
		{
			selectedPreset = preset;
			PresetHelper.applyPresetFromGUI(client, preset);
			presetDirty = false;
			setDirtyWhole();
		}
	}
	
	
	void selectFactoryNext()
	{
		int presetCount = PresetHelper.countFactoryPresets(client);
		if (selectedIndex < 0)
			selectedIndex = 0;
		else
			selectedIndex = (selectedIndex + 1) % presetCount;
		applyFactoryPreset(selectedIndex);
		setDirtyWhole();
	}
	
	void selectFactoryPrevious()
	{
		int presetCount = PresetHelper.countFactoryPresets(client);
		if (selectedIndex < 0)
			selectedIndex = 0;
		else
			selectedIndex = (selectedIndex + presetCount - 1) % presetCount;
		applyFactoryPreset(selectedIndex);
		setDirtyWhole();
	}
	
	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		UIHelper.fillColorRaw(rawMap, dirtyRects, color);
		
		bool hasPreset = (selectedPreset !is null);
		RGBA c = hasPreset ? colorLabel : colorLabelDisabled;
		char[4096] modifiedName;
		auto presetName = (hasPreset ? selectedPreset.name() : "Kits");
		StringHelper.formatString(modifiedName, (hasPreset && presetDirty) ? "* %s *" : "%s", CString(presetName).storage);
		UIHelper.drawLabelRaw(rawMap, dirtyRects, box2f.rectangle(0, cast(int)(position.size.y*(1.0-labelSize)*0.5), position.size.x, cast(int)(position.size.y*labelSize)), StringHelper.getStringSlice(modifiedName), font, c);
		
		int arrowSize = position.size.y;
	}
	
	/* //TODO: Show a popup menu
	override Click onMouseClick(int x, int y, int button, bool isDoubleClick, MouseState mstate)
	{
		return Click.startDrag;
	}
	//*/

private:
	OneTrickClient client;
	Preset selectedPreset = null;
	Font font;
	
	UIButton buttonPrevious;
	UIButton buttonNext;
	debug
	{
		//UIButton buttonNew;
		UIButton buttonSave;
		UIButton buttonOpen;
	}
	
	int selectedIndex = -1;
	Canvas canvas;
	bool presetDirty = false;

}
