/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.tab;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import std.algorithm;
import core.stdc.string;


import onetrick.ui.activatable;
import onetrick.ui.tabcontainer;
import onetrick.ui.helper;

/// A tab button that belongs to a UITabContainer
final class UITab : UIElement, IActivatable
{ mixin Activatable;
public:
nothrow:
@nogc:

	// TODO: Move these to regular @properties that can setDirty
	@ScriptProperty float labelOffsetY = 0;
	@ScriptProperty float labelRelativeHeight = 0.4f;
	@ScriptProperty float styleOutlineThickness = 0.05;
	float horizontalSkew = 0;
	@ScriptProperty RGBA colorFillInactive = RGBA(64, 64, 64, 255);
	@ScriptProperty RGBA colorFillActive = RGBA(96, 96, 96, 255);
	@ScriptProperty RGBA colorFillHover = RGBA(96, 192, 255, 255);
	@ScriptProperty RGBA colorOutlineInactive = RGBA(96, 96, 96, 255);
	@ScriptProperty RGBA colorOutlineActive = RGBA(196, 196, 0, 255);
	@ScriptProperty RGBA colorOutlineHover = RGBA(255, 255, 0, 255);
	@ScriptProperty RGBA colorLabelInactive = RGBA(96, 96, 96, 255);
	@ScriptProperty RGBA colorLabelActive = RGBA(196, 196, 0, 255);
	@ScriptProperty RGBA colorLabelHover = RGBA(255, 255, 0, 255);

	this(UIContext context, const(char)[] labelText, Font labelFont)
	{
		super(context, flagRaw);
		
		this.labelText = labelText;
		this.labelFont = labelFont;
	}
	this(UIContext context, Icon icon)
	{
		super(context, flagRaw);
		
		this.icon = icon;
	}

	box2f getLabelPosition()
	{
		auto p = position;
		float x=0;
		float y=labelOffsetY;
		float w=p.size.x;
		float h=p.size.y;
		return box2f.rectangle(cast(int)(x), cast(int)(y+h/2-(h*labelRelativeHeight/2)), cast(int)(w), cast(int)(h*labelRelativeHeight));
	}

	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		float W = position.width;
		float H = position.height;
		float outlineH = H * styleOutlineThickness;
		bool active = isActive();
		
		auto fillColor = isMouseOver ? colorFillHover : (active ? colorFillActive : colorFillInactive);
		auto highlightColor = isMouseOver ? colorOutlineHover : (active ? colorOutlineActive : colorOutlineInactive);
		
		//if (fillColor.a > 0) // More efficient, but has edge artifacts, we'll use Canvas below instead with a margin
		//	UIHelper.fillColorRaw(rawMap, dirtyRects, fillColor);
		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			canvas.initialize(croppedRaw);
			canvas.translate(-dirtyRect.min.x, -dirtyRect.min.y); // Offset so we're using a consistent coordinate system

			// Background
			const int margin = 1; //Avoid issues with AA with neighboring tabs
			if (fillColor.a > 0 && W>margin*2 && H>margin) {
				canvas.fillStyle = fillColor;
				canvas.beginPath();
				canvas.moveTo(margin+max(0,horizontalSkew),margin);
				canvas.lineTo(W-margin+min(0, horizontalSkew),margin);
				canvas.lineTo(W-margin-max(0, horizontalSkew),H);
				canvas.lineTo(margin-min(0, horizontalSkew),H);
				//canvas.lineTo(0,0);
				canvas.closePath();
				canvas.fill();
			}

			// Highlight
			canvas.fillStyle = highlightColor;
			canvas.beginPath();
			canvas.moveTo(-min(0, horizontalSkew),H-outlineH);
			canvas.lineTo(W-max(0,horizontalSkew),H-outlineH);
			canvas.lineTo(W-max(0,horizontalSkew),H);
			canvas.lineTo(-min(0, horizontalSkew),H);
			//canvas.lineTo(0,0);
			canvas.closePath();
			canvas.fill();
		}
		if (labelText !is null && labelFont !is null && labelRelativeHeight > 0)
		{
			RGBA labelColor = (isMouseOver ? colorLabelHover : (active ? colorLabelActive : colorLabelInactive));
			if (icon != Icon.None)
				UIHelper.drawIconRaw(rawMap, dirtyRects, getLabelPosition(), icon, labelColor);
			else if(labelText !is null)
				UIHelper.drawLabelRaw(rawMap, dirtyRects, getLabelPosition(), labelText, labelFont, labelColor);
		}
	}

	override Click onMouseClick(int x, int y, int button, bool isDoubleClick, MouseState mstate)
	{
		if (!isEnabled)
			return Click.handled; // Blocked
		if (!isVisible() || !parent().isVisible())
			return Click.unhandled; // Pass through
		(cast(UITabContainer)parent()).setTab(this);
		return Click.handled;
	}

	override void onMouseEnter()
	{
		if (!isEnabled)
			return;
		isMouseOver = true;
		setDirtyWhole();
	}

	override void onMouseExit()
	{
		if (!isEnabled)
			return;
		isMouseOver = false;
		setDirtyWhole();
	}
	final bool enabled() pure const
    {
        return isEnabled;
    }
    final void enabled(bool isEnabled)
    {
    	if (this.isEnabled != isEnabled)
		{
			this.isEnabled = isEnabled;
			setDirtyWhole();
		}
    }

private:
	bool isMouseOver = false; // Shadow the built-in to make Indicator hack more efficient
	bool isEnabled = true;
	Canvas canvas;
	
	Icon icon = Icon.None;
	//UILabel _label;
	const(char)[] labelText=null;
	Font labelFont=null;

}
