/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.tabcontainer;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import std.algorithm;

import onetrick.ui.tab;
import onetrick.ui.pager;
import onetrick.ui.helper;

/// A tab Container that contains multiple UITab instances and controls a UITabPanel
final class UITabContainer : UIElement
{
public:
nothrow:
@nogc:
	// TODO: Move these to regular @properties that can setDirty
	@ScriptProperty RGBA colorFill = RGBA(64, 64, 64, 0);
	@ScriptProperty int horizontalSkew = 0;
	
	this(UIContext context, UIPager pager)
	{
		super(context, flagRaw);
		_pager = pager;
	}

	void setTab(UITab tab)
	{
		int tabIndex = _children.indexOf(tab);
		for(int i=0; i<_children.length; i++)
		{
			(cast(UITab)_children[i]).setActive(i==tabIndex);
		}
		_pager.setPage(tabIndex);
	}
	void setTab(int index)
	{
		setTab(cast(UITab)_children[index]);
	}

	override void reflow()
	{
		box2i totalRect = _position;
		int horizontalSkewAbs = horizontalSkew<0?-horizontalSkew:horizontalSkew; //abs() was being wonk?
		int count = cast(int)_children.length;
		int availableWidth = totalRect.width - horizontalSkewAbs;
		int tabWidth = availableWidth / count;
		int totalWidth = tabWidth * cast(int)_children.length;
		int pixelsShort = totalRect.width - totalWidth;
		int offset = pixelsShort/2;
		int startX=totalRect.min.x + offset;
		startX += min(0, horizontalSkew)/2;
		int expand = horizontalSkewAbs;
		
		int i=0;
		foreach(tab; _children[])
		{
			int tx = startX+(tabWidth*i);
			(cast(UITab)tab).horizontalSkew = horizontalSkew;
			tab.position = box2i.rectangle(tx, totalRect.min.y, tabWidth+expand, totalRect.height);
			i++;
		}
	}

	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		if (colorFill.a > 0)
			UIHelper.fillColorRaw(rawMap, dirtyRects, colorFill);
	}


private:
	Canvas canvas;
	UIPager _pager;
	
}
