/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.themebase;

import dplug.gui.element;
import dplug.flatwidgets;
import onetrick.ui;
import onetrick.string;

/*
	This class is a quick-and-dirty port of the Wren script functionality used during development.
	This whole system should be replaced at some point.
*/

class UIThemeBase
{
	public:
	nothrow:
	@nogc:

	struct Coord
	{
		float x;
		float y;
	}
	struct Size
	{
		float width;
		float height;
	}
	struct Rectangle
	{
	public:
	nothrow:
	@nogc:
		float x;
		float y;
		float width;
		float height;
		@property Coord origin() { return Coord(x, y); }
		@property Size size() { return Size(width, height); }
		this(float x, float y, float width, float height)
		{
			assert(x>=0);
			assert(y>=0);
			this.x = x;
			this.y = y;
			this.width = width;
			this.height = height;
		}
		Rectangle scaleByFactor(float scale)
		{
			return Rectangle(x*scale, y*scale, width*scale, height*scale);
		}
		
		@property box2i toBox()
		{
			return box2i(cast(int)(x+0.5f),cast(int)(y+0.5f),cast(int)(x+width+0.5f),cast(int)(y+height+0.5f));
		}
    	alias toBox this;
	}
	
	UIElement rootElement = null;
	this(UIElement rootElement)
	{
		this.rootElement = rootElement;
		createUI();
	}
	
	float W;
	float DW;
	float H;
	float DH;
	float S;

    void createUI() {
    }

    void reflow() {
	}

	void parameterChanged() {
	}

	bool initLayout() {
		if(rootElement.position.width < 30 || rootElement.position.height < 30)
			return false;
        W = rootElement.position.width;
        DW = rootElement.context.getDefaultUIWidth();
        H = rootElement.position.height;
        DH = rootElement.context.getDefaultUIHeight();
        S = W / DW;
        return true;
	}

	
	UIElement get(string id) {
		return rootElement.getElementById(cast(char*)id);
	}
	UIElement get(string id1, string id2) {
		char[1000] combined;
		StringHelper.formatString(combined, "%s%s", cast(char*)id1, cast(char*)id2);
		return get(cast(string)combined);
	}
	UIElement get(string id1, string id2, string id3) {
		char[1000] combined;
		StringHelper.formatString(combined, "%s%s%s", cast(char*)id1, cast(char*)id2, cast(char*)id3);
		return get(cast(string)combined);
	}

	static RGBA correct(RGBA c) { return lightenColor(c, 0.1); }

	static const string[] NUMBERS = ["0","1","2","3","4","5","6","7","8","9","10"]; // quick hack

	static const float GAIN_START_CLIP = 0.708;
	
	static RGBA RGBA_new(ubyte rgb) { return RGBA(rgb, rgb, rgb, 255); }
	static RGBA RGBA_new(ubyte rgb, ubyte a) { return RGBA(rgb, rgb, rgb, a); }
	static RGBA RGBA_new(ubyte r, ubyte g, ubyte b) { return RGBA(r, g, b, 255); }
	static RGBA RGBA_new(ubyte r, ubyte g, ubyte b, ubyte a) { return RGBA(r, g, b, a); }
		
	static const RGBA COLOR_CLEAR = correct(RGBA_new(255,0));
	static const RGBA COLOR_WHITE = correct(RGBA_new(255));
	static const RGBA COLOR_BLACK = correct(RGBA_new(0));
    static const RGBA COLOR_GREY1 = correct(RGBA_new(16));
    static const RGBA COLOR_GREY2 = correct(RGBA_new(32));
    static const RGBA COLOR_GREY3 = correct(RGBA_new(48));
    static const RGBA COLOR_GREY4 = correct(RGBA_new(64));
    static const RGBA COLOR_GREY5 = correct(RGBA_new(96));
    static const RGBA COLOR_GREY6 = correct(RGBA_new(128));
    static const RGBA COLOR_GREY7 = correct(RGBA_new(192));
    static const RGBA COLOR_GREY8 = correct(RGBA_new(224));
    static const RGBA COLOR_GREY9 = correct(RGBA_new(235));
    static const RGBA COLOR_BROWN = correct(RGBA_new(48,32,24,255));
    	
	static const RGBA COLOR_ORANGE = correct(RGBA_new(255, 155, 55, 255));
    static const RGBA COLOR_BLUE = correct(RGBA_new(55, 200, 255, 255));
	static const RGBA COLOR_PINK = correct(RGBA_new(247, 61, 200, 255));
	static const RGBA COLOR_AQUA = correct(RGBA_new(0, 255, 255, 255));
	static const RGBA COLOR_PURPLE = correct(RGBA_new(176, 146, 255, 255));
	static const RGBA COLOR_RED = correct(RGBA_new(252, 100, 100, 255));
	static const RGBA COLOR_YELLOW = correct(RGBA_new(252, 249, 100, 255));
	static const RGBA COLOR_GREEN = correct(RGBA_new(100, 255, 100, 255));
	
	static const RGBA COLOR_FILL_GREEN = RGBA_new(78, 209, 100, 255);
	static const RGBA COLOR_CLIP = correct(RGBA_new(255, 64, 32, 255));
		
	UIKnob getModuleKnob(string m, string name) { return cast(UIKnob)get(m,"Knob",name); }
	UISlider getModuleSlider(string m, string name) { return cast(UISlider)get(m,"Slider",name); }
	UIParameterEditor getModuleControl(string moduleName, string editorName)
	{
		if (auto knob = getModuleKnob(moduleName, editorName)) return knob;
		if (auto slider = getModuleSlider(moduleName, editorName)) return slider;
		return null;
	}	
	UIButton getModuleButton(string m, string name) { return cast(UIButton)get(m,"Button",name); }
	UIParameterReader getModuleParam(string m, string name) { return cast(UIParameterReader)get(m,"Param",name); }
	UIValue getModuleValue(string m, string name) { return cast(UIValue)get(m,"Value",name); }
	UIPeakIndicator getModuleIndicator(string m) { return cast(UIPeakIndicator)get(m,"Indicator"); }
	UITab getModuleTab(string m) { return cast(UITab)get(m,"Tab"); }
	UIColorPanel getModulePanel(string m) { return cast(UIColorPanel)get(m,"Panel"); }
	UISeparator getModuleSeparator(string m, string name) { return cast(UISeparator)get(m,"Separator",name); }
	UIConnector getModuleConnector(string m, string name) { return cast(UIConnector)get(m,"Connector",name); }
	UILabel getModuleLabel(string m, string name) { return cast(UILabel)get(m,"Label",name); }
	
	Rectangle flowPadding(Rectangle r, float padding) { return Rectangle(r.origin.x+padding, r.origin.y+padding, r.size.width-padding*2, r.size.height-padding*2); }
	Rectangle flowPaddingX(Rectangle r, float padding) { return Rectangle(r.origin.x+padding, r.origin.y, r.size.width-padding*2, r.size.height); }
	Rectangle flowPaddingY(Rectangle r, float padding) { return Rectangle(r.origin.x, r.origin.y+padding, r.size.width, r.size.height-padding*2); }
	Rectangle flowPaddingTop(Rectangle r, float padding) { return Rectangle(r.origin.x, r.origin.y+padding, r.size.width, r.size.height-padding); }
	Rectangle flowPaddingBottom(Rectangle r, float padding) { return Rectangle(r.origin.x, r.origin.y, r.size.width, r.size.height-padding); }
	Rectangle flowPaddingLeft(Rectangle r, float padding) { return Rectangle(r.origin.x+padding, r.origin.y, r.size.width-padding, r.size.height); }
	Rectangle flowPaddingRight(Rectangle r, float padding) { return Rectangle(r.origin.x, r.origin.y, r.size.width-padding, r.size.height); }
	Rectangle flowOffset(Rectangle r, float x, float y) { return Rectangle(r.origin.x+x, r.origin.y+y, r.size.width, r.size.height); }
	
	Rectangle flowSplitX(Rectangle r, float i, float count, float spacing) {
		auto chunkSize = (r.size.width - (spacing*(count-1)))/count;
		auto stride = chunkSize + spacing;
		return Rectangle(r.origin.x+(stride*i), r.origin.y, chunkSize, r.size.height);
	}
	Rectangle flowSplitX(Rectangle r, float i, float count) { return flowSplitX(r, i, count, 0); }
	
	Rectangle flowSplitY(Rectangle r, float i, float count, float spacing) {
		auto chunkSize = (r.size.height - (spacing*(count-1)))/count;
		auto stride = chunkSize + spacing;
		return Rectangle(r.origin.x, r.origin.y+(stride*i), r.size.width, chunkSize);
	}
	Rectangle flowSplitY(Rectangle r, float i, float count) { return flowSplitY(r, i, count, 0); }

	Rectangle __fsg_r;
	float __fsg_cx, __fsg_cy;
	void flowSetGrid(Rectangle r, float countX, float countY) {
		__fsg_r = r;
		__fsg_cx = countX;
		__fsg_cy = countY;
	}
	Rectangle flowUseGrid(float iX, float iY) { return flowSplitY(flowSplitX(__fsg_r, iX, __fsg_cx), iY, __fsg_cy); }
	
	Rectangle flowGrid(Rectangle r, float iX, float iY, float countX, float countY) { return flowSplitY(flowSplitX(r, iX, countX), iY, countY); }
	Rectangle flowCenter(Rectangle r, float w, float h) { return Rectangle(r.origin.x+(r.size.width-w)/2, r.origin.y+(r.size.height-h)/2, w, h); }
	Rectangle flowCenter(Rectangle r, float s) { return Rectangle(r.origin.x+(r.size.width-s)/2, r.origin.y+(r.size.height-s)/2, s, s); }
	Rectangle flowCenterX(Rectangle r, float w) { return Rectangle(r.origin.x+(r.size.width-w)/2, r.origin.y, w, r.size.height); }
	Rectangle flowCenterY(Rectangle r, float h) { return Rectangle(r.origin.x, r.origin.y+(r.size.height-h)/2, r.size.width, h); }
	Rectangle flowTop(Rectangle r, float h) { return Rectangle(r.origin.x, r.origin.y, r.size.width, h); }
	Rectangle flowBottom(Rectangle r, float h) { return Rectangle(r.origin.x, r.origin.y+r.size.height-h, r.size.width, h); }
	Rectangle flowLeft(Rectangle r, float w) { return Rectangle(r.origin.x, r.origin.y, w, r.size.height); }
	Rectangle flowRight(Rectangle r, float w) { return Rectangle(r.origin.x+r.size.width-w, r.origin.y, w, r.size.height); }
	Rectangle flowWidth(Rectangle r, float w) { return Rectangle(r.origin.x, r.origin.y, w, r.size.height); }
	Rectangle flowHeight(Rectangle r, float h) { return Rectangle(r.origin.x, r.origin.y, r.size.width, h); }
	
	static RGBA darkenColor(RGBA color, float t) { return RGBA_new(cast(ubyte)(color.r*t),cast(ubyte)(color.g*t),cast(ubyte)(color.b*t),color.a); }
	static RGBA lightenColor(RGBA color, float t) { return RGBA_new(cast(ubyte)(color.r+(255-color.r)*t),cast(ubyte)(color.g+(255-color.g)*t),cast(ubyte)(color.b+(255-color.b)*t),color.a); }
	static RGBA blendColor(RGBA a, RGBA b, float t) { return RGBA_new(cast(ubyte)(a.r+(b.r-a.r)*t), cast(ubyte)(a.g+(b.g-a.g)*t), cast(ubyte)(a.b+(b.b-a.b)*t), cast(ubyte)(a.a+(b.a-a.a)*t)); }
}
