/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
	This class is a quick-and-dirty port of the Wren script used during development.
	Doing layout here should be faster and less leaky.
*/

module theme;

import dplug.gui.element;
import dplug.flatwidgets;

import onetrick.ui;
import onetrick.string.helper;


class Theme : UIThemeBase {
public:
nothrow:
@nogc:
	this(UIElement rootElement)
	{
		super(rootElement);
	}

	const string[] GLOBAL_KNOB_NAMES = ["Gain", "Drive", "Sensitivity", "Saturation", "ReverbDecay", "ReverbGate"];
	const string[] MODULE_NAMES = ["module1", "module2", "module3", "module4", "module5", "module6", "module7", "module8", "module9", "module10"];
	const string[] MODULE_KNOB_NAMES = ["Gain", "Pan", "Noise", "Tone", "Bend", "Decay", "Punch", "ToneNoise", "Click", "GainDynamics", "BendDynamics", "LowpassQ", "Saturation", "FilterEnvBend", "SendReverb", "FilterEnvBendDynamics", "OscWaveform"];
	const string[] MODULE_BUTTON_NAMES = ["Icon", "Color"];
	RGBA[] MODULE_COLORS = [COLOR_RED, COLOR_ORANGE, COLOR_BLUE, COLOR_AQUA, COLOR_GREEN, COLOR_PURPLE, COLOR_PINK, COLOR_YELLOW];
	
	const float KNOB_SIZE = 80;
	
	void styleKnobBasic(UIKnob knob) {
		//knob.colorBacking = correct(RGBA_new(0,0,0,0));
		//knob.margin = 0.05;
		knob.indicatorStyle = 0;
		knob.indicatorOrigin = 0;
		if (knob.indicatorStyle == UIKnob.IndicatorStyle.Continuous) {
			knob.indicatorWidth = 0.1;
		} else {
			knob.indicatorWidth = 0.12;
			knob.indicatorLedCount = 25;
		}
		knob.indicatorGlow = 0.075;
		knob.indicatorGlowRadius=2;

		knob.knobRadius = 0.75;
		knob.hoverScale = 1.05;
		knob.knobColor = COLOR_WHITE;
		//knob.knobShadeColor = correct(RGBA_new(255,255,255,16));
		knob.knobShadeColor = correct(RGBA_new(10,22,55,32));
		knob.knobShadeColorHover = lightenColor(knob.knobShadeColor, 0.5);//correct(RGBA_new(100,200,255,92))
		knob.knobShadeR1=0.33;
		knob.knobShadeR2=0.93;
		knob.knobShadeR3=0.98;
		knob.indicatorColor = COLOR_GREY4;
		knob.colorHover = COLOR_WHITE;
		knob.colorDrag = COLOR_YELLOW;
		knob.labelColor = COLOR_WHITE;
		//knob.labelBackgroundColor = COLOR_CLEAR; //COLOR_WHITE.withAlpha(32);
		knob.labelValueColor = COLOR_GREY6;
		
		knob.labelBackgroundColor = RGBA_new(0, 0, 0, 0);
		knob.labelValueBackgroundColor = RGBA_new(0, 0, 0, 32);


		knob.knobMarkerStyle = UIKnob.MarkerStyle.Line;
		knob.knobMarkerOffset = 0.9;
		knob.knobMarkerWidth = 0.125;
		knob.knobMarkerLength = 0.4;
	}
	void styleKnobDark(UIKnob knob) {
		knob.backingColor = correct(RGBA_new(32,40,48,255));
		knob.indicatorStyle = UIKnob.IndicatorStyle.Led;
		knob.indicatorColor = COLOR_GREY5;
		knob.indicatorFillColor = COLOR_YELLOW;
		knob.labelColor = COLOR_WHITE; //COLOR_BLACK;
		knob.labelValueColor = COLOR_WHITE;
		knob.knobMarkerColor = COLOR_GREY2;
		//knob.indicatorWidth = 0.1;
		knob.margin = 0.1;
		knob.labelValueBackgroundColor = knob.backingColor;
		knob.labelValueColor = COLOR_GREY9;

	}
	
	void styleModule(string moduleName, RGBA color) {
	    color = MODULE_COLORS[getModuleParam(moduleName, "Color").intValue % 8];
	
		//getModulePanel(moduleName).color =  darkenColor(color, 0.1);
		
		getModulePanel(moduleName).color = COLOR_GREY2; // blendColor(darkenColor(color, 0.7), COLOR_GREY1, 0.8);
		
		auto tab = getModuleTab(moduleName);
		if (tab) {
			auto dColor = darkenColor(color,0.85);
			tab.colorFillInactive = COLOR_CLEAR;
			tab.colorFillActive = COLOR_WHITE;
			tab.colorFillHover = COLOR_WHITE;
			tab.colorOutlineInactive = COLOR_GREY6;
			tab.colorOutlineActive = dColor;
			tab.colorOutlineHover = dColor;
			tab.colorLabelActive = dColor;
			tab.colorLabelHover = dColor;
			tab.styleOutlineThickness = 0.05;
			tab.labelOffsetY = -7*S;
			//tab.labelOffsetY = 0;
			//tab.labelRelativeHeight = 0.7;
			tab.labelRelativeHeight = 0;
		}
		
		auto indicator=getModuleIndicator(moduleName);
		if (indicator) {
			indicator.oneShot = true;
			indicator.peakHoldMS = 16.67*3;
			indicator.peakFallMS = 225;
			indicator.color = COLOR_GREY3; //darkenColor(color, 0.5);
			indicator.colorActive = darkenColor(color, 0.7);
			indicator.colorPeaking = lightenColor(color, 0.5);
			//if ( ["module3", "module4", "module5", "module6", "module10"].indexOf(moduleName)>=0 ) {
			if (!getModuleValue(moduleName, "Enabled").boolValue) {
				indicator.color = COLOR_GREY7;
				indicator.colorActive = COLOR_GREY7;
				indicator.colorPeaking = COLOR_GREY7;
			}
			indicator.scale = 0.8;
			indicator.scalePeaking = 0.97;
		}
		
		foreach (knobName; MODULE_KNOB_NAMES) {
			auto knob = getModuleKnob(moduleName, knobName);
			styleKnobBasic(knob);
			knob.indicatorFillColor = color;
			knob.labelColor = color;
			knob.knobMarkerColor = darkenColor(color, 0.33);
			knob.labelBackgroundColor = RGBA_new(0, 0, 0, 0);
			//knob.labelValueBackgroundColor = RGBA_new(255, 255, 255, 32);
			knob.labelValueBackgroundColor = COLOR_GREY8;
			knob.labelValueColor = COLOR_GREY2;
		}
		
		foreach (knobName; ["Pan", "ToneNoise", "Bend", "FilterEnvBend"]) {
			auto knob = getModuleKnob(moduleName, knobName);
			knob.indicatorOrigin = 1;
			knob.indicatorDetent = true; //(knobName=="Bend");
		}
		
		foreach (knobName; ["GainDynamics", "BendDynamics", "FilterEnvBendDynamics"]) {
			auto knob = getModuleKnob(moduleName, knobName);
			knob.indicatorStyle = UIKnob.IndicatorStyle.Led;
			knob.indicatorFillColor = lightenColor(color, 0.75);
		}
		
		/*
		for (knobName; ["SendReverb"]) {
			auto knob = getModuleKnob(moduleName, knobName);
			knob.indicatorFillColor = COLOR_PURPLE;
		}
		*/
		
		foreach (i; 1..4+1) {
			getModuleSeparator(moduleName, NUMBERS[i]).color = COLOR_GREY9;
		}
		foreach (i; 1..5+1) {
			auto label = getModuleLabel(moduleName, NUMBERS[i]);
			label.color = COLOR_GREY9;
			//label.shadowColor = correct(RGBA_new(0,0,0,96));
			label.shadowOffset = 0.1;
		}
		foreach (i; 1..3+1) {
			auto connector = getModuleConnector(moduleName, NUMBERS[i]);
			connector.lineWidth = 3*S;
			connector.style = UIConnector.Style.VerticalLeft2;
			connector.color = COLOR_GREY6;
			connector.colorEditing = COLOR_YELLOW;
		}
		/*
		foreach (i; 4..5+1) {
			auto connector = getModuleConnector(moduleName, NUMBERS[i]);
			connector.style = UIConnector.Style.HorizontalDown2;
		}
		*/
		{
			auto knob = getModuleKnob(moduleName, "Gain");
			knob.indicatorClipStart = GAIN_START_CLIP;
			knob.indicatorClipColor = COLOR_CLIP;
		}
		
		foreach (buttonName; MODULE_BUTTON_NAMES) {
			auto button = getModuleButton(moduleName, buttonName);
			button.color = COLOR_GREY5;
			button.backgroundColorHover = COLOR_CLEAR;
			button.colorHover = COLOR_PINK;
			button.colorPressed = COLOR_YELLOW;
		}

			

	}
	
	void reflowModule(int index, float x, float y, float w, float h) {
		string moduleName = MODULE_NAMES[index];
		auto panel = getModulePanel(moduleName);
		if (panel) {
			panel.position = Rectangle(x,y,w,h);
		}
		
		auto indicator=getModuleIndicator(moduleName);
		if (indicator) {
			//auto tab = getModuleTab(moduleName)
			//if (tab) { indicator.position = flowCenter(tab.position, 50*S) }
			auto spacing = (w/MODULE_NAMES.length);
			auto indicatorX = (w/MODULE_NAMES.length)*index;
			auto indicatorSize = 50*S;
			auto offset = -indicatorSize/2;
			auto offsetX = spacing / 2;
			indicator.position = Rectangle(indicatorX + offsetX + offset, y - (32*S) + offset, indicatorSize, indicatorSize);
		}
		
		{
			auto fullR = flowPaddingX(Rectangle(x,y,w,h), 20*S);
			auto r = flowPaddingY(fullR, 40*S);
			
			auto gx = 11;
			auto gy = 2;
			auto gs = KNOB_SIZE*S;
			flowSetGrid(r, gx, gy);
			
			//get("module1ButtonIcon").position = Rectangle(0,0,10,10);
			getModuleButton(moduleName, "Icon").position = Rectangle(fullR.origin.x+10*S, fullR.origin.y, 30*S, 30*S);
			getModuleButton(moduleName, "Color").position = Rectangle(fullR.origin.x-20*S, fullR.origin.y, 30*S, 30*S);
			
			float ix = 0;
			getModuleKnob(moduleName, "Tone").position = flowCenter(flowUseGrid(ix, 0), gs);
			getModuleKnob(moduleName, "Bend").position = flowCenter(flowUseGrid(ix+1, 0), gs);
			getModuleKnob(moduleName, "OscWaveform").position = flowCenter(flowUseGrid(ix, 1), gs);
			getModuleKnob(moduleName, "BendDynamics").position = flowCenter(flowUseGrid(ix+1, 1), gs);
			
			ix = ix + 2.5;
			getModuleKnob(moduleName, "Noise").position = flowCenter(flowUseGrid(ix, 0), gs);
			getModuleKnob(moduleName, "FilterEnvBend").position = flowCenter(flowUseGrid(ix+1, 0), gs);
			getModuleKnob(moduleName, "LowpassQ").position = flowCenter(flowUseGrid(ix, 1), gs);
			getModuleKnob(moduleName, "FilterEnvBendDynamics").position = flowCenter(flowUseGrid(ix+1, 1), gs);
			
			ix = ix + 2.5;
			getModuleKnob(moduleName, "Decay").position = flowCenter(flowUseGrid(ix, 0), gs);
			getModuleKnob(moduleName, "Punch").position = flowCenter(flowUseGrid(ix, 1), gs);

			ix = ix + 1.5;
			getModuleKnob(moduleName, "Pan").position = flowCenter(flowUseGrid(ix, 0), gs);
			getModuleKnob(moduleName, "SendReverb").position = flowCenter(flowUseGrid(ix+1, 0), gs);
			getModuleKnob(moduleName, "ToneNoise").position = flowCenter(flowUseGrid(ix, 1), gs);
			getModuleKnob(moduleName, "Click").position = flowCenter(flowUseGrid(ix+1, 1), gs);

			ix = ix + 2.5;
			getModuleKnob(moduleName, "Gain").position = flowCenter(flowUseGrid(ix+1, 0), gs);
			getModuleKnob(moduleName, "Saturation").position = flowCenter(flowUseGrid(ix, 0.5), gs);
			getModuleKnob(moduleName, "GainDynamics").position = flowCenter(flowUseGrid(ix+1, 1), gs);

			auto sw = 2*S;
			auto sh = 300*S;
			getModuleSeparator(moduleName, "1").position = flowCenter(flowSplitX(r, 1.75, gx), sw, sh);
			getModuleSeparator(moduleName, "2").position = flowCenter(flowSplitX(r, 4.25, gx), sw, sh);
			getModuleSeparator(moduleName, "3").position = flowCenter(flowSplitX(r, 5.75, gx), sw, sh);
			getModuleSeparator(moduleName, "4").position = flowCenter(flowSplitX(r, 8.25, gx), sw, sh);
			auto cw = 10*S;
			auto ch = 130*S;
			getModuleConnector(moduleName, "1").position = flowCenter(flowSplitX(r, 1.6, gx), cw, ch);
			getModuleConnector(moduleName, "2").position = flowCenter(flowSplitX(r, 4.1, gx), cw, ch);
			getModuleConnector(moduleName, "3").position = flowCenter(flowSplitX(r, 10.6, gx), cw, ch);
			// getModuleConnector(moduleName, "4").position = flowTop(flowUseGrid(0.5, 0.05), 8*S);
			// getModuleConnector(moduleName, "5").position = flowTop(flowUseGrid(3.0, 0.05), 8*S);

			
			auto labelR = flowTop(flowPaddingY(fullR, 20*S), 25*S);
			getModuleLabel(moduleName, "1").position = flowCenterX(flowSplitX(labelR, 0.5, gx), 100*S);
			getModuleLabel(moduleName, "2").position = flowCenterX(flowSplitX(labelR, 3, gx), 100*S);
			getModuleLabel(moduleName, "3").position = flowCenterX(flowSplitX(labelR, 5, gx), 100*S);
			getModuleLabel(moduleName, "4").position = flowCenterX(flowSplitX(labelR, 7, gx), 100*S);
			getModuleLabel(moduleName, "5").position = flowCenterX(flowSplitX(labelR, 9.5, gx), 100*S);
		}
	}

	override void parameterChanged() {
	    styleModules();
	}
	
	void styleModules() {
	    styleModule("module1", COLOR_RED);
	    styleModule("module2", COLOR_ORANGE);
	    styleModule("module3", COLOR_ORANGE);
	    styleModule("module4", COLOR_ORANGE);
	    styleModule("module5", COLOR_BLUE);
	    styleModule("module6", COLOR_AQUA);
	    styleModule("module7", COLOR_GREEN);
	    styleModule("module8", COLOR_PINK);
	    styleModule("module9", COLOR_PURPLE);
	    styleModule("module10", COLOR_YELLOW);
	}
	
	void styleUI() {
        //System.print("plugin.createUI()");
		//initLayout(); // Call from GUI.d
		
		(cast(UIColorPanel)get("background")).color = darkenColor(COLOR_AQUA, 0.9); //COLOR_GREY7;
		//get("background").color = darkenColor(MODULE_COLORS[getModuleParam("module1", "Color").intValue % 8], 0.9);
		
        
		auto logo = cast(UILabel)get("logoLabel");
		logo.color = COLOR_GREY2;
        logo.shadowColor = COLOR_CLEAR;
		logo.shadowOffset = 0.035;
		//(cast(UIIcon)get("logoIcon")).color = COLOR_GREY2;
		(cast(UIIcon)get("simianIcon")).color = COLOR_GREY8;

        auto versionLabel = cast(UILabel)get("versionLabel");
        if (versionLabel) {
        	versionLabel.color = COLOR_GREY5;
        	//versionLabel.backgroundColor = COLOR_WHITE;
        }

		
		auto tabContainer = cast(UITabContainer)get("tabContainer");
		tabContainer.colorFill = COLOR_GREY9;
		
		foreach (moduleName; MODULE_NAMES) {
			getModulePanel(moduleName).color = COLOR_GREY5;
			}

		styleModules();
	    
	    //UI.root.color = correct(RGBA_new(255,255,255,255))

		foreach (b; [get("creditsButton"), get("upgradeButton")]) {
			auto button = cast(UIButton)b;
			button.backgroundColor = COLOR_CLEAR;
			button.backgroundColorHover = COLOR_CLEAR;
			button.backgroundColorPressed = COLOR_CLEAR;
			button.color = COLOR_WHITE; //COLOR_GREY5; //.withAlpha(128); //Alpha not working on fonts/icons yet
			button.colorHover = COLOR_YELLOW;
			button.colorPressed = COLOR_YELLOW;
			button.shadowColor = blendColor(COLOR_AQUA, COLOR_BLACK, 0.333);
			button.shadowOffset = 0.075;
		}
		(cast(UIButton)get("upgradeButton")).color = COLOR_YELLOW;
		(cast(UIButton)get("upgradeButton")).colorHover = COLOR_AQUA;


        //(cast(UIColorPanel)get("globalPanel")).color = COLOR_CLEAR;
        (cast(UIColorPanel)get("globalPanel")).color = darkenColor(COLOR_RED, 0.8);
        //(cast(UIColorPanel)get("globalPanel")).color = darkenColor(MODULE_COLORS[getModuleParam("module1", "Color").intValue % 8], 0.8);

        
        foreach (knobName; GLOBAL_KNOB_NAMES) {
        	auto knob = getModuleKnob("global", knobName);
			styleKnobBasic(knob);
			styleKnobDark(knob);
        }
        auto knob = getModuleKnob("global", "Gain");
		knob.indicatorStyle = UIKnob.IndicatorStyle.Continuous;
		knob.indicatorClipStart = GAIN_START_CLIP;
		knob.indicatorClipColor = COLOR_CLIP;

        knob = getModuleKnob("global", "Drive");
		knob.indicatorStyle = UIKnob.IndicatorStyle.Continuous;
		knob.indicatorClipStart = GAIN_START_CLIP;
		knob.indicatorClipColor = COLOR_CLIP;

        knob = getModuleKnob("global", "Sensitivity");
		knob.indicatorFillColor = COLOR_ORANGE;
		knob.indicatorStyle = UIKnob.IndicatorStyle.Led;
		
        knob = getModuleKnob("global", "Saturation");
		knob.indicatorFillColor = COLOR_PINK;
		knob.indicatorStyle = UIKnob.IndicatorStyle.Continuous;

        knob = getModuleKnob("global", "ReverbDecay");
		knob.indicatorFillColor = COLOR_PURPLE;
		knob.indicatorStyle = UIKnob.IndicatorStyle.Continuous;
		
        knob = getModuleKnob("global", "ReverbGate");
		knob.indicatorFillColor = COLOR_PURPLE;
		knob.indicatorStyle = UIKnob.IndicatorStyle.Continuous;

	    
	    auto resizer = cast(UIWindowResizer)get("resizerHint");
	    resizer.color = COLOR_GREY9;
	    resizer.colorHovered = COLOR_YELLOW;
	    resizer.colorDragged = COLOR_YELLOW;
	    resizer.colorCannotResize = COLOR_RED;
	    
	}


    override void createUI() {
		styleUI();
    }

    override void reflow() {
	    if(!initLayout())
	    	return;
    	styleUI();

		//System.print("plugin.reflow()");
		
      	
        get("background").position = Rectangle(0,0,W,H);
        get("logoIcon").position = Rectangle(10,10,60,60).scaleByFactor(S);
        get("logoLabel").position = Rectangle(70,15,160,50).scaleByFactor(S);
        get("simianIcon").position = Rectangle(802,467,180,180).scaleByFactor(S);
        auto versionLabel = get("versionLabel");
        if (versionLabel) {
        	versionLabel.position = Rectangle(70,58,160,22).scaleByFactor(S);
        }
		auto buttonSize = 35*S;
		get("creditsButton").position = Rectangle(240*S, 10*S, buttonSize, buttonSize);
        get("presetSelector").position = Rectangle(600,25,380,30).scaleByFactor(S);
        
        get("upgradeButton").position = Rectangle(0,515,112,140).scaleByFactor(S);

        //get("GG").position = Rectangle(DW-70, 6, 64, 64).scaleByFactor(S);
        //UI.getElementById("GG").position = Rectangle(DW-70, 6, 64, 64).scaleByFactor(S);
        //UIElement.new("GG").position = Rectangle(DW-70, 6, 64, 64).scaleByFactor(S);
        
	    (cast(UICreditsPanel)get("creditsPanel")).textSize = 25 * S;

        
        auto y = 80*S;
        auto tabHeight = 64*S;
        auto tabContainer = cast(UITabContainer)get("tabContainer");
		tabContainer.horizontalSkew = 0; //-20*S;
        tabContainer.position = Rectangle(0, y, W, tabHeight);

		y = y + tabHeight;
		auto panelHeight = 330*S;
		foreach (int i; 0..cast(int)MODULE_NAMES.length) {
			reflowModule(i, 0, y, W, panelHeight);
		}
		y = y + panelHeight;
		auto gs = KNOB_SIZE*S;
		
		panelHeight = H-y;
		auto fullR = Rectangle(0, y, W, panelHeight);
		auto r = fullR;
		get("globalPanel").position = r;
		//r = Rectangle(r.origin.x, r.origin.y-15*S, r.size.width, r.size.height);
		auto gx = 11;
		r = flowPaddingX(r, 20*S);
		flowSetGrid(r, gx, 1);
		auto offset = 1; //1.75;
        getModuleKnob("global", "Sensitivity").position = flowCenter(flowUseGrid(0.0+offset, 0), gs);
        getModuleKnob("global", "ReverbDecay").position = flowCenter(flowUseGrid(1.5+offset, 0), gs);
        getModuleKnob("global", "ReverbGate").position = flowCenter(flowUseGrid(2.5+offset, 0), gs);
        getModuleKnob("global", "Drive").position = flowCenter(flowUseGrid(4.0+offset, 0), gs);
        getModuleKnob("global", "Saturation").position = flowCenter(flowUseGrid(5.0+offset, 0), gs);
        getModuleKnob("global", "Gain").position = flowCenter(flowUseGrid(6.0+offset, 0), gs);
		auto sw = 2*S;
		auto sh = 140*S;
		getModuleSeparator("global", "1").position = flowCenter(flowSplitX(r, 0.75+offset, gx), sw, sh);
		getModuleSeparator("global", "2").position = flowCenter(flowSplitX(r, 3.25+offset, gx), sw, sh);
		auto labelR = flowTop(flowPaddingY(r, 5*S), 25*S);
		getModuleLabel("global", "1").position = flowCenterX(flowSplitX(labelR, 0.0+offset, gx), 100*S);
		getModuleLabel("global", "2").position = flowCenterX(flowSplitX(labelR, 2.0+offset, gx), 100*S);
		getModuleLabel("global", "3").position = flowCenterX(flowSplitX(labelR, 5.0+offset, gx), 100*S);

        get("resizerHint").position = Rectangle(W-30, H-30, 30, 30);
    }
}
