
/+ dub.sdl:
	name "dsp_drum"
	dependency "dplug:core" version="*"
+/
module dsp_drum;
/* ------------------------------------------------------------
author: "Oren Kurtz"
copyright: "Punk Labs LLC 2022"
license: "GPLv3-or-later"
name: "OneTrick SIMIAN"
version: "1.00"
Code generated with Faust 2.50.6 (https://faust.grame.fr)
Compilation options: -a arch.fastmath.d.template -lang dlang -cn DSP_Drum -pn processDrum -es 1 -mcd 16 -single -ftz 0
------------------------------------------------------------ */
/************************************************************************
 IMPORTANT NOTE : this file contains two clearly delimited sections :
 the ARCHITECTURE section (in two parts) and the USER section. Each section
 is governed by its own copyright and license. Please check individually
 each section for license and copyright information.
 *************************************************************************/

/*******************BEGIN ARCHITECTURE SECTION (part 1/2)****************/

/************************************************************************
 FAUST Architecture File
 Copyright (C) 2003-2019 GRAME, Centre National de Creation Musicale
 ---------------------------------------------------------------------
 This Architecture section is free software; you can redistribute it
 and/or modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 3 of
 the License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; If not, see <http://www.gnu.org/licenses/>.
 
 EXCEPTION : As a special exception, you may create a larger work
 that contains this FAUST architecture section and distribute
 that work under terms of your choice, so long as this FAUST
 architecture section is not modified.
 
 ************************************************************************
 ************************************************************************/

import std.math;
import std.stdio;
import std.conv;
import std.algorithm;
import std.format;
import core.stdc.stdlib;

/******************************************************************************
 *******************************************************************************
 
 VECTOR INTRINSICS
 
 *******************************************************************************
 *******************************************************************************/


/********************END ARCHITECTURE SECTION (part 1/2)****************/

/**************************BEGIN USER SECTION **************************/

/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This section is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

import onetrick.client;
alias pow = FastMath.pow;
alias sqrt = FastMath.sqrt;

import std.math;
import std.algorithm : min, max;
import dplug.core.nogc: mallocNew, mallocSlice, destroyFree, assumeNothrowNoGC;

const __gshared float[] fDSP_DrumSIG0Wave0 = [-0.09098f,-0.5436f,-0.39318f,0.82093f,0.69092f,-0.27279f,0.13717f,0.27376f,-0.12363f,-0.11062f,0.08682f,0.49528f,-0.20334f,-0.10878f,0.14994f,0.00672f,0.08203f,0.11961f,0.06626f,-0.12063f,-0.37199f,0.0503f,0.55583f,-0.35192f,-0.45857f,-0.43206f,-0.34873f,0.42989f,0.12933f,-0.44314f,-0.21954f,0.17441f,-0.13935f,-0.67854f,0.04735f,0.8383f,0.58805f,0.56923f,-0.28464f,-1.32574f,0.02389f,0.71874f,0.96192f,0.97789f,0.03081f,-0.73876f,-0.2119f,1.1035f,0.25436f,-0.33794f,-0.34592f,0.0295f,0.26665f,-0.21369f,-0.48633f,-0.31975f,0.33775f,0.06549f,-0.11502f,-0.41112f,-0.50897f,-0.54926f,-0.43796f,0.41373f,-0.21262f,-0.21485f,-0.35352f,-0.27816f,0.06636f,-0.28566f,0.33741f,0.17141f,0.01944f,-0.09746f,-0.37417f,0.0326f,1.2558f,0.58273f,-0.64735f,-0.60135f,0.25548f,0.88662f,0.19419f,0.52764f,0.21407f,-0.45484f,-0.43037f,0.25707f,0.66611f,-0.16909f,-0.65982f,0.03753f,0.54079f,-0.19032f,0.07173f,-0.05185f,-0.42998f,0.21378f,-0.20077f,-0.59433f,-0.78548f,-0.1909f,0.7967f,0.10675f,-0.65267f,-0.43772f,-0.4014f,0.03511f,0.8398f,0.66543f,-0.43917f,-0.81499f,0.18172f,0.76831f,0.43303f,-0.15792f,0.06597f,0.5066f,0.4606f,-0.13214f,-0.34017f,0.10031f,0.32488f,0.29185f,0.30321f,0.7716f,-0.04222f,-1.02848f,-0.26036f,0.61562f,0.53533f,0.17016f,-0.16938f,-0.90422f,-0.61407f,0.6769f,0.91878f,-0.14626f,-1.05058f,-0.6487f,0.04421f,0.47767f,-0.0206f,-0.67603f,0.11908f,0.03763f,-0.34408f,0.09296f,-0.16295f,-0.15371f,0.15429f,0.23985f,-0.17325f,-0.10496f,0.15516f,-0.49871f,-0.23535f,0.52299f,0.23216f,-0.23405f,-0.23574f,0.16092f,0.49992f,0.1496f,0.05407f,-0.04459f,-0.16126f,-0.01499f,0.15671f,0.33112f,-0.15062f,0.04208f,-0.00658f,-0.23115f,-0.19869f,-0.15961f,0.19269f,-0.20498f,-0.16691f,-0.38418f,-0.59419f,-0.24237f,0.20232f,0.69001f,0.34761f,-0.34495f,-0.42853f,-0.2341f,0.04634f,0.28271f,0.66843f,0.59825f,-0.85237f,-0.73687f,0.31327f,0.80188f,1.27863f,0.20116f,-0.6782f,-0.58002f,-0.18196f,0.40048f,0.2888f,0.20919f,0.28179f,-0.02307f,0.08329f,0.33586f,-0.40498f,-0.15763f,0.0607f,-0.2251f,-0.0089f,-0.26413f,-0.2813f,-0.06723f,-0.09606f,-0.23187f,-0.06515f,0.1687f,0.13857f,-0.57426f,-0.10563f,0.93252f,0.1391f,-0.3435f,-0.40861f,-0.05915f,0.6384f,0.00406f,-0.42732f,-0.23603f,0.13369f,0.17881f,0.206f,0.61005f,0.07884f,-0.54587f,0.29073f,0.44111f,-0.40193f,-0.61639f,0.05543f,0.94408f,0.12043f,-0.25383f,-0.57001f,-0.44164f,0.36275f,0.33988f,0.12334f,-0.2074f,-0.18109f,-0.22525f,0.03628f,-0.16958f,-0.28367f,-0.22302f,-0.50713f,-0.34113f,0.10641f,0.75104f,0.50016f,0.07869f,0.07453f,-0.25876f,-0.4723f,-0.25257f,0.47588f,0.23535f,-0.44353f,-0.31835f,0.5422f,0.65591f,0.0119f,-0.04339f,0.12919f,0.52159f,0.09649f,-0.46989f,-0.40677f,-0.08082f,0.0652f,0.32832f,0.70137f,0.21185f,-0.674f,-0.11802f,0.74553f,0.66829f,-0.23792f,-0.79356f,0.00987f,0.1673f,0.1154f,-0.00677f,-0.40909f,-0.23076f,0.27758f,-0.01388f,0.0858f,-0.14196f,-1.10006f,-0.14438f,0.78331f,0.34297f,-0.29204f,-0.48362f,-0.01021f,0.19207f,0.29049f,-0.21809f,-0.29959f,0.3925f,-0.18254f,-0.1835f,0.57499f,0.50215f,-0.32841f,-0.44386f,0.21456f,0.44604f,0.02858f,-0.31898f,0.04904f,0.0725f,-0.20169f,0.01398f,0.12097f,0.38877f,0.00813f,-0.01736f,0.29504f,-0.39361f,-0.585f,-0.36464f,0.5051f,0.74776f,0.21809f,-0.72372f,-0.79177f,0.17441f,0.9908f,0.59535f,-0.47052f,-0.37639f,-0.35264f,-0.15792f,0.39743f,0.52856f,-0.13122f,-0.3939f,-0.09712f,0.06573f,-0.01088f,0.13891f,0.14515f,-0.00106f,-0.00058f,-0.30026f,-0.12658f,0.29731f,0.01683f,-0.06128f,0.19497f,0.10713f,0.22597f,-0.12788f,-0.33509f,-0.11908f,-0.34766f,-0.45373f,0.12513f,0.33934f,0.21045f,-0.01693f,-0.44483f,0.03265f,0.47163f,0.31758f,-0.35463f,-0.80928f,0.33639f,0.39284f,-0.36628f,-0.09625f,0.10713f,0.38423f,0.42877f,0.21354f,-0.06631f,-0.30012f,-0.32657f,-0.05978f,0.18317f,0.06588f,-0.36904f,0.49015f,1.07331f,-0.29054f,-0.8856f,-0.06176f,0.62447f,0.27279f,-0.20759f,-0.38549f,-0.31095f,0.39366f,0.25809f,-0.36401f,-0.30795f,-0.21552f,0.08348f,-0.14921f,-0.16285f,0.13151f,0.30142f,-0.16092f,-0.29576f,-0.07231f,0.11105f,0.35622f,-0.47864f,-0.12658f,-0.03449f,-0.35535f,0.28416f,0.27961f,-0.19376f,-0.06351f,0.24575f,0.39463f,0.22728f,-0.24948f,-0.074f,-0.18583f,-0.04358f,0.29117f,0.33562f,0.55612f,-0.01799f,-0.52193f,-0.07918f,0.20764f,0.45489f,0.37209f,-0.34196f,-0.24101f,-0.20556f,-0.27569f,-0.01949f,0.19714f,0.24145f,-0.16904f,-0.00397f,0.25886f,0.15245f,-0.32159f,-0.23516f,0.21141f,0.06486f,-0.12029f,0.02133f,0.37378f,-0.24019f,-0.73296f,-0.10699f,0.68173f,0.85682f,-0.23511f,-0.75312f,-0.08972f,-0.21422f,-0.27463f,0.09006f,0.70215f,0.44425f,-0.47511f,-0.33649f,-0.14205f,-0.13122f,0.08014f,0.36662f,0.01528f,-0.45189f,-0.19898f,0.0326f,-0.36474f,-0.22379f,0.33823f,0.18834f,-0.05016f,-0.06694f,0.012f,0.06505f,-0.07274f,-0.20629f,-0.1186f,0.33025f,0.29538f,0.25456f,0.7167f,0.00793f,-0.89943f,-0.18974f,0.37397f,0.21925f,0.29954f,0.20948f,-0.24566f,-0.45794f,0.25006f,0.18447f,-0.09973f,0.21794f,-0.12875f,-0.41794f,-0.12527f,-0.09132f,-0.38118f,-0.09615f,0.47637f,0.40609f,-0.23468f,-0.95989f,-0.53833f,0.14191f,0.20996f,0.043f,-0.39535f,-0.0697f,0.10791f,0.54074f,0.62045f,-0.10607f,-0.37983f,-0.17049f,0.03216f,0.22036f,0.32246f,0.26433f,0.38863f,0.13949f,-0.01993f,0.1053f,0.36145f,0.13388f,-0.44532f,0.00992f,0.58515f,0.19255f,-0.1318f,-0.39613f,-0.11952f,0.13954f,0.0147f,-0.05528f,0.04276f,-0.18172f,-0.41518f,-0.08193f,0.02215f,0.18626f,-0.49741f,-0.71603f,-0.04247f,0.0016f,-0.24991f,-0.18912f,0.32919f,0.74539f,-0.10718f,-0.7968f,-0.33726f,0.3109f,0.35946f,-0.50785f,-0.59854f,0.10578f,0.45712f,0.67032f,0.31405f,-0.35622f,-0.44817f,-0.27618f,0.45412f,0.50964f,-0.23294f,-0.25606f,0.42974f,0.34578f,-0.01156f,0.14278f,0.40609f,0.50974f,-0.2324f,-0.87719f,-0.79844f,0.19492f,0.8397f,0.29262f,-0.37267f,-0.03158f,0.32778f,-0.04285f,-0.5096f,-0.64522f,0.22036f,0.34703f,0.02297f,0.17219f,0.0029f,-0.40952f,-0.1304f,0.6413f,0.22491f,-0.41871f,-0.37257f,0.13857f,0.38708f,0.13886f,-0.474f,0.10341f,0.53833f,-0.54872f,-0.2371f,0.17088f,0.33451f,-0.15211f,-0.22897f,0.34795f,0.38553f,0.19226f,-0.62273f,-0.39202f,0.09282f,0.4458f,-0.03647f,-0.00281f,0.1793f,-0.26462f,0.03981f,0.31593f,0.23061f,-0.68527f,-0.28246f,0.59893f,0.50785f,0.18631f,-0.44807f,-0.25035f,0.41141f,0.07521f,-0.2608f,0.06825f,0.15352f,-0.32072f,-0.6723f,0.31003f,0.58587f,-0.2755f,0.3998f,0.22655f,-0.28362f,-0.45446f,-0.93436f,-0.1778f,0.51284f,0.72415f,0.14239f,-0.26771f,-0.35376f,-0.07241f,0.05079f,0.39023f,0.28633f,-0.72589f,-0.73794f,-0.41257f,0.66447f,0.29494f,-0.31013f,-0.01156f,-0.00082f,0.19342f,0.01548f,-0.06307f,-0.12546f,-0.06027f,-0.4903f,-0.20537f,0.37189f,0.22186f,0.01103f,0.02921f,0.17746f,0.06791f,0.02089f,-0.12508f,0.31782f,0.33015f,-0.09364f,-0.12779f,0.41078f,0.23124f,-0.63598f,-0.32735f,0.4399f,0.62379f,0.24745f,-0.19071f,-0.63801f,-0.2681f,0.41576f,0.09722f,-0.13548f,0.1393f,-0.01591f,-0.2341f,-0.13156f,0.03584f,-0.24053f,-0.01969f,0.44749f,0.16145f,-0.74892f,-0.55995f,0.48643f,0.63772f,0.55545f,-0.69131f,-1.10683f,0.19284f,0.84807f,0.17775f,-0.49126f,-0.01316f,-0.25939f,-0.1556f,0.5656f,0.59153f,-0.06157f,-0.47027f,0.02752f,-0.05659f,0.10118f,0.20793f,0.05751f,0.43453f,0.43114f,-0.62505f,-0.35453f,0.58892f,0.41175f,-0.14636f,-0.30157f,0.03739f,0.15922f,0.10926f,-0.56793f,-0.06863f,0.18597f,0.36087f,0.18824f,-0.48778f,-0.29963f,-0.62055f,-0.05359f,0.27855f,-0.25881f,-0.07632f,0.34631f,0.12991f,-0.42698f,-0.02002f,0.3362f,0.05838f,-0.36014f,-0.34336f,0.0f,-0.00779f,0.44604f,0.28928f,-0.07221f,-0.39724f,-0.18336f,0.50079f,0.71675f,0.25572f,-0.50873f,-0.66079f,-0.34505f,0.23182f,0.38892f,0.36604f,0.14186f,-0.16392f,-0.151f,-0.23226f,-0.24261f,0.60348f,0.45939f,-0.10549f,-0.47061f,-0.78606f,0.11719f,0.38636f,0.29572f,0.17901f,-0.07236f,0.22708f,0.0934f,-0.25973f,-0.07052f,0.36677f,0.18945f,-0.23042f,-0.02728f,0.06863f,-0.13204f,-0.02472f,0.05098f,0.22065f,0.39995f,-0.24759f,-0.24367f,0.10747f,0.06791f,0.30786f,0.15734f,-0.32217f,-0.28517f,-0.13494f,-0.32362f,-0.17959f,0.11835f,0.02085f,-0.42128f,0.27734f,0.2769f,-0.50544f,-0.1452f,0.52241f,0.34476f,-0.9197f,-0.76884f,-0.10646f,0.42984f,0.74311f,-0.20314f,-0.56667f,0.07531f,0.63163f,0.52464f,-0.09997f,0.24527f,0.28764f,-0.3926f,-0.5138f,-0.03599f,0.05209f,0.24658f,0.69624f,0.27826f,0.00924f,-0.58278f,-0.24338f,0.38268f,0.22839f,0.06771f,-0.12672f,-0.13262f,-0.19976f,-0.6724f,0.09809f,0.7747f,0.22191f,-0.4443f,-0.79665f,-0.1258f,0.65736f,0.39574f,-0.65136f,-0.37813f,-0.00469f,-0.354f,-0.14085f,0.50442f,0.31381f,-0.44406f,-0.47003f,0.39975f,0.30331f,-0.27816f,-0.18868f,0.48444f,0.60585f,-0.21199f,-0.60019f,0.46108f,1.2115f,-0.07971f,-0.35526f,-0.14447f,0.15062f,-0.10249f,-0.47821f,0.51511f,0.22084f,-0.0252f,-0.01108f,-0.35506f,0.2102f,-0.36f,-0.51666f,0.0962f,0.43095f,-0.23187f,-0.90006f,-0.28043f,0.45673f,0.32338f,-0.18616f,0.13064f,0.30718f,0.1598f,-0.53276f,-0.35632f,0.08372f,0.28459f,0.23061f,0.09649f,-0.27023f,-0.19497f,0.55632f,0.69958f,0.13456f,-0.91796f,-0.72894f,0.03821f,0.80817f,0.62471f,-0.27947f,-0.40696f,-0.31284f,-0.35279f,-0.41451f,0.20111f,0.98887f,0.49068f,-0.76918f,-0.88633f,0.0828f,0.41107f,0.57301f,0.14114f,0.00595f,-0.21717f,-0.72836f,0.02718f,0.32067f,0.6902f,0.22655f,-0.37842f,0.25165f,0.15797f,-0.16837f,0.16537f,0.51119f,0.35564f,-0.35264f,-0.37223f,0.36208f,0.44972f,-0.15627f,-0.72072f,0.09543f,0.60696f,-0.05383f,0.14849f,0.19105f,-0.08063f,-0.08798f,-0.33006f,-0.17301f,-0.25223f,-0.44212f,-0.16919f,-0.05262f,0.4339f,0.3926f,-0.64822f,-0.50399f,0.03332f,-0.07777f,-0.02249f,-0.33349f,-0.53794f,-0.24392f,-0.0179f,0.26423f,0.26142f,0.03062f,0.02355f,-0.12077f,0.18022f,-0.05649f,-0.35148f,0.61644f,0.68909f,0.01654f,-0.20038f,-0.0059f,0.41634f,0.194f,0.55937f,0.82606f,-0.53305f,-0.5629f,0.0666f,0.22636f,0.35705f,-0.20005f,-0.26728f,-0.19821f,-0.12508f,0.04677f,-0.22858f,0.13146f,0.18496f,-0.47443f,-0.7317f,-0.1733f,0.52841f,0.17432f,0.02099f,-0.19037f,-0.55259f,-0.16072f,0.28875f,0.32324f,-0.07768f,-0.14994f,0.2711f,0.40058f,0.04605f,-0.16687f,-0.11937f,0.3124f,0.47526f,-0.06322f,-0.14089f,-0.49808f,-0.04692f,0.69721f,0.11037f,-0.11976f,-0.22471f,-0.17997f,0.14321f,0.42041f,-0.145f,-0.48807f,0.01045f,0.37267f,-0.10718f,-0.56759f,0.42021f,0.49436f,-0.03729f,-0.27971f,-0.02515f,0.07168f,-0.16798f,-0.01949f,-0.0015f,0.19023f,0.06346f,-0.03985f,-0.31937f,-0.32977f,0.29069f,0.70742f,0.23144f,-0.46887f,-0.34882f,-0.17011f,0.1378f,0.01098f,-0.08774f,-0.27719f,-0.6279f,0.1006f,0.64657f,0.34834f,-0.45528f,-0.26491f,0.07434f,-0.41281f,-0.4696f,0.02525f,0.10892f,0.01006f,0.42244f,0.11748f,-0.43356f,-0.25325f,0.27177f,0.54935f,0.29194f,-0.38447f,-0.60154f,0.0727f,0.82954f,0.50128f,-0.0207f,0.33978f,0.22167f,-0.48212f,-0.38863f,0.36943f,0.41364f,-0.25122f,0.04276f,0.19168f,-0.07013f,0.0873f,0.05209f,0.12164f,-0.29417f,-0.47588f,-0.12063f,0.18858f,0.35995f,0.04251f,-0.21833f,0.29557f,0.16687f,-0.56604f,-0.4857f,0.18326f,0.49562f,0.2324f,-0.5125f,-0.36377f,0.31705f,-0.01296f,-0.12329f,-0.05204f,0.23366f,-0.15153f,-0.31811f,0.22834f,-0.03647f,-0.60744f,-0.02946f,0.38984f,-0.02839f,-0.12904f,-0.41968f,-0.18113f,0.3405f,0.66887f,0.37402f,-0.10554f,-0.17949f,-0.29156f,-0.006f,0.50461f,0.64623f,-0.14481f,-0.45049f,0.20793f,0.30752f,0.22994f,0.1185f,-0.0993f,-0.11536f,-0.56067f,-0.89508f,0.6309f,0.93919f,-0.12469f,-0.34225f,-0.25963f,0.22118f,-0.154f,-0.22674f,0.45484f,0.18084f,-0.56996f,-0.40957f,-0.01325f,0.23589f,0.35115f,0.26055f,0.16977f,-0.4533f,-0.8382f,-0.36938f,0.21925f,0.46752f,0.16619f,-0.26945f,-0.53184f,0.00411f,0.41581f,0.32619f,0.04556f,-0.63061f,-0.10515f,0.15037f,0.14133f,0.10293f,-0.10447f,-0.27666f,0.28846f,0.78892f,-0.13427f,-0.25847f,0.15995f,0.36807f,0.18215f,0.17378f,-0.23893f,-0.26452f,0.31709f,0.16513f,0.01712f,0.09001f,0.0015f,-0.55013f,-0.53586f,-0.13552f,0.09838f,0.00943f,-0.15279f,0.37852f,0.17852f,-0.64198f,-0.27714f,0.39351f,0.16914f,-0.50186f,-0.33257f,0.25581f,0.35787f,-0.05819f,-0.36f,0.28807f,0.5627f,0.14612f,-0.72101f,-0.94534f,0.25712f,0.54321f,-0.06568f,-0.24566f,-0.39637f,0.11589f,0.57334f,0.4429f,0.37833f,-0.38616f,-0.34969f,0.3109f,0.34631f,-0.02767f,-0.39129f,-0.02917f,0.40822f,0.25485f,-0.11211f,0.20953f,0.18118f,0.19792f,0.39598f,-0.11511f,-0.38752f,-0.20629f,0.16837f,-0.17383f,-0.5347f,0.15922f,0.59564f,-0.16164f,-0.4844f,-0.04469f,0.21156f,0.28532f,-0.2267f,-0.67013f,-0.82413f,0.15424f,1.07723f,0.2576f,-0.1971f,-0.01756f,0.09045f,0.01054f,-0.19216f,-0.41944f,-0.12716f,0.32633f,0.3393f,0.01852f,-0.49591f,-0.21209f,0.2059f,0.42283f,0.1778f,-0.3896f,-1.05344f,-0.75017f,0.53654f,0.78771f,0.0665f,-0.73155f,-0.14646f,0.23139f,-0.02249f,0.27632f,0.04977f,-0.22911f,-0.30515f,-0.05296f,0.31139f,0.29514f,0.48082f,0.5761f,-0.0695f,-0.3909f,0.25175f,0.15565f,-0.2504f,0.05901f,0.24662f,-0.1822f,-0.23966f,0.16977f,0.42287f,0.09852f,-0.06892f,0.24856f,0.17886f,-0.10931f,-0.69373f,0.14815f,0.76227f,0.1643f,-0.39806f,-0.61083f,0.08672f,0.52493f,0.23685f,-0.53402f,-0.63545f,0.10041f,0.39535f,-0.51917f,-0.82035f,0.19676f,0.01388f,-0.13843f,0.61344f,0.34099f,-0.29296f,-0.24275f,-0.12426f,0.01021f,0.34873f,0.04745f,-0.05407f,-0.06118f,-0.0415f,0.24687f,0.50074f,0.52783f,0.15502f,-0.24488f,-0.48493f,0.08145f,0.40556f,0.56227f,-0.02036f,-0.32657f,0.02399f,-0.26602f,0.11487f,0.30931f,0.01993f,-0.58742f,-0.66026f,-0.18399f,-0.22084f,-0.11976f,0.03831f,0.46104f,-0.10433f,-0.59608f,-0.58195f,-0.02027f,0.83951f,0.24527f,-0.27255f,-0.53794f,0.20957f,0.11744f,0.06757f,0.91849f,-0.08517f,-1.58489f,-0.55342f,1.24676f,0.80347f,-0.31405f,-0.81392f,-0.13983f,0.40793f,0.3155f,0.11797f,0.10002f,0.00479f,0.02462f,-0.07337f,-0.06501f,0.32106f,0.08087f,0.04455f,0.22844f,0.0281f,-0.22394f,-0.20029f,0.17896f,0.52372f,0.02012f,-0.49247f,-0.16556f,0.45097f,0.26283f,-0.28711f,-0.06849f,0.59796f,0.02602f,-0.3362f,0.40198f,0.08193f,0.02829f,-0.17175f,-0.30515f,0.00401f,0.09848f,0.123f,0.17978f,0.07773f,-0.6161f,-0.28836f,0.50873f,0.76067f,-0.60826f,-1.23849f,0.02109f,0.60677f,0.42964f,-0.76536f,-0.64633f,0.20435f,0.44029f,0.31869f,-0.52923f,-0.24072f,0.28957f,-0.42829f,-0.18834f,0.86911f,0.04203f,-0.6293f,0.22031f,0.3983f,-0.37485f,-0.33422f,0.43656f,0.68014f,0.39405f,-0.14965f,-0.36261f,0.19937f,0.50892f,-0.29049f,-0.2341f,0.39119f,0.45194f,-0.30442f,-0.83989f,0.04087f,0.36024f,0.13533f,-0.14104f,-0.03623f,0.13456f,-0.33842f,-0.04034f,0.25751f,0.02181f,-0.44677f,-0.44029f,0.20575f,0.08372f,-0.09494f,-0.39564f,0.00629f,0.32271f,-0.00102f,-0.02264f,-0.20986f,-0.27951f,0.0045f,0.09291f,-0.23032f,0.08479f,0.17799f,0.14394f,0.66592f,0.37238f,-0.22747f,-0.23376f,0.02322f,0.55138f,-0.11444f,-0.12953f,0.22941f,-0.38287f,-0.05214f,0.15047f,0.01519f,0.08788f,0.05499f,0.13344f,-0.19821f,-0.18689f,0.12764f,-0.11787f,-0.11995f,0.07328f,0.33098f,0.15531f,-0.07008f,-0.60589f,-0.1643f,0.45862f,0.15313f,0.19318f,-0.16725f,-0.35202f,-0.41581f,0.04605f,0.67124f,0.27303f,-0.06462f,0.00866f,0.19202f,0.20246f,-0.2296f,-0.67085f,-0.32285f,0.28208f,0.09891f,-0.1882f,-0.02409f,0.15357f,-0.09625f,-0.17649f,0.08532f,-0.05944f,-0.52614f,-0.1939f,0.39192f,0.2843f,-0.22568f,-0.41905f,0.30462f,0.55235f,-0.05006f,-0.28445f,-0.20329f,0.01374f,0.11932f,-0.27158f,-0.19023f,0.50805f,0.40406f,0.02621f,-0.03202f,0.25717f,0.29315f,-0.40106f,-0.20377f,0.43453f,0.61707f,0.07942f,0.13436f,0.28038f,-0.57649f,-0.60778f,0.28019f,0.57155f,-0.29828f,-0.57233f,-0.29915f,-0.07163f,0.39429f,0.08982f,-0.45838f,-0.60096f,-0.11168f,0.43917f,-0.24217f,-0.09794f,-0.22563f,-0.79506f,0.05253f,0.63714f,0.1082f,-0.65697f,-0.0564f,0.68227f,0.06641f,-0.52169f,0.0667f,0.68957f,0.70205f,-0.1556f,-0.91588f,-0.04285f,0.68019f,0.38336f,0.21751f,0.47052f,0.58708f,-0.12498f,0.24111f,0.36894f,-0.72009f,-0.53165f,0.07241f,0.61039f,0.22529f,-0.24164f,-0.23028f,-0.08948f,0.24614f,-0.60623f,-0.62848f,0.49412f,0.30142f,-0.65552f,-0.89349f,-0.22612f,0.35637f,0.16904f,-0.06675f,0.20638f,-0.10583f,-0.43864f,-0.17915f,0.1067f,0.21117f,-0.07371f,0.01872f,0.71414f,0.55356f,-0.25625f,-0.42321f,-0.19415f,0.58394f,0.1865f,-0.21615f,0.16251f,0.77324f,0.56256f,-0.73953f,-0.51259f,0.28609f,0.11589f,-0.12556f,0.46273f,0.37262f,-0.65368f,-0.91046f,0.28875f,0.64285f,0.08324f,-0.18795f,-0.32527f,-0.32817f,-0.1081f,-0.24662f,-0.03753f,0.28827f,0.24387f,-0.03439f,-0.50447f,0.07342f,0.19831f,-0.10907f,-0.24067f,0.09838f,0.10675f,-0.20251f,0.29789f,0.27893f,-0.20474f,0.00479f,0.02878f,0.47501f,0.08619f,-0.56236f,0.2429f,0.32943f,0.04285f,-0.47704f,-0.05538f,0.30515f,-0.23893f,-0.17669f,0.12817f,0.1127f,-0.15782f,-0.11105f,-0.0933f,0.05649f,-0.08227f,-0.45383f,0.02438f,0.44938f,-0.18549f,-0.89474f,-0.27119f,1.06746f,0.93537f,-0.41838f,-0.6783f,-0.17789f,0.2563f,0.11637f,0.19357f,0.45252f,-0.10549f,-0.15893f,-0.16672f,0.33615f,0.67177f,0.14989f,-0.31443f,-0.11226f,0.16706f,-0.25596f,-0.23976f,0.19608f,0.37707f,-0.35052f,-0.51022f,0.15468f,0.2606f,-0.01785f,-0.4547f,-0.46012f,-0.10873f,-0.08851f,0.30418f,0.363f,-0.18544f,-0.34974f,0.01741f,0.11961f,-0.22084f,-0.31685f,0.07095f,0.61658f,-0.00948f,-0.05964f,0.31139f,-0.01736f,-0.05935f,0.24213f,0.10085f,-0.35371f,0.00909f,0.64401f,0.3289f,-0.21577f,0.2297f,0.18225f,0.41117f,0.39356f,-0.46732f,-0.68222f,-0.52914f,0.33296f,0.5923f,0.24474f,-0.28362f,-0.25344f,-0.09166f,-0.27245f,-0.00489f,-0.12121f,0.30171f,-0.16609f,-0.36764f,0.03274f,-0.03744f,0.26969f,0.01132f,-0.6295f,-0.74065f,0.18824f,0.30084f,0.21364f,-0.02505f,-0.32256f,0.03265f,0.28856f,0.3911f,0.03347f,-0.10022f,-0.03357f,-0.16251f,-0.04019f,0.40648f,0.17504f,0.32503f,0.76551f,0.10776f,-0.89397f,-1.14664f,-0.02317f,0.83477f,0.62824f,-0.28333f,-0.31516f,0.23158f,-0.12372f,-0.07695f,-0.19298f,0.14781f,0.24851f,-0.25228f,-0.10689f,0.10012f,-0.00924f,0.15473f,0.20242f,-0.11826f,0.11033f,0.08672f,-0.26404f,-0.48507f,0.22259f,0.29465f,-0.17364f,0.25635f,0.48382f,-0.14878f,-0.2325f,0.10234f,-0.23511f,-0.41987f,0.10351f,0.65044f,0.31579f,0.35956f,-0.11613f,-0.72149f,-0.45804f,-0.21785f,0.35589f,0.35516f,-0.16841f,-0.10278f,0.222f,0.52116f,0.1127f,-0.18742f,-0.21818f,-0.13306f,-0.0414f,0.43767f,0.08155f,-0.73707f,0.0579f,0.08145f,0.22292f,0.43337f,0.13751f,-0.4844f,-0.96226f,-0.04566f,0.59922f,0.22655f,-0.05069f,-0.06288f,0.21509f,0.12614f,-0.49567f,-0.34568f,0.60889f,0.53064f,-0.60067f,-0.55279f,0.37949f,0.56972f,-0.37354f,-0.4532f,0.3437f,-0.30404f,-0.66306f,-0.18815f,0.33668f,0.86442f,0.16116f,-0.51071f,-0.30524f,0.2267f,0.00788f,0.14762f,0.21572f,-0.30757f,-0.83438f,-0.45939f,0.72048f,0.48928f,0.22916f,0.05253f,0.0948f,0.5006f,-0.09132f,-0.64865f,-0.17717f,0.26249f,0.27772f,0.00305f,0.20614f,-0.07763f,-0.4503f,-0.10152f,0.23487f,0.30167f,-0.40585f,-0.63646f,-0.04832f,0.33688f,-0.00861f,-0.25364f,-0.15057f,-0.12048f,-0.23134f,0.12561f,0.66771f,0.50195f,-0.04721f,-0.32493f,-0.13973f,0.43134f,0.47796f,0.31361f,0.2341f,-0.25625f,-0.31027f,-0.50534f,0.07071f,0.99273f,0.4931f,-0.56425f,-0.67356f,0.49291f,0.35023f,-0.7686f,-0.16953f,0.62002f,0.25011f,-0.54157f,-0.52169f,0.06728f,0.29992f,0.0f,-0.37983f,-0.14205f,-0.23313f,-0.06646f,0.19289f,0.24595f,-0.08092f,-0.63598f,-0.23148f,0.28798f,0.29407f,0.02012f,-0.42979f,0.06559f,0.51506f,-0.04329f,-0.12072f,0.21886f,0.2887f,-0.22674f,-0.20348f,0.33368f,0.0739f,-0.28561f,-0.11598f,-0.08387f,0.51613f,0.54249f,-0.05538f,-0.09175f,-0.36788f,-0.39322f,-0.02147f,0.09248f,0.08024f,-0.16953f,-0.33093f,0.27274f,0.5406f,-0.01964f,-0.74534f,-0.36532f,0.17301f,-0.19898f,-0.502f,0.02481f,0.46582f,0.13248f,0.09291f,0.1423f,-0.14292f,-0.35197f,0.10883f,0.95511f,0.3599f,-0.72942f,-0.74925f,0.05378f,0.71366f,0.6841f,0.3229f,0.03004f,-0.03149f,-0.36991f,-0.43617f,-0.049f,0.69518f,0.43438f,0.02438f,-0.10341f,-0.4532f,-0.47729f,0.07661f,0.50442f,0.1199f,0.00856f,-0.25828f,-0.29451f,0.15966f,0.12058f,0.00861f,0.00377f,-0.49983f,-0.348f,0.08842f,0.22089f,0.12469f,-0.17398f,-0.12314f,-0.09901f,-0.05688f,0.55482f,0.45963f,-0.18549f,-0.54824f,-0.66287f,-0.4207f,0.27492f,0.56057f,0.02752f,0.14829f,-0.21238f,-0.37344f,-0.07274f,-0.08459f,0.17543f,0.2087f,0.18612f,0.41915f,-0.23385f,-0.68149f,-0.2089f,0.20633f,0.14544f,0.1897f,0.25741f,0.2517f,0.20014f,-0.20188f,-0.05079f,-0.12189f,-0.01886f,-0.03976f,-0.05751f,-0.03589f,-0.19589f,-0.08213f,0.15743f,0.21117f,0.3554f,0.2355f,-0.53509f,-0.44638f,0.27729f,0.26892f,-0.29344f,-0.35647f,-0.09702f,0.02191f,0.36803f,0.56773f,0.04295f,-0.64464f,-0.6694f,0.06612f,0.36208f,0.53213f,0.2652f,-0.08039f,0.08803f,-0.20474f,-0.64469f,-0.19531f,0.44962f,0.34839f,0.03453f,-0.11744f,0.07879f,-0.30002f,-0.1821f,0.3378f,0.30036f,0.42307f,-0.02443f,-0.49141f,-0.31482f,0.01828f,0.18926f,0.29054f,0.02505f,-0.23366f,0.07352f,0.1022f,0.27627f,-0.06409f,-0.47506f,-0.25223f,0.08039f,0.54752f,0.35018f,-0.18713f,-0.14476f,-0.11627f,-0.51704f,-0.51192f,0.32092f,0.64294f,-0.14404f,-0.00769f,-0.03318f,-0.46607f,-0.26969f,0.3065f,0.8174f,0.03125f,-0.63912f,-0.32711f,-0.07777f,0.00735f,-0.15192f,0.15913f,0.82195f,0.37088f,-0.39666f,-0.47825f,0.13224f,0.54819f,0.3331f,-0.50868f,-0.33233f,0.14737f,-0.19951f,-0.07603f,0.35589f,0.53141f,-0.26786f,-0.40024f,0.30423f,0.8189f,0.19893f,-0.46331f,-0.12106f,-0.04169f,-0.35879f,-0.36382f,0.4353f,0.42718f,-0.01446f,0.17901f,-0.07144f,-0.08754f,0.02742f,0.30524f,0.48072f,-0.16677f,-0.40445f,-0.70393f,-0.28329f,0.12503f,0.34834f,0.431f,-0.32599f,-0.32082f,0.30945f,0.15642f,-0.50437f,-0.29886f,-0.10631f,-0.23071f,-0.02138f,0.19661f,-0.14578f,-0.26549f,-0.006f,0.08271f,-0.19023f,-0.19782f,0.17112f,-0.08508f,0.26379f,0.51816f,0.09881f,-0.18036f,-0.23424f,0.00885f,-0.14912f,0.18157f,0.20716f,-0.19231f,0.19739f,0.90127f,0.24459f,-0.65634f,-0.13441f,0.60232f,0.41659f,-0.99361f,-0.96192f,0.126f,0.51076f,0.26534f,0.41949f,0.35255f,-0.39767f,-0.5597f,-0.03463f,0.37131f,0.21238f,-0.12769f,-0.35134f,-0.39913f,-0.19439f,0.00276f,-0.02041f,0.41741f,0.05243f,-0.42931f,-0.53484f,0.08924f,0.70195f,0.40188f,-0.12532f,-0.34147f,0.23748f,-0.05248f,-0.6013f,0.10123f,0.92105f,0.44406f,-0.20087f,-0.01132f,0.22829f,-0.1689f,-0.0475f,0.19806f,0.09335f,-0.10418f,-0.22307f,0.01272f,-0.05915f,0.30196f,0.3985f,-0.19889f,-0.46036f,-0.00106f,-0.00861f,-0.43351f,-0.01669f,0.00305f,-0.1732f,-0.01586f,0.38984f,0.16454f,-0.53199f,-0.18641f,0.08222f,-0.55346f,-0.44164f,0.30341f,0.25799f,0.44348f,0.20556f,0.12653f,-0.18409f,-0.59259f,0.00261f,0.54118f,0.36532f,-0.08024f,-0.07502f,-0.03328f,-0.1763f,-0.19487f,0.4502f,0.14452f,-0.10099f,0.04145f,-0.40604f,-0.09838f,0.60396f,0.30873f,-0.08803f,-0.16247f,-0.07526f,-0.08842f,-0.30969f,-0.52788f,-0.32343f,0.35376f,0.43322f,-0.16285f,-0.53596f,0.42133f,0.57296f,-0.07748f,-0.55675f,-0.39163f,-0.1659f,0.00135f,0.68009f,0.6367f,0.14984f,-0.57692f,0.04687f,0.76052f,-0.15487f,-0.55163f,0.2489f,0.88947f,0.34268f,-0.81697f,-0.85421f,0.05838f,0.32261f,0.35816f,0.28595f,-0.00992f,-0.01025f,-0.33252f,-0.23298f,0.18699f,0.19139f,0.05122f,-0.27129f,-0.35908f,-0.40164f,-0.28159f,0.09799f,0.29596f,0.35144f,0.39443f,-0.13485f,-0.26118f,0.30795f,0.09964f,-0.60401f,-0.51211f,0.07361f,-0.01823f,0.16363f,0.72807f,0.58834f,-0.35661f,-0.70413f,-0.23792f,0.12885f,0.48551f,0.25765f,-0.04846f,-0.2415f,-0.44333f,-0.34815f,-0.26297f,0.55274f,0.85276f,0.00895f,-0.6693f,-0.63327f,-0.0104f,0.3792f,0.57252f,0.31782f,-0.00377f,-0.72701f,-0.6101f,0.03429f,-0.20396f,-0.17814f,-0.11178f,0.54945f,0.54824f,-0.34747f,-0.14868f,0.22239f,0.04527f,-0.28895f,-0.1391f,0.50026f,0.157f,-0.18355f,0.28928f,0.55917f,0.34524f,-0.45291f,-0.53712f,0.36919f,0.41214f,-0.15748f,-0.10786f,-0.04513f,-0.32919f,-0.00542f,-0.04967f,-0.24672f,0.13204f,0.23415f,0.01504f,-0.38278f,-0.09393f,-0.09736f,-0.37678f,0.22694f,0.52836f,-0.29364f,-0.76613f,-0.10123f,0.22936f,0.33064f,0.30031f,-0.14737f,0.11448f,0.11942f,-0.15207f,-0.03115f,0.33634f,-0.1466f,-0.82596f,0.0815f,0.76788f,0.53349f,-0.00798f,-0.16469f,-0.31555f,0.08793f,0.19753f,-0.29843f,-0.21402f,-0.09997f,0.17683f,0.05422f,-0.00358f,0.22868f,0.04982f,0.13311f,0.44425f,0.25959f,-0.30495f,-0.87951f,-0.3731f,0.02157f,-0.00972f,0.26824f,0.45876f,0.31294f,-0.28058f,-0.60701f,-0.49146f,0.19511f,0.52338f,0.23463f,-0.09857f,-0.06476f,-0.20866f,-0.3021f,-0.21891f,0.17852f,0.49122f,0.01301f,-0.22665f,0.01833f,0.19061f,-0.17335f,0.13983f,0.57794f,-0.1673f,-0.693f,0.10118f,0.32899f,0.13011f,-0.24309f,-0.44014f,0.60502f,0.68754f,0.13359f,-0.11511f,-0.25291f,-0.21494f,-0.46316f,-0.51545f,0.40358f,0.57421f,-0.31342f,-0.27666f,0.04769f,0.33069f,-0.09499f,-0.27013f,0.43801f,-0.03961f,-0.38442f,0.13374f,0.02684f,-0.26665f,-0.19579f,0.10728f,0.53886f,0.046f,-0.55738f,-0.36633f,0.16943f,0.5599f,0.61373f,-0.02114f,-0.4621f,-0.40493f,-0.0742f,-0.00605f,-0.3894f,0.54839f,0.64933f,-0.38249f,-0.24077f,0.25746f,0.08614f,0.25731f,0.09088f,-0.09112f,-0.24866f,-0.70311f,-0.22452f,0.45871f,0.6872f,0.03241f,-0.07661f,0.27366f,0.09117f,-0.26839f,-0.58278f,-0.10539f,0.46926f,0.11318f,0.01911f,0.63051f,0.01852f,-0.67225f,-0.22399f,0.32266f,0.19821f,-0.45199f,-0.23221f,0.20798f,0.24948f,-0.25383f,-0.30936f,0.20053f,0.25901f,-0.20184f,-0.56411f,-0.14834f,0.15913f,-0.13678f,-0.14926f,0.18974f,0.01852f,0.10133f,0.42621f,-0.222f,-0.40246f,-0.1999f,-0.29073f,0.66205f,0.78708f,0.24285f,-0.31603f,-0.27226f,0.36459f,-0.01166f,-0.10563f,0.17693f,0.15187f,0.2503f,-0.27192f,-0.3939f,0.45978f,0.34113f,0.03366f,-0.22186f,-0.21301f,-0.07777f,-0.40856f,-0.00102f,0.24363f,-0.02491f,0.02447f,-0.31076f,-0.50573f,0.05698f,0.14941f,-0.00063f,0.01911f,0.15081f,0.10191f,-0.38312f,-0.13688f,0.18713f,-0.02147f,-0.59042f,-0.18143f,0.38728f,0.09722f,0.14554f,0.04034f,0.32092f,0.42887f,0.01509f,-0.25093f,-0.53678f,-0.23961f,0.01103f,0.09978f,0.80633f,0.38249f,-0.12309f,0.24832f,0.38718f,0.1038f,-0.55124f,-0.57291f,0.04638f,0.55937f,0.12126f,-0.34597f,0.04063f,0.12585f,-0.12445f,-0.03894f,0.52164f,0.20725f,-0.81025f,-0.26462f,0.21519f,-0.22515f,-0.48865f,-0.0325f,0.63907f,0.23148f,-0.35564f,-0.55864f,-0.11995f,0.3704f,0.265f,0.01906f,0.29533f,0.31279f,-0.04837f,-0.13731f,-0.4132f,0.03753f,0.08527f,0.35859f,0.80323f,-0.32774f,-1.01629f,-0.37509f,0.46176f,0.53064f,-0.04605f,-0.47859f,-0.1732f,0.18147f,-0.31148f,-0.40091f,0.03642f,0.06457f,0.48444f,-0.01862f,-0.54403f,-0.05407f,0.45605f,0.07076f,-0.57141f,0.16924f,0.17296f,-0.22544f,-0.09451f,0.33078f,0.40846f,-0.38752f,-0.39434f,0.27695f,0.41059f,0.29789f,-0.21968f,-0.33625f,-0.01998f,-0.01959f,-0.09572f,-0.06264f,0.44285f,-0.1585f,-0.54684f,0.12343f,0.57223f,0.22568f,-0.67733f,-0.09838f,0.48159f,-0.19889f,-0.46249f,0.11613f,0.17059f,-0.05577f,0.04102f,0.48836f,0.52256f,-0.42727f,-0.51811f,0.05669f,0.22863f,0.02805f,-0.18176f,-0.06887f,0.05485f,-0.08208f,0.38055f,0.62689f,0.36188f,-0.10626f,-0.51463f,-0.1895f,0.04716f,-0.03589f,-0.14815f,-0.05204f,0.31477f,0.1318f,-0.69605f,0.10655f,0.41039f,-0.47279f,-0.17224f,-0.23603f,-0.14907f,0.03303f,-0.27768f,-0.0222f,0.0577f,0.04f,-0.14176f,-0.26191f,0.19521f,0.03797f,-0.32164f,0.19603f,0.62568f,0.32803f,-0.24275f,-0.28783f,-0.06476f,-0.04435f,-0.17669f,0.08474f,0.25794f,0.11168f,0.19139f,-0.23105f,-0.00339f,0.47206f,0.1762f,-0.36009f,-0.51825f,-0.22394f,0.21132f,0.0547f,0.03869f,0.13315f,-0.03618f,-0.30002f,0.00537f,0.63598f,0.09717f,-0.29165f,-0.43197f,0.00953f,0.24014f,-0.09306f,-0.03284f,0.1081f,-0.16145f,-0.24609f,-0.02592f,0.42805f,0.6649f,0.05751f,-0.16682f,-0.38868f,-0.09378f,-0.1125f,-0.13291f,0.41368f,-0.11337f,-0.52672f,-0.52193f,0.55066f,0.99124f,0.20445f,-0.2119f,-0.04827f,0.18974f,-0.22984f,-0.22858f,0.02941f,0.02675f,-0.14481f,0.23197f,0.80584f,0.20532f,-1.01576f,-0.52652f,0.1822f,0.12314f,0.08024f,-0.01466f,0.0105f,-0.25272f,-0.20111f,-0.09185f,0.22413f,0.44628f,-0.32769f,-0.46795f,0.4191f,-0.11976f,-0.52382f,0.07618f,0.71414f,0.68386f,-0.08184f,-0.21823f,-0.33073f,0.01896f,0.0606f,-0.33078f,-0.19506f,0.13901f,0.27298f,0.10626f,-0.00179f,-0.04595f,-0.23168f,-0.31826f,0.09567f,0.40333f,0.33412f,-0.16822f,-0.2724f,0.03424f,0.02389f,-0.12111f,-0.12803f,0.26771f,0.65557f,-0.06733f,-0.60265f,-0.01059f,0.28469f,0.31584f,0.18152f,0.16111f,-0.07482f,-0.50287f,-0.58481f,0.02428f,0.29867f,0.2206f,0.40246f,0.06235f,-0.26394f,-0.55671f,-0.23105f,0.21122f,0.21789f,0.04929f,-0.10002f,-0.19424f,-0.07855f,0.14109f,0.09427f,0.01175f,-0.33291f,-0.19245f,-0.03889f,0.49721f,0.51153f,0.00914f,0.00798f,-0.00474f,-0.17968f,-0.84323f,-0.10573f,0.61794f,0.10447f,-0.22645f,-0.27057f,0.07802f,0.42921f,0.25688f,0.41968f,0.05233f,-0.27618f,-0.32217f,-0.49615f,-0.11192f,-0.26674f,-0.14143f,0.37925f,0.50858f,0.04377f,-0.33983f,-0.39327f,0.08005f,0.20991f,-0.09049f,-0.33069f,-0.20977f,0.19506f,-0.11192f,-0.54708f,-0.17494f,0.19323f,0.24072f,0.44512f,0.01258f,-0.12532f,0.13272f,0.37233f,-0.15584f,-0.40009f,0.09364f,0.32517f,0.36846f,0.23608f,-0.08421f,-0.40101f,0.16474f,0.53175f,0.3554f,0.39114f,-0.05722f,-0.41378f,-0.0947f,-0.32111f,-0.23013f,-0.21775f,0.15463f,0.61329f,0.1689f,-0.58713f,-0.68676f,0.40638f,0.29446f,-0.24479f,-0.01229f,0.30413f,-0.31076f,-0.52498f,-0.46602f,0.10065f,0.43482f,-0.34723f,0.42379f,0.31835f,-0.55632f,-0.40662f,0.67511f,0.69421f,-0.11574f,-0.22771f,-0.01345f,0.15835f,-0.10418f,-0.03299f,-0.08774f,0.348f,0.28725f,-0.0207f,0.0754f,0.15342f,0.04247f,0.03328f,-0.03739f,-0.62442f,-0.4073f,-0.09993f,0.2354f,0.47627f,-0.01224f,-0.63704f,-0.21673f,0.56391f,0.277f,-0.11419f,-0.20779f,-0.52991f,-0.50447f,0.28546f,0.74577f,0.27013f,-0.13852f,-0.08967f,-0.16991f,-0.07773f,-0.28807f,-0.38041f,0.18568f,0.5731f,0.33069f,-0.43883f,-0.38428f,0.18283f,0.54166f,0.22602f,-0.30104f,-0.61765f,-0.43709f,0.07734f,0.45954f,0.55211f,0.17045f,-0.19124f,-0.63129f,0.09693f,0.47008f,0.05867f,-0.40914f,-0.31197f,0.17997f,0.17006f,-0.03734f,-0.18723f,0.04493f,-0.02201f,-0.03613f,0.12203f,0.32063f,-0.03091f,-0.05316f,0.34249f,0.30225f,-0.05137f,-0.32212f,-0.19879f,-0.09228f,0.00798f,0.32604f,0.4576f,0.38916f,-0.19739f,-0.40575f,0.0534f,0.28029f,0.12039f,-0.37025f,-0.16116f,-0.00329f,-0.20532f,-0.32266f,0.11328f,0.4325f,0.16343f,-0.43081f,-0.66698f,-0.28512f,-0.02467f,0.12542f,0.29073f,0.48048f,0.05088f,-0.67491f,-0.56643f,0.07545f,0.04972f,0.01886f,0.01732f,-0.01287f,0.36744f,0.35734f,0.28043f,0.44754f,0.10583f,-0.61131f,-0.54012f,0.16687f,0.19187f,0.10409f,0.02051f,0.09673f,0.23932f,-0.16904f,0.06844f,0.17722f,0.26017f,0.04266f,-0.53538f,-0.74945f,-0.51487f,0.06742f,0.18341f,0.4858f,0.48831f,-0.13673f,-0.67646f,-0.43221f,-0.19463f,0.33968f,0.66423f,0.0104f,-0.58263f,-0.53455f,0.01727f,0.0088f,0.1452f,0.32261f,0.02655f,-0.13093f,-0.07013f,0.19327f,0.34418f,-0.14708f,-0.2148f,0.12725f,0.0384f,-0.32962f,-0.12527f,0.75753f,0.35192f,-0.22297f,-0.12363f,-0.27298f,0.11497f,0.55854f,0.37949f,-0.01258f,-0.36416f,-0.45973f,0.05935f,0.52972f,0.49673f,0.17968f,-0.08537f,-0.34965f,-0.21132f,-0.09814f,-0.01891f,0.45189f,0.0475f,-0.13789f,-0.16677f,-0.03482f,-0.41852f,-0.4427f,0.70858f,0.345f,-0.29156f,-0.49402f,-0.30553f,0.19874f,0.24077f,0.00982f,-0.0044f,-0.03773f,-0.38713f,-0.49518f,-0.01761f,0.52498f,0.31516f,-0.15458f,-0.58776f,-0.1139f,0.45315f,0.32217f,0.09049f,0.046f,-0.03797f,-0.77958f,-0.21446f,0.70713f,0.6962f,0.11656f,-0.15743f,0.2072f,-0.25562f,-0.22147f,-0.15806f,-0.28899f,0.43032f,0.45426f,-0.34998f,-0.5376f,0.06771f,0.43506f,0.23008f,-0.00285f,0.19042f,-0.26955f,-0.79196f,-0.20546f,0.35888f,0.42626f,-0.03303f,0.09756f,0.09122f,-0.23647f,-0.0875f,-0.08005f,0.39187f,0.38036f,-0.10965f,-0.23298f,-0.4813f,-0.26394f,0.54795f,0.93537f,0.0784f,-0.74587f,-0.64773f,-0.34984f,0.52909f,1.09924f,0.41712f,-0.53141f,-0.89876f,-0.4147f,0.10902f,0.38157f,1.06877f,0.32416f,-0.83124f,-0.7554f,0.02118f,0.70616f,0.57397f,0.09277f,-0.5049f,-0.23071f,-0.32488f,0.04624f,0.57407f,0.48382f,0.02829f,-0.68193f,-0.13654f,0.11526f,0.15748f,0.10375f,0.23913f,0.30578f,-0.44009f,-0.80338f,-0.10762f,0.79878f,0.43327f,-0.31642f,-0.51395f,-0.44527f,-0.23724f,-0.09988f,0.1881f,0.32769f,0.10036f,0.1096f,-0.16938f,-0.35042f,-0.44962f,-0.41881f,0.31352f,0.83699f,0.20324f,-0.61112f,-0.56904f,0.21615f,0.74437f,0.26641f,-0.16672f,-0.42423f,-0.43429f,-0.1554f,0.35468f,0.53654f,0.55138f,0.15182f,-0.37528f,-0.15666f,0.47932f,0.46698f,-0.14766f,0.0579f,-0.45538f,-0.60062f,0.15942f,0.70679f,0.88047f,-0.02559f,-0.62055f,-0.48928f,-0.17253f,-0.1007f,0.27758f,0.36914f,-0.0637f,-0.34312f,-0.6545f,-0.23124f,0.20029f,0.22491f,0.1835f,0.283f,0.19797f,-0.57455f,-0.90263f,0.07381f,0.58945f,0.31081f,-0.10713f,-0.2473f,0.14172f,-0.35661f,-0.47516f,0.81344f,0.96995f,-0.26674f,-0.65126f,-0.47027f,-0.02525f,0.22887f,0.15758f,0.37368f,0.23918f,-0.39298f,-0.27153f,0.42273f,0.6752f,0.46021f,-0.21693f,-0.56522f,-0.70132f,-0.26152f,0.56875f,0.89644f,0.41765f,-0.30292f,-0.76609f,-0.68023f,0.02365f,0.54258f,0.67511f,0.1865f,-0.49726f,-0.7747f,-0.10085f,0.39535f,0.33272f,0.03463f,-0.29141f,-0.4206f,-0.52894f,-0.10056f,0.2223f,0.39695f,0.28948f,0.14022f,-0.17436f,-0.05306f,0.09035f,-0.03487f,0.08571f,0.13625f,0.01601f,-0.61504f,-0.22094f,0.39709f,0.74355f,0.46408f,-0.45189f,-0.87274f,-0.40082f,0.19473f,0.38428f,0.49359f,-0.08135f,-0.48464f,-0.33127f,-0.03144f,-0.04938f,0.28561f,0.77837f,0.54263f,-0.19734f,-0.73895f,-0.61934f,0.01282f,0.50321f,0.49929f,0.08503f,-0.70926f,-0.58074f,0.01674f,0.55109f,0.63153f,-0.06442f,-0.30176f,-0.02655f,-0.0386f,-0.25315f,0.08198f,0.46945f,0.15473f,-0.25267f,-0.10413f,0.06046f,0.19395f,-0.11357f,-0.11821f,0.0932f,0.05296f,-0.26747f,-0.58911f,0.18365f,0.36261f,0.2843f,0.0918f,-0.3525f,-0.32474f,-0.25838f,-0.01345f,0.43187f,-0.01775f,-0.43521f,-0.36609f,-0.46065f,0.06931f,0.56546f,0.33431f,0.12991f,-0.03091f,-0.72957f,-0.18457f,0.56807f,0.5849f,0.35501f,0.06539f,-0.08237f,-0.36058f,0.02573f,0.35801f,0.38989f,0.15115f,-0.29886f,-0.68183f,-0.49915f,0.37586f,0.52188f,0.02878f,-0.1718f,-0.33335f,-0.39782f,-0.44909f,-0.04803f,0.55255f,0.49663f,-0.15927f,-0.56478f,-0.00953f,-0.02907f,-0.12788f,0.3585f,0.85948f,-0.15153f,-1.3387f,-0.38539f,0.40783f,0.64333f,0.51177f,0.2297f,0.04039f,-0.30718f,-0.50224f,-0.14679f,0.43825f,0.35211f,-0.24271f,-0.0754f,0.09906f,-0.35579f,-0.17412f,0.17659f,0.6382f,0.24048f,-0.60106f,-0.49083f,0.26404f,0.48149f,-0.03183f,-0.06573f,-0.24392f,-0.13548f,-0.14321f,-0.03482f,0.64164f,0.39806f,-0.08329f,-0.14786f,-0.031f,0.00058f,-0.01553f,0.21165f,-0.13581f,-0.48614f,-0.27294f,-0.18355f,0.25639f,0.55598f,0.09804f,-0.20242f,0.25044f,-0.35144f,-0.84604f,-0.13572f,0.59922f,0.61441f,-0.17968f,-0.33988f,-0.36159f,0.09224f,0.21446f,0.00363f,0.24866f,-0.01649f,-0.61194f,-0.62669f,0.18462f,0.75695f,0.40696f,-0.0312f,-0.29427f,-0.46287f,-0.58302f,-0.4324f,0.43545f,0.79046f,0.77136f,-0.17528f,-0.61649f,0.01277f,0.25311f,-0.04174f,0.16276f,0.7197f,-0.32561f,-0.67356f,-0.00967f,0.76169f,0.80014f,0.0414f,-0.45789f,-0.13707f,0.08663f,-0.16604f,-0.07313f,0.17185f,0.17069f,-0.36217f,-0.08406f,0.03163f,0.11879f,-0.06302f,-0.2458f,-0.10631f,-0.46021f,-0.22863f,-0.17862f,0.08546f,0.60875f,0.07037f,-0.57596f,-0.48425f,-0.18655f,0.21315f,0.48348f,0.50253f,0.21615f,-0.48149f,-0.24793f,0.08068f,-0.15052f,0.06128f,0.30312f,0.57025f,0.02118f,-0.87008f,-0.52391f,0.7421f,1.19133f,0.32899f,-0.65286f,-0.68638f,-0.2429f,-0.10742f,0.09311f,0.81266f,0.77455f,-0.42403f,-0.44503f,0.20159f,-0.10355f,-0.37862f,-0.10017f,-0.11864f,-0.26655f,-0.2783f,-0.02757f,0.2948f,0.34568f,0.09543f,-0.21059f,-0.459f,-0.15874f,-0.08909f,-0.11386f,0.17224f,0.16464f,0.35569f,-0.32953f,-0.4576f,0.07898f,0.42249f,0.37257f,-0.36522f,-0.13964f,0.12469f,-0.01016f,-0.05291f,0.28996f,0.74568f,0.05751f,-0.47569f,-0.12803f,0.11908f,-0.0608f,0.17185f,0.382f,0.32488f,0.06965f,-0.47743f,-0.23226f,0.27356f,0.4961f,-0.24643f,-0.33693f,-0.17132f,-0.21635f,0.34079f,0.67511f,0.6235f,-0.30176f,-0.64401f,-0.43521f,0.28324f,0.58602f,0.13611f,-0.08721f,-0.32977f,-0.42742f,-0.68976f,-0.06172f,0.86558f,0.41039f,-0.07057f,-0.06534f,-0.60556f,-0.59424f,0.26495f,0.66718f,0.72299f,-0.0741f,-0.79941f,-0.52498f,-0.13112f,-0.03855f,0.0964f,0.68594f,0.26762f,-0.34123f,-0.42399f,-0.21717f,0.00982f,-0.08735f,0.05349f,-0.20077f,-0.19366f,0.21746f,0.30466f,0.10147f,0.01277f,0.30558f,-0.03773f,-0.50336f,-0.21538f,0.18733f,0.16585f,0.0651f,0.11429f,0.37949f,0.29838f,-0.04401f,0.03076f,0.45035f,0.48614f,-0.57929f,-0.6073f,0.05809f,0.52304f,0.44396f,0.05683f,0.02549f,-0.46732f,-0.74568f,-0.45857f,0.02907f,0.37697f,0.1999f,-0.20275f,-0.30718f,-0.32348f,-0.13731f,-0.18147f,0.25804f,0.45068f,-0.29064f,-0.63535f,-0.54829f,0.07216f,0.6295f,0.40072f,0.14515f,0.06404f,-0.14496f,-0.44474f,-0.33214f,0.668f,0.63264f,-0.07086f,0.30553f,0.29354f,-0.09973f,-0.02002f,0.58732f,0.80497f,-0.53779f,-0.8221f,0.09543f,0.43975f,0.17359f,-0.21896f,0.15584f,0.41881f,-0.33755f,-0.82031f,-0.17016f,0.31337f,0.17369f,-0.13388f,-0.06544f,0.01006f,-0.53504f,-0.49842f,0.31419f,0.39279f,0.10191f,-0.70771f,-0.68686f,0.14791f,0.4146f,0.37634f,-0.06162f,-0.11507f,-0.60159f,-0.58635f,0.18196f,0.52614f,0.64082f,0.20614f,-0.60178f,-0.11308f,0.59777f,0.43574f,0.35874f,-0.23753f,-0.4486f,-0.32353f,-0.22379f,0.07013f,0.49006f,0.50766f,0.0917f,-0.17915f,-0.43023f,-0.49006f,-0.28614f,0.30036f,0.55279f,0.27105f,-0.39946f,-0.70853f,0.0903f,0.55603f,0.26157f,-0.1095f,-0.48503f,-0.3495f,0.01533f,0.11284f,0.41015f,0.33349f,-0.15845f,0.00977f,0.14525f,0.0117f,-0.51627f,-0.27792f,0.69184f,0.42108f,-0.2267f,-0.34123f,0.13935f,0.13465f,-0.06375f,0.32314f,0.20667f,-0.65963f,-0.64251f,0.26694f,0.55192f,0.0267f,-0.228f,0.25659f,0.14873f,-0.2488f,-0.37528f,-0.04014f,0.41306f,-0.0414f,-0.60502f,-0.08575f,0.45127f,-0.01083f,-0.07173f,-0.09785f,-0.04063f,0.02157f,-0.08962f,0.14805f,0.03574f,0.52585f,0.53804f,-0.24058f,-0.3319f,-0.20401f,0.19898f,0.21194f,0.17601f,0.51003f,0.07632f,-0.47854f,-0.30292f,-0.01103f,0.23274f,0.53436f,0.02249f,-0.13577f,-0.5733f,-0.8132f,-0.16846f,0.74031f,1.29851f,-0.16193f,-1.16802f,-0.99003f,-0.0947f,0.50998f,0.26297f,0.1023f,0.26994f,-0.18404f,-0.7895f,-0.41905f,0.56797f,0.77373f,-0.18084f,-0.30476f,-0.25373f,-0.18031f,0.07657f,0.37257f,0.32556f,0.20087f,-0.06094f,-0.45218f,0.12561f,0.61852f,0.31569f,-0.18965f,-0.16367f,-0.07952f,0.03739f,0.50548f,0.32382f,0.0444f,-0.11966f,-0.31753f,-0.07124f,0.09209f,0.19347f,0.12203f,-0.00019f,-0.3465f,-0.51598f,-0.16774f,0.11971f,0.07961f,-0.15463f,0.14795f,0.19066f,-0.31908f,-0.63898f,-0.1512f,0.56285f,0.56696f,0.08629f,-0.7005f,-0.49417f,0.00392f,0.12068f,0.20764f,0.21354f,0.31342f,-0.22142f,-0.39753f,0.26858f,0.51105f,0.25514f,0.17577f,-0.10123f,-0.21586f,-0.31308f,-0.27661f,0.41185f,0.76183f,0.17296f,-0.57678f,-0.33712f,0.02317f,0.09441f,0.08609f,0.27932f,0.07584f,-0.14409f,-0.01833f,-0.5271f,-0.28561f,0.48943f,0.51777f,0.11071f,0.05054f,-0.03671f,-0.61305f,-0.26123f,0.80454f,0.79177f,-0.16735f,-0.73716f,-0.37049f,0.13896f,0.22723f,0.07081f,0.21354f,0.25451f,-0.36372f,-0.44198f,0.29136f,0.39738f,-0.23695f,-0.32696f,0.03681f,-0.06597f,-0.3139f,0.01973f,0.67956f,0.31937f,-0.68585f,-0.56454f,-0.00247f,0.35047f,0.08082f,-0.11004f,0.15197f,-0.00068f,-0.33049f,-0.27845f,0.28614f,0.71312f,0.22631f,-0.60309f,-0.27158f,-0.0993f,0.00198f,0.43293f,0.58587f,0.46229f,-0.26157f,-0.72952f,-0.21267f,0.29639f,0.1644f,0.26384f,0.13204f,0.21886f,0.01001f,-0.41804f,-0.21204f,0.23042f,0.89677f,0.16875f,-0.75293f,-0.67003f,-0.32711f,0.17436f,0.74964f,0.65615f,-0.39695f,-0.92671f,-0.25315f,0.35322f,0.25664f,0.3081f,0.06742f,-0.07555f,-0.16904f,-0.29209f,-0.13877f,0.15405f,0.48309f,0.27545f,-0.50466f,-0.64536f,0.07013f,0.6473f,0.39603f,-0.19932f,-0.17577f,-0.30936f,-0.01461f,0.24159f,0.07724f,-0.08764f,-0.10147f,-0.18626f,-0.20396f,-0.08256f,0.20899f,0.5155f,-0.16991f,-0.33383f,-0.22075f,0.13175f,0.54079f,0.26984f,0.20387f,0.0726f,-0.13335f,-0.3272f,-0.12488f,0.46379f,0.43003f,-0.24895f,-0.09596f,0.36928f,0.02559f,-0.24479f,-0.51139f,-0.03811f,0.62277f,0.31942f,-0.67371f,-0.9774f,0.0725f,0.18268f,-0.1378f,0.38079f,0.52691f,-0.28469f,-0.72942f,-0.64715f,-0.0341f,0.6575f,0.0489f,-0.20619f,0.05451f,-0.03066f,-0.22592f,0.04092f,0.60425f,0.48599f,-0.15066f,-0.42394f,-0.071f,0.19149f,0.61392f,0.34558f,-0.305f,-0.09577f,-0.05156f,-0.16537f,0.06143f,0.44633f,0.62577f,0.29146f,-0.48933f,-0.69489f,0.07748f,0.50679f,0.25586f,-0.13959f,-0.31066f,-0.36595f,-0.2443f,-0.09968f,0.42698f,0.50621f,-0.46718f,-0.44938f,-0.29552f,-0.27371f,-0.083f,0.00063f,0.34882f,0.04962f,-0.45218f,-0.76338f,-0.14022f,0.82892f,0.67608f,0.05374f,-0.37325f,-0.52304f,-0.02254f,0.58814f,0.37267f,-0.07163f,-0.39971f,-0.00856f,-0.08556f,0.0117f,0.45388f,0.37122f,0.3036f,0.034f,-0.10747f,-0.14529f,0.19487f,0.12653f,-0.01712f,-0.14075f,-0.02312f,-0.30321f,-0.40445f,0.22858f,0.21746f,0.20483f,0.012f,-0.22036f,-0.20266f,-0.14017f,-0.14844f,0.15018f,0.34046f,0.06428f,-0.36464f,-0.42771f,0.12421f,0.37867f,0.00716f,-0.12566f,0.00711f,0.07289f,-0.0267f,-0.27632f,0.07187f,0.24251f,0.14138f,-0.15371f,-0.27318f,0.01016f,-0.01504f,0.13112f,0.24933f,0.108f,-0.32716f,-0.34645f,-0.09746f,0.2443f,0.36643f,-0.34055f,-0.30167f,0.22113f,0.30786f,0.00015f,-0.47598f,0.03661f,0.42171f,0.03603f,-0.21054f,-0.41968f,-0.10447f,0.22191f,0.21794f,0.02704f,0.01499f,-0.11376f,-0.29934f,0.03202f,0.19531f,0.22757f,-0.15739f,-0.19671f,0.07434f,0.03913f,-0.05146f,-0.12774f,0.22428f,0.34887f,-0.18781f,-0.48556f,-0.15255f,0.24155f,0.30974f,-0.04073f,-0.15671f,-0.15845f,-0.23376f,-0.13702f,0.15386f,0.19003f,0.28672f,-0.06143f,-0.50152f,-0.06829f,-0.10883f,-0.08063f,0.0327f,0.22496f,0.01886f,-0.17156f,-0.04392f,0.16895f,0.35753f,0.42611f,0.30263f,-0.16343f,0.05417f,-0.2488f,-0.30742f,0.16256f,0.55274f,0.3702f,-0.37011f,-0.49663f,-0.19163f,0.35966f,0.88507f,0.43782f,-0.60672f,-0.65794f,-0.21905f,0.28764f,0.19927f,-0.08232f,-0.14321f,-0.33959f,-0.29257f,-0.60652f,0.01248f,0.65296f,0.09997f,-0.17799f,-0.32416f,-0.35129f,-0.25146f,0.04179f,0.43535f,0.3969f,-0.35719f,-0.74495f,-0.08218f,0.53252f,0.58703f,0.2532f,-0.08938f,-0.14205f,-0.00406f,-0.26645f,-0.06341f,0.62979f,0.39738f,0.06568f,-0.26907f,-0.39119f,0.1866f,0.6473f,0.36164f,-0.18317f,-0.14201f,0.04885f,-0.05054f,-0.21964f,0.45813f,0.44169f,-0.44691f,-0.71569f,-0.61499f,0.17373f,0.594f,0.05993f,-0.06926f,0.31835f,-0.4886f,-1.04192f,-0.13698f,0.58959f,0.38408f,-0.40304f,-0.39811f,-0.12929f,0.24454f,0.23173f,-0.32778f,0.08459f,0.57538f,0.17935f,-0.30945f,-0.23922f,-0.03507f,0.16841f,0.20329f,0.43134f,0.42805f,-0.12711f,-0.25011f,-0.37151f,0.22834f,0.47066f,-0.00489f,-0.12856f,-0.23913f,-0.11521f,-0.11536f,-0.01291f,0.17315f,0.19052f,-0.24019f,-0.23003f,-0.21668f,0.03236f,0.64962f,0.42268f,-0.32691f,-0.7258f,-0.10099f,0.47893f,0.34457f,-0.01349f,-0.23632f,-0.2652f,-0.09016f,-0.05214f,0.23245f,0.27419f,0.32449f,0.1629f,-0.69605f,-0.81823f,0.1749f,0.85416f,0.47598f,-0.0964f,-0.51492f,-0.78389f,-0.42752f,0.72788f,0.63501f,-0.05248f,-0.34979f,-0.28338f,0.22399f,0.12396f,0.23081f,0.19647f,0.24222f,0.14365f,-0.57344f,-0.60454f,0.14955f,0.59206f,0.53199f,0.23927f,-0.126f,-0.08962f,-0.54544f,-0.28585f,0.61412f,0.56619f,-0.02805f,-0.38312f,-0.17862f,0.13407f,0.03337f,-0.10689f,0.29436f,0.22225f,-0.2709f,-0.76899f,-0.25165f,0.6887f,0.38278f,-0.23526f,-0.2309f,0.05741f,-0.12991f,-0.39337f,0.08803f,0.49088f,-0.09465f,-0.54863f,-0.02138f,0.32014f,0.09108f,0.14525f,0.23797f,-0.07623f,-0.63869f,-0.57252f,0.34853f,0.98045f,0.31332f,-0.62224f,-0.33233f,0.08474f,0.01881f,0.31279f,0.60638f,0.15057f,-0.57126f,-0.24464f,-0.02709f,-0.12445f,0.26491f,0.16958f,0.08029f,0.07976f,-0.10926f,-0.55859f,-0.29199f,0.57881f,0.69068f,0.03279f,-0.38176f,-0.22791f,-0.2059f,0.06394f,0.20875f,0.38026f,0.49852f,-0.09606f,-0.44933f,0.11178f,0.305f,0.18384f,0.19061f,0.06346f,-0.17557f,-0.74911f,-0.34578f,0.39685f,0.63772f,0.27502f,-0.50824f,-0.54602f,-0.0429f,0.26849f,-0.05611f,-0.07415f,0.15913f,0.00077f,-0.36575f,-0.51269f,-0.13664f,0.06423f,0.18143f,0.14868f,0.03405f,-0.22254f,-0.26645f,-0.00585f,0.13306f,0.50718f,0.29238f,-0.08542f,-0.08116f,-0.24377f,0.02796f,0.54587f,0.47216f,-0.26176f,-0.33775f,0.1423f,0.23255f,0.40827f,0.23178f,-0.05765f,-0.25731f,-0.23898f,-0.36261f,0.29499f,0.88787f,0.04866f,-0.71985f,-0.58103f,0.16048f,0.19013f,0.12648f,0.43579f,0.33862f,-0.60894f,-1.1745f,-0.865f,0.7273f,1.23771f,0.21944f,-0.51879f,-0.84807f,-0.25277f,-0.12958f,0.33156f,0.45958f,0.06588f,-0.117f,-0.46998f,-0.27303f,0.39714f,0.37997f,0.11173f,0.29393f,-0.0221f,-0.28479f,-0.35971f,0.11342f,0.71627f,0.19434f,0.09388f,0.04871f,-0.08827f,0.05606f,-0.1273f,-0.03985f,-0.12184f,0.26321f,0.31884f,-0.47888f,-0.41383f,0.29625f,0.25901f,-0.28585f,-0.34437f,-0.08068f,0.31187f,0.00305f,-0.22259f,0.34703f,0.53509f,-0.18974f,-0.84018f,-0.22046f,0.40179f,0.10012f,-0.03511f,0.24406f,-0.12251f,-0.37634f,0.05848f,0.70988f,0.28184f,-0.59695f,-0.22399f,0.11284f,0.12455f,-0.14152f,0.04488f,0.48217f,0.23526f,0.0711f,-0.13654f,-0.49847f,-0.323f,0.08285f,0.44406f,0.45562f,-0.38868f,-0.69368f,-0.02631f,0.46128f,0.50979f,-0.01069f,-0.43951f,-0.48594f,-0.31622f,-0.0533f,0.29736f,0.59158f,0.27966f,-0.0976f,-0.29576f,-0.36899f,-0.44643f,0.10607f,0.42181f,0.53538f,0.38046f,-0.55085f,-0.2592f,0.53131f,0.60836f,0.40735f,-0.19071f,-0.76009f,-0.29412f,-0.1053f,0.14181f,0.46408f,0.32406f,-0.22476f,-0.91003f,-0.2637f,0.28614f,0.44764f,-0.06718f,-0.68589f,-0.21968f,0.03937f,-0.15714f,0.00174f,0.29891f,0.16145f,-0.57233f,-0.73242f,0.13282f,0.40217f,-0.11831f,-0.08324f,0.37692f,0.35521f,-0.35376f,-0.95612f,0.21814f,0.893f,0.21596f,-0.11724f,0.05461f,0.42757f,0.03487f,0.18394f,0.87622f,0.14418f,-0.51714f,-0.2858f,-0.02883f,0.11516f,0.22766f,0.58568f,0.43888f,-0.30379f,-0.64043f,-0.54002f,0.10413f,0.5081f,-0.05166f,-0.14887f,-0.22631f,-0.34041f,-0.27424f,-0.03129f,0.36512f,0.24527f,-0.21891f,-0.6502f,-0.68996f,-0.17224f,0.05741f,0.19564f,0.59404f,0.31008f,-0.25098f,-0.79801f,-0.02138f,0.62858f,0.20072f,-0.0533f,0.09199f,0.42355f,0.23492f,0.06501f,0.09997f,0.22002f,0.02829f,-0.43487f,-0.34979f,0.23922f,0.17025f,-0.00851f,0.23197f,0.30486f,0.02791f,-0.29567f,-0.02152f,-0.01538f,-0.05296f,-0.04455f,-0.37131f,-0.41088f,-0.12285f,0.62698f,0.42239f,-0.0519f,-0.03337f,-0.30839f,-0.73286f,-0.57112f,0.44029f,0.75216f,0.53097f,-0.16546f,-0.76826f,-0.28459f,-0.04126f,0.06157f,0.77968f,0.5271f,-0.54374f,-0.57339f,-0.36981f,-0.22273f,0.01451f,0.35796f,0.58911f,-0.13833f,-0.57518f,-0.5983f,0.19879f,0.76028f,0.53122f,0.05335f,-0.27782f,-0.15845f,-0.45542f,0.18926f,0.48638f,0.02075f,-0.44488f,-0.25606f,0.27642f,0.34263f,0.34413f,-0.01601f,0.03487f,0.06481f,-0.35826f,-0.34089f,0.51806f,0.61562f,0.03226f,-0.04198f,0.05248f,0.08348f,-0.38205f,-0.02559f,0.52812f,0.24527f,-0.16024f,-0.5745f,0.10312f,0.30312f,0.14675f,0.07047f,-0.18026f,-0.0327f,-0.36798f,-0.37422f,-0.0532f,0.24106f,0.36609f,0.15158f,-0.18737f,-0.29557f,-0.20522f,-0.05756f,-0.2489f,-0.33625f,0.15221f,-0.04962f,-0.65383f,0.17557f,1.00783f,0.41925f,-0.06462f,-0.13843f,-0.44933f,-0.79869f,-0.23816f,0.86814f,1.15602f,0.09722f,-0.88676f,-0.50476f,0.43787f,0.46553f,0.1822f,-0.08343f,-0.05615f,0.03057f,-0.5095f,-0.15971f,0.43908f,0.50519f,0.39284f,-0.00111f,-0.33901f,-0.59569f,-0.3405f,0.46365f,0.4279f,-0.15221f,-0.1407f,-0.40832f,-0.29543f,-0.11163f,-0.25359f,0.11806f,0.03613f,0.04711f,-0.30921f,-0.1302f,0.27448f,0.02815f,-0.14916f,-0.06118f,0.08174f,-0.38979f,-0.28938f,0.44527f,0.93392f,0.39095f,-0.37809f,-0.71385f,0.31531f,0.59129f,-0.10423f,0.15202f,0.20571f,-0.31424f,-0.60498f,0.40517f,0.84323f,0.69097f,-0.00696f,-0.61924f,-0.50186f,-0.40865f,-0.16005f,0.14583f,0.66524f,0.58219f,-0.27908f,-0.73968f,-0.33011f,0.20648f,0.1274f,0.04527f,0.27763f,-0.20029f,-0.35922f,-0.33456f,0.2045f,0.76555f,0.26167f,-0.55293f,-0.78451f,0.11816f,0.0341f,-0.17562f,0.40416f,0.49044f,-0.01412f,-0.23216f,-0.38752f,-0.40565f,0.13282f,0.10834f,-0.04184f,0.01659f,-0.10186f,-0.15037f,0.10283f,0.64812f,0.33905f,-0.20493f,-0.06762f,-0.25146f,-0.1287f,-0.08788f,0.20909f,0.52449f,-0.1407f,-0.6695f,-0.25494f,0.60391f,0.42152f,-0.00169f,-0.33983f,-0.15047f,0.1333f,-0.28106f,0.03545f,0.5108f,0.1791f,-0.40159f,-0.45276f,-0.12445f,0.15095f,0.34495f,0.64038f,0.54829f,0.13838f,-1.0263f,-0.98543f,0.59743f,0.68599f,0.06486f,-0.15951f,0.09291f,-0.32125f,-0.49784f,0.22674f,0.61092f,0.30631f,-0.33712f,-0.43211f,-0.37876f,-0.25064f,-0.10201f,0.26713f,0.59932f,-0.06858f,-0.71327f,-0.51085f,-0.07299f,0.31903f,0.39134f,0.2519f,-0.23922f,-0.26423f,0.02302f,0.16958f,0.51308f,0.61151f,0.3111f,-0.6457f,-0.41596f,0.13833f,0.16493f,0.52164f,0.39816f,-0.27381f,-0.44522f,-0.37809f,-0.33789f,0.32551f,0.25843f,-0.19743f,-0.27666f,-0.38278f,-0.1688f,0.26268f,0.28188f,0.40246f,-0.25809f,-0.55419f,-0.02675f,0.07366f,0.03429f,-0.21074f,0.08585f,0.15308f,-0.01475f,-0.2162f,-0.11531f,0.39734f,0.33818f,-0.15753f,-0.18017f,0.03366f,-0.17485f,-0.28769f,0.30757f,0.71685f,0.28121f,-0.26578f,-0.78954f,-0.34495f,0.37383f,0.53954f,0.32836f,-0.34167f,-0.76048f,-0.54084f,0.19908f,0.75496f,0.59371f,0.03158f,-0.01828f,-0.15748f,-0.45876f,-0.36474f,0.26829f,0.49847f,0.29088f,-0.03342f,-0.7377f,-0.76459f,0.00271f,0.73919f,0.58626f,0.14776f,-0.56362f,-0.3318f,0.16537f,0.0963f,0.03371f,-0.00987f,0.09804f,-0.20213f,-0.00721f,0.21678f,0.37383f,0.32885f,0.09582f,-0.35356f,-0.5272f,-0.34123f,-0.17165f,0.22529f,0.38844f,0.25209f,-0.52386f,-0.74074f,-0.16527f,0.49615f,0.67714f,0.32479f,-0.34819f,-0.83032f,-0.56846f,0.03255f,0.74568f,0.82829f,0.17238f,-0.5849f,-0.65721f,-0.44459f,0.12817f,0.64488f,0.60469f,0.09944f,-0.48096f,-0.08251f,0.40159f,0.2473f,0.17432f,0.1792f,-0.28242f,-0.58094f,-0.60115f,0.35535f,1.17585f,0.11434f,-0.62621f,0.01175f,-0.02351f,-0.28092f,0.12155f,0.47792f,0.48827f,-0.49978f,-0.91119f,-0.12924f,0.31753f,0.21161f,0.19971f,0.23589f,0.05214f,-0.53054f,-0.54573f,-0.19139f,0.13935f,0.44029f,0.03405f,-0.12421f,-0.31434f,-0.61131f,-0.44135f,0.50645f,0.91767f,0.44154f,-0.21586f,-0.72048f,-0.07231f,-0.08068f,0.09141f,0.63641f,0.1512f,-0.35202f,-0.18234f,0.43724f,0.54945f,0.50031f,0.1007f,-0.33315f,-0.46974f,-0.30911f,0.00798f,0.57252f,0.61349f,-0.23956f,-0.70127f,-0.28445f,0.64875f,0.69325f,-0.11627f,-0.51051f,-0.37436f,-0.30849f,0.01316f,0.09833f,0.20319f,0.3302f,0.01243f,-0.1615f,-0.34491f,-0.50258f,-0.46998f,0.00174f,0.77605f,0.44367f,-0.58157f,-0.74007f,-0.02384f,0.38002f,0.10089f,0.16542f,0.11086f,0.09074f,-0.50703f,-0.43961f,0.36628f,0.53973f,0.42147f,-0.28338f,-0.20595f,-0.42558f,-0.28348f,0.39153f,0.69997f,0.46694f,-0.43347f,-0.91259f,-0.39168f,0.9329f,0.76352f,0.18868f,-0.07574f,-0.74761f,-0.81489f,-0.0977f,0.87728f,0.96357f,0.01877f,-0.53494f,-0.75395f,-0.46292f,0.63598f,0.87593f,0.11676f,-0.47733f,-0.35245f,-0.18254f,0.09751f,0.52299f,0.10945f,-0.09794f,-0.16638f,-0.13509f,0.20701f,-0.06384f,0.20058f,0.47704f,0.12817f,-0.5791f,-0.66374f,0.09059f,0.08532f,0.06002f,0.38621f,0.32904f,-0.27197f,-0.57973f,0.06051f,0.61257f,0.03115f,-0.32063f,-0.08556f,-0.36232f,0.04087f,0.07265f,-0.1615f,0.46394f,0.78597f,0.36212f,-0.71008f,-0.43371f,0.10544f,0.01088f,0.21214f,0.42471f,0.27318f,-0.48125f,-0.42205f,0.09219f,0.49557f,0.32672f,-0.3304f,-0.50234f,-0.20295f,0.02636f,-0.03632f,0.12508f,0.28435f,-0.02723f,-0.48415f,-0.40841f,-0.06341f,0.72938f,0.18317f,-0.66123f,-0.10592f,-0.12701f,-0.13664f,-0.12484f,0.32948f,0.27182f,-0.14568f,-0.08575f,-0.34902f,-0.22689f,0.03811f,0.34853f,0.56918f,0.35419f,-0.25731f,-0.48058f,-0.19444f,0.18181f,0.53097f,0.40696f,-0.16503f,-0.20377f,-0.16406f,-0.06515f,0.74181f,0.36425f,-0.37809f,-0.13838f,0.13814f,-0.01548f,-0.29576f,0.13843f,0.34341f,0.03052f,-0.11739f,-0.09228f,0.1985f,0.29896f,0.0089f,0.07511f,-0.23637f,-0.67051f,-0.33572f,0.39545f,0.64004f,-0.00252f,-0.28691f,-0.62089f,-0.48933f,0.22699f,0.57905f,0.30379f,0.1526f,-0.18708f,-0.72183f,-0.67666f,-0.2798f,0.25359f,0.18534f,0.18505f,-0.12469f,0.02036f,0.25388f,-0.01364f,0.09262f,0.11086f,0.26887f,-0.16063f,-0.72705f,-0.27405f,0.36981f,0.3777f,0.27028f,0.431f,0.17112f,-0.05388f,0.06099f,-0.01557f,0.14007f,-0.20972f,-0.76251f,-0.35197f,0.36319f,0.41939f,0.19013f,0.05872f,-0.129f,-0.49712f,-0.6545f,-0.1139f,0.35226f,0.3541f,-0.16633f,-0.31032f,-0.08764f,-0.01451f,-0.34524f,-0.16348f,0.28072f,0.19763f,-0.1776f,-0.43671f,0.15037f,0.46394f,0.42123f,0.16498f,-0.07386f,-0.11961f,-0.29369f,-0.23265f,0.14718f,0.37992f,0.17069f,-0.15318f,0.19855f,0.25983f,0.0194f,-0.05698f,0.04576f,-0.06975f,-0.17078f,0.09272f,0.21465f,-0.01751f,-0.48464f,-0.1808f,0.28479f,0.44764f,-0.38868f,-0.49151f,0.02525f,0.36096f,0.59245f,0.26902f,-0.16222f,-0.61866f,-0.48507f,-0.06496f,0.47153f,0.41857f,-0.10114f,-0.54534f,-0.07618f,0.56309f,0.43081f,-0.1244f,-0.43511f,-0.11642f,-0.14075f,-0.12455f,-0.12063f,-0.01524f,0.58592f,0.63709f,-0.15966f,-0.62316f,-0.34795f,-0.15265f,0.01872f,0.53155f,0.26694f,-0.58766f,-0.38732f,0.31347f,0.43168f,0.02796f,-0.21146f,-0.61605f,-0.36145f,0.09417f,-0.01973f,0.13344f,0.11918f,0.01775f,0.06612f,0.44503f,0.08247f,-0.07294f,-0.08039f,0.01156f,0.14815f,-0.16421f,-0.23057f,0.22786f,0.92434f,0.10684f,-0.63748f,-0.19192f,0.62374f,-0.09393f,-0.04218f,0.50176f,-0.32324f,-0.4665f,-0.539f,0.07308f,0.23076f,-0.21277f,-0.0564f,0.551f,-0.13804f,-0.84139f,-0.18897f,0.45958f,0.49746f,-0.14592f,-0.17838f,-0.34041f,-0.16513f,-0.23681f,-0.36009f,0.25959f,0.36798f,0.43971f,0.46065f,-0.20469f,-0.31995f,0.12087f,0.31134f,0.37494f,0.24101f,-0.15429f,-0.42573f,0.20188f,0.30109f,-0.01703f,-0.14167f,0.27874f,0.39027f,0.18225f,0.09436f,-0.45562f,-0.397f,0.1349f,0.03681f,-0.35526f,-0.21668f,0.05707f,0.24735f,-0.19743f,-0.42379f,-0.18665f,0.31506f,0.52827f,0.12677f,-0.30171f,-0.12837f,-0.29954f,-0.48362f,-0.24087f,-0.21712f,0.34481f,0.31182f,0.34829f,0.32933f,-0.15874f,-0.27429f,-0.11932f,0.41871f,0.44536f,-0.67501f,-0.56372f,-0.00073f,-0.01378f,-0.00527f,0.19869f,0.6664f,0.1734f,-0.11023f,-0.31511f,-0.17857f,0.09257f,0.11681f,0.07265f,-0.19003f,-0.14457f,-0.04401f,0.15008f,0.2459f,0.05132f,-0.48909f,-0.43535f,-0.11507f,0.27685f,0.3466f,0.20875f,0.27922f,-0.33852f,-0.62205f,-0.30123f,0.09833f,0.15652f,0.25712f,0.31719f,0.20329f,-0.35071f,-0.50935f,0.42713f,0.54756f,0.01238f,-0.15903f,-0.05132f,0.06975f,-0.10496f,-0.32053f,0.10457f,0.42829f,0.29146f,-0.14442f,-0.33702f,-0.11182f,-0.02172f,0.03623f,0.20217f,0.25509f,-0.02752f,-0.32029f,-0.17422f,0.16169f,-0.2221f,-0.2769f,-0.2073f,0.27231f,0.85895f,0.40672f,-0.25988f,-0.50326f,0.33378f,0.03095f,-0.62365f,0.1452f,0.43627f,0.0119f,-0.13354f,-0.27526f,-0.11627f,0.02501f,0.00624f,0.1186f,0.16527f,-0.03231f,-0.9924f,-0.5079f,0.70582f,0.5819f,-0.04817f,-0.07947f,0.07081f,-0.08189f,-0.31564f,0.00513f,0.1525f,-0.20837f,-0.07255f,0.61721f,1.00845f,-0.09625f,-0.60904f,-0.38229f,0.24169f,0.2193f,-0.26544f,0.02206f,-0.32029f,-0.27739f,-0.10549f,0.38157f,0.67293f,0.15062f,-0.14346f,-0.27086f,0.00996f,0.07608f,-0.11874f,0.06409f,0.24367f,-0.29044f,-0.38989f,-0.16324f,0.45455f,0.43405f,-0.29127f,-0.16029f,0.05698f,0.13233f,-0.02457f,0.19637f,0.03386f,-0.4251f,-0.49939f,-0.11303f,0.40638f,0.31168f,-0.24996f,0.0014f,0.18791f,0.14563f,0.25867f,-0.11303f,-0.11681f,-0.02496f,-0.16222f,-0.34703f,0.01935f,0.55733f,0.49629f,-0.14259f,-0.22965f,0.23884f,-0.04474f,-0.28406f,-0.00058f,0.03299f,0.07308f,-0.19922f,-0.15236f,0.04068f,-0.29891f,-0.10486f,-0.41451f,-0.10868f,0.54519f,-0.4961f,-0.36396f,0.38312f,0.1126f,-0.13451f,-0.10704f,0.02776f,-0.01746f,-0.40764f,-0.39322f,0.17306f,0.50882f,0.60735f,0.07734f,-0.03932f,0.04387f,-0.17112f,-0.10099f,0.15453f,0.20653f,-0.21707f,-0.64439f,0.0653f,0.66273f,0.22254f,-0.04251f,0.02215f,0.24943f,-0.19327f,-0.4784f,0.21504f,0.36875f,-0.58031f,-0.30626f,0.55187f,0.40183f,-0.24043f,-0.54074f,0.29799f,0.22423f,-0.15913f,-0.0342f,-0.00498f,-0.14694f,-0.39632f,0.07715f,0.38437f,0.40377f,-0.0297f,-0.56565f,0.03531f,0.50152f,0.08532f,-0.19971f,-0.09722f,0.34795f,-0.27888f,-0.40159f,0.42935f,0.24309f,-0.35656f,-0.43318f,0.39593f,0.27134f,0.03647f,-0.1333f,-0.50181f,0.00648f,0.09045f,-0.4133f,-0.3347f,0.27937f,0.61117f,0.09867f,-0.80386f,-0.86432f,0.36372f,0.74234f,0.07526f,0.16759f,-0.25427f,-0.31405f,-0.06994f,0.08958f,0.41402f,0.0976f,0.10665f,0.38007f,0.19081f,-0.41813f,-0.51675f,-0.10515f,0.54901f,0.39588f,-0.37489f,-0.41325f,0.02554f,0.13818f,-0.71961f,-0.40014f,0.57151f,0.22965f,-0.15332f,-0.0074f,0.16875f,-0.19869f,-0.22645f,0.45915f,0.38133f,-0.3037f,-0.67709f,-0.23652f,0.31821f,0.36314f,0.04014f,-0.01751f,0.77383f,0.47114f,-0.03144f,-0.02641f,-0.47318f,-0.41842f,-0.12159f,0.22142f,0.47845f,0.11124f,0.10617f,0.09606f,0.24986f,0.24222f,-0.55603f,-0.32706f,0.15526f,0.08493f,-0.35685f,-0.16856f,0.21344f,0.05654f,0.00788f,0.04285f,0.20014f,-0.24895f,-0.52106f,-0.0711f,-0.01224f,0.08991f,0.23806f,-0.302f,-0.26452f,-0.06878f,-0.30147f,-0.07139f,0.1407f,0.53151f,0.05707f,-0.5243f,0.08213f,0.21523f,0.1643f,0.07521f,-0.17818f,0.26989f,0.11816f,-0.4577f,-0.02326f,0.50229f,0.91525f,0.12571f,-0.33717f,-0.21315f,-0.47293f,0.18965f,0.34321f,-0.04039f,0.1584f,0.04861f,-0.23231f,0.04672f,0.09074f,-0.42611f,-0.46771f,0.07835f,0.30413f,-0.0414f,-0.41078f,0.03666f,0.29359f,0.02709f,-0.45552f,-0.1689f,0.38979f,0.08532f,-0.52735f,-0.48101f,0.50539f,0.36493f,-0.08072f,0.38249f,0.27279f,-0.45006f,-0.48159f,-0.06887f,0.22539f,0.15095f,-0.1717f,0.41649f,0.6798f,0.14181f,-0.56449f,-0.54398f,-0.08256f,-0.07463f,0.26365f,0.12687f,0.19192f,0.09572f,-0.48561f,0.09137f,0.86655f,-0.03671f,-0.29257f,0.09045f,-0.28991f,-0.45276f,-0.33838f,0.36348f,0.65639f,0.394f,-0.10215f,-0.54534f,-0.34413f,0.41407f,0.38021f,-0.09577f,-0.25262f,-0.1717f,0.18399f,-0.12672f,-0.37349f,0.35269f,0.12716f,-0.3629f,0.23148f,0.21199f,-0.10002f,-0.4162f,0.10259f,0.71743f,-0.12629f,-0.59874f,-0.29446f,-0.22766f,-0.2148f,-0.00295f,0.37214f,0.84604f,0.04116f,-0.50785f,0.02239f,0.20092f,0.3614f,-0.36657f,-0.60788f,0.32745f,0.47289f,-0.14534f,0.01591f,0.17925f,-0.14873f,-0.17906f,0.29494f,0.43917f,-0.08551f,-0.26592f,-0.29881f,-0.09514f,-0.10617f,0.03018f,0.42466f,0.45804f,-0.17993f,-0.62606f,-0.49252f,0.02273f,0.46253f,0.3805f,0.11927f,-0.028f,-0.05809f,-0.7464f,-0.67559f,0.02709f,0.40856f,0.32928f,0.01117f,0.17407f,0.1407f,-0.04875f,0.18143f,0.03816f,-0.16232f,-0.29794f,-0.09446f,0.2933f,-0.20203f,-0.2472f,-0.12174f,0.24092f,0.73629f,0.55966f,-0.50103f,-0.65199f,0.26612f,0.163f,-0.13988f,-0.05364f,0.21683f,-0.08953f,0.04073f,0.40246f,0.24638f,0.16072f,-0.49518f,-0.6946f,0.20575f,0.52222f,-0.1556f,-0.31705f,0.0223f,-0.31158f,-0.22592f,0.17596f,0.22752f,-0.06684f,-0.23647f,0.08575f,-0.22597f,-0.2193f,-0.00058f,-0.21528f,0.0976f,0.55608f,-0.08783f,-0.6842f,-0.02486f,0.53146f,0.4784f,0.01964f,-0.13915f,0.4636f,0.22384f,-0.33122f,-0.1243f,0.17456f,0.11492f,-0.38945f,0.05069f,0.76023f,0.15086f,-0.16798f,-0.22645f,-0.05195f,-0.18133f,-0.87162f,-0.33485f,0.17436f,0.48227f,0.51883f,0.21504f,-0.32745f,-0.76643f,-0.39356f,0.48062f,0.8945f,0.06423f,-0.34853f,-0.39535f,-0.38655f,-0.11835f,0.05819f,0.45127f,0.33784f,-0.04982f,-0.48798f,-0.68841f,0.15071f,0.79008f,0.47279f,-0.02858f,0.08919f,-0.17136f,-0.70176f,-0.27777f,0.33693f,0.03908f,-0.34229f,0.18781f,0.41509f,0.31168f,0.23758f,-0.0977f,-0.38761f,-0.37344f,-0.34679f,-0.11424f,0.14936f,0.21746f,0.02215f,-0.39376f,-0.13881f,-0.00938f,0.10689f,0.3822f,0.31487f,0.29044f,-0.39163f,-1.04371f,-0.27086f,0.77532f,0.6798f,-0.2297f,-0.5406f,0.40778f,0.60851f,-0.47806f,-0.26447f,0.44623f,0.74984f,0.36977f,-0.17567f,-0.20135f,-0.57344f,-0.35482f,0.24551f,0.69861f,-0.1302f,-0.51032f,0.20875f,0.48725f,0.03773f,-0.53151f,-0.42031f,-0.19922f,-0.03676f,-0.14099f,-0.14118f,0.05079f,0.56556f,0.33151f,-0.40387f,-0.17301f,0.04029f,-0.10926f,0.44101f,0.13804f,-0.2857f,-0.06355f,-0.04605f,-0.11105f,0.09731f,0.55235f,0.33818f,-0.36861f,-0.50577f,0.08237f,0.09756f,0.39419f,0.32904f,0.31835f,0.00716f,-0.68976f,-0.19192f,0.07299f,-0.47564f,-0.25804f,0.42674f,0.17901f,-0.12904f,-0.12841f,0.09238f,-0.27869f,-0.29973f,0.13707f,-0.27917f,-0.42582f,0.12406f,0.46979f,0.05538f,-0.01664f,-0.08832f,-0.22955f,0.02423f,0.15023f,0.14147f,-0.21194f,0.05238f,0.19124f,-0.19613f,0.03565f,0.82089f,0.63666f,-0.39617f,-0.43772f,0.25369f,-0.0606f,-0.3539f,0.15777f,0.61054f,0.52232f,-0.33905f,-0.52227f,-0.25417f,0.3568f,0.20769f,-0.06089f,-0.03158f,-0.34699f,-0.38665f,-0.2711f,0.04682f,0.33959f,0.24203f,-0.34558f,-0.59047f,-0.02384f,0.51332f,0.26762f,0.40488f,0.43468f,-0.39202f,-1.2545f,-0.28367f,0.71801f,0.27042f,-0.5123f,-0.36943f,0.78098f,0.61871f,-0.11318f,-0.04f,0.31956f,-0.05698f,-0.06036f,-0.08266f,-0.2724f,-0.05117f,0.31381f,0.51946f,-0.02612f,0.07071f,0.03995f,-0.29775f,-0.01461f,0.07487f,-0.13683f,-0.28764f,0.07183f,0.14442f,-0.17499f,-0.1821f,-0.12725f,-0.49857f,-0.22974f,0.38171f,0.01141f,-0.47748f,0.0681f,0.51259f,0.18375f,0.01335f,-0.6087f,-0.56203f,-0.3242f,0.16566f,0.51966f,0.09432f,0.00222f,-0.15652f,0.10844f,0.07516f,0.12126f,0.6057f,0.01006f,-0.47472f,-0.17649f,-0.00909f,-0.33876f,-0.27569f,0.65987f,0.64841f,-0.07864f,-0.50379f,0.18578f,0.55946f,0.117f,-0.20735f,-0.01969f,0.03855f,-0.61547f,-0.08677f,0.54321f,0.40498f,-0.20948f,-0.5494f,-0.06162f,0.73421f,0.56827f,-0.12f,-0.05354f,-0.01582f,-0.16798f,-0.53528f,-0.45334f,0.56435f,0.67342f,-0.03715f,-0.42321f,-0.31303f,0.11787f,-0.21006f,0.24957f,0.67738f,-0.16503f,-0.61813f,-0.30553f,-0.06239f,-0.23584f,0.04251f,0.55966f,0.33915f,-0.11226f,-0.39313f,-0.36367f,0.21311f,0.17533f,0.11153f,-0.08634f,-0.19071f,0.228f,0.2487f,0.12493f,-0.1391f,-0.45726f,-0.3407f,0.29112f,0.36846f,0.32846f,0.15182f,-0.39134f,-0.04977f,0.00044f,-0.10355f,0.19908f,0.46316f,0.20982f,-0.54427f,-0.48503f,0.02375f,0.00208f,0.00972f,0.49112f,0.44285f,0.28217f,-0.31216f,-0.79424f,-0.32643f,0.37373f,0.55259f,0.1539f,-0.03449f,-0.21717f,-0.41083f,-0.02118f,0.40125f,-0.06172f,-0.65276f,-0.15531f,0.40802f,0.24846f,0.20943f,0.04372f,-0.17383f,-0.13248f,-0.00445f,-0.0934f,0.0222f,-0.11361f,-0.04972f,-0.16121f,-0.03869f,0.29988f,0.21722f,0.26824f,0.11705f,0.21872f,0.05233f,-0.49818f,-0.60067f,0.2015f,0.34994f,0.1364f,0.2562f,0.01756f,-0.08068f,-0.1866f,-0.41891f,0.06592f,0.04131f,0.26902f,0.57992f,-0.12667f,-0.58534f,-0.65049f,-0.39269f,0.11463f,0.65813f,0.23182f,-0.14752f,-0.06868f,-0.19313f,-0.04121f,0.22041f,0.17282f,-0.08619f,-0.54379f,-0.24213f,0.06631f,0.0088f,0.21905f,-0.06075f,-0.37248f,-0.08174f,0.02631f,0.33794f,0.41039f,-0.2756f,-0.52217f,-0.29078f,-0.0015f,0.14162f,0.71148f,0.36038f,-0.31922f,-0.45504f,-0.19211f,0.02157f,0.12396f,0.53344f,0.75898f,0.47419f,-0.4072f,-0.87752f,-0.27613f,0.868f,0.60976f,-0.2458f,-0.28493f,0.01843f,0.16479f,0.04227f,0.33523f,0.35516f,-0.3274f,-0.43545f,0.06205f,0.01151f,-0.7035f,-0.45533f,0.34921f,0.51424f,-0.07763f,-0.34249f,0.00658f,0.08798f,0.07642f,-0.20658f,-0.46292f,-0.194f,-0.09698f,0.18921f,0.70534f,0.15289f,-0.11729f,-0.20459f,-0.19168f,0.31555f,0.10351f,-0.35061f,0.05944f,0.33175f,0.0428f,-0.1821f,0.26703f,0.44449f,-0.32183f,-0.24769f,0.20551f,0.32894f,0.09935f,-0.54713f,-0.42853f,0.17301f,0.22583f,0.10723f,-0.2235f,-0.14447f,-0.27903f,-0.26916f,0.22776f,0.23729f,-0.21204f,-0.67603f,-0.17277f,0.55627f,0.35922f,-0.67438f,-0.3213f,0.71549f,-0.12338f,-0.82659f,-0.06027f,0.90979f,0.65354f,-0.39023f,-0.39226f,-0.31985f,-0.02418f,0.33625f,0.34447f,0.55714f,0.14283f,-0.29944f,-0.23816f,0.22384f,0.68043f,0.44933f,-0.13712f,-0.188f,-0.10699f,-0.21673f,-0.1939f,0.22104f,0.23622f,-0.38936f,-0.65755f,-0.00489f,0.36063f,0.06892f,-0.30399f,-0.37296f,0.18641f,0.06293f,0.1185f,0.33344f,0.09132f,-0.48338f,-0.50505f,0.33388f,0.35516f,-0.02738f,0.18176f,0.28087f,0.75211f,0.22157f,-0.86195f,-0.58171f,-0.03342f,0.29441f,-0.02394f,0.18259f,0.01359f,-0.51313f,-0.27739f,0.63675f,0.7066f,0.02239f,-0.42244f,-0.66882f,0.17997f,0.03071f,-0.64817f,0.15497f,0.64464f,-0.08692f,-0.78224f,-0.05843f,0.79907f,0.12658f,-0.57629f,-0.19506f,0.27816f,0.05311f,-0.08725f,0.04305f,0.46553f,0.4385f,-0.6279f,-0.868f,-0.10263f,0.30283f,0.42142f,0.03937f,0.20841f,-0.40928f,-0.51405f,0.49102f,0.15129f,-0.09393f,-0.19332f,-0.20648f,-0.10539f,0.38708f,0.4296f,0.03574f,-0.19671f,0.05103f,0.50118f,0.04082f,-0.34636f,-0.62761f,-0.54113f,0.26113f,0.28416f,0.5184f,0.4886f,-0.25262f,-0.39671f,-0.01785f,0.00218f,-0.33746f,0.2146f,0.48875f,-0.03961f,-0.01911f,0.44193f,-0.20251f,-0.46418f,-0.06114f,0.1613f,-0.03937f,-0.00131f,0.27158f,0.01272f,0.06022f,0.03927f,-0.25006f,-0.04048f,0.34341f,-0.11802f,-0.10568f,-0.14887f,0.08247f,0.06239f,-0.50611f,0.19637f,0.85029f,0.5451f,-0.69112f,-1.19597f,-0.41426f,0.55477f,0.64328f,0.11361f,0.36338f,0.21867f,-0.1792f,-0.08213f,-0.05098f,-0.09523f,0.24285f,0.38897f,0.17407f,-0.16745f,-0.26834f,0.18099f,0.20256f,0.3851f,-0.29668f,-0.84522f,-0.29794f,0.23008f,0.12106f,-0.03265f,-0.08779f,-0.45828f,0.22491f,0.06534f,-0.07565f,-0.1141f,-0.24638f,-0.07337f,-0.391f,-0.01054f,0.43854f,0.60314f,-0.13731f,-0.68033f,-0.64043f,-0.06462f,0.57673f,0.18926f,0.29151f,-0.03715f,-0.06027f,0.63419f,0.38133f,-0.2845f,-0.17659f,0.27066f,0.0888f,-0.10897f,0.1763f,0.21606f,-0.21248f,0.39879f,0.44633f,-0.21548f,-0.3052f,-0.05533f,0.16082f,-0.20275f,-0.34655f,0.09224f,0.45431f,0.05562f,-0.09345f,-0.25422f,-0.55385f,-0.23608f,0.14699f,0.29127f,-0.1674f,-0.60865f,0.08155f,0.3866f,-0.0414f,-0.3363f,-0.0237f,0.12479f,-0.3141f,-0.13049f,0.4561f,0.63699f,-0.00943f,-0.15589f,-0.20087f,0.18926f,0.43105f,0.04116f,0.25968f,0.07734f,-0.44561f,-0.72967f,0.23951f,1.27346f,0.44638f,-0.77489f,-0.64333f,0.24508f,0.34878f,-0.31468f,-0.32962f,0.07492f,-0.04213f,-0.09064f,-0.16324f,-0.04092f,0.18225f,-0.00938f,-0.29427f,0.02675f,0.29427f,-0.4252f,-0.23003f,0.2045f,0.40715f,-0.20043f,-0.43864f,0.50055f,0.19695f,0.03299f,-0.13586f,-0.03095f,0.43971f,-0.01117f,-0.12174f,0.2193f,-0.12135f,-0.21238f,0.03313f,0.26244f,-0.28933f,-0.36628f,0.19216f,0.26167f,0.19947f,-0.24034f,-0.13712f,-0.21025f,-0.36358f,-0.2606f,-0.08967f,0.36159f,0.67714f,-0.37528f,-1.00139f,-0.10617f,0.58577f,0.13257f,-0.11173f,0.31792f,-0.03347f,-0.8532f,-0.14089f,0.90253f,0.44416f,-0.05727f,-0.15942f,0.55922f,0.22336f,-0.39753f,-0.39376f,0.09678f,0.81707f,0.36106f,-0.13195f,0.08832f,0.20996f,-0.03086f,-0.33809f,-0.36295f,0.15076f,0.06665f,-0.02423f,0.34713f,0.14612f,-0.62602f,-0.91162f,-0.19685f,0.60232f,-0.02177f,-0.88294f,-0.07516f,0.16909f,-0.23448f,-0.22607f,-0.11216f,0.55516f,0.38355f,-0.21939f,-0.78742f,-0.222f,0.55322f,-0.13383f,0.24319f,0.63936f,0.29286f,0.24933f,0.57953f,0.48217f,-0.6353f,-0.70224f,0.38737f,0.77915f,0.44914f,-0.33151f,-0.36285f,0.50776f,0.45634f,-0.53252f,-0.37035f,0.49702f,0.48178f,-0.48328f,-0.65155f,0.03062f,-0.32599f,0.1512f,0.30858f,-0.47158f,-0.57064f,-0.23589f,0.36092f,0.18829f,-0.22525f,-0.41586f,-0.5851f,0.19013f,0.59999f,-0.19652f,-0.76865f,-0.12363f,0.35168f,-0.03797f,0.38945f,0.61813f,0.13954f,-0.44672f,-0.0667f,0.40391f,0.06065f,-0.03091f,0.17175f,0.57794f,0.1822f,-0.36459f,-0.04551f,0.12556f,-0.11545f,-0.25644f,-0.2399f,0.2694f,0.58568f,-0.1213f,-0.43511f,0.0518f,0.41736f,0.16421f,-0.48285f,-0.32822f,0.09988f,-0.04566f,-0.20232f,0.06355f,0.11047f,-0.15207f,-0.09451f,0.27642f,0.04658f,-0.61223f,-0.50524f,0.03318f,0.24793f,0.09519f,-0.05156f,0.03115f,0.56048f,0.46234f,-0.30036f,-0.3836f,-0.31453f,0.06089f,0.16566f,0.30529f,0.305f,-0.57615f,-0.33499f,0.05915f,0.27913f,0.23584f,-0.04798f,-0.1601f,0.02322f,0.03884f,-0.30989f,0.01877f,0.10559f,0.06452f,0.05901f,0.04987f,0.15584f,-0.33417f,-0.09553f,0.20904f,-0.34196f,0.3393f,0.35995f,-0.17712f,-0.2608f,-0.00643f,0.668f,0.40754f,0.07923f,0.22645f,-0.14355f,-0.09731f,0.13606f,-0.17282f,-0.28643f,-0.09562f,-0.02215f,0.14399f,0.29944f,-0.31603f,-0.24977f,-0.01669f,0.21925f,-0.07434f,-0.59056f,-0.18708f,0.16793f,0.20261f,-0.30447f,-0.10887f,0.14699f,-0.17891f,0.00334f,0.24948f,0.34471f,-0.47327f,-0.75912f,0.18162f,0.50103f,0.12421f,-0.44556f,0.14684f,0.51216f,-0.16498f,-0.34012f,0.12508f,0.58887f,0.64357f,0.06365f,-0.52541f,0.02215f,0.49455f,-0.271f,-0.52744f,0.67733f,0.46365f,-0.56309f,-0.00077f,0.17485f,-0.58089f,-0.11744f,0.49509f,0.34012f,-0.13161f,-0.97257f,-0.14225f,0.653f,0.42268f,-0.64294f,-1.26098f,0.25064f,0.77895f,0.13054f,-0.60807f,-0.29606f,0.23187f,0.1155f,0.1319f,0.05901f,-0.13185f,-0.34031f,0.01466f,-0.03086f,0.33978f,0.07947f,-0.38012f,0.48856f,0.32154f,-0.30466f,-0.05485f,0.47975f,0.18549f,-0.42853f,-0.32271f,0.36648f,0.42602f,-0.05751f,-0.2504f,-0.12387f,0.1703f,0.1687f,-0.01519f,0.25901f,0.16725f,-0.15719f,0.02631f,-0.01678f,0.07091f,0.0919f,0.20575f,0.40125f,0.24063f,-0.58103f,-0.69953f,0.21548f,1.26262f,0.55027f,-1.42954f,-1.0086f,0.68319f,0.58403f,-0.12213f,0.36585f,-0.24067f,-0.4859f,-0.05673f,0.64348f,0.56053f,-0.45992f,-0.58045f,-0.47985f,0.32063f,0.27961f,-0.07536f,-0.1852f,-0.17997f,-0.37451f,-0.7731f,-0.25417f,0.19395f,0.30979f,-0.16309f,-0.12063f,-0.08309f,-0.34829f,0.38684f,0.50597f,0.30283f,-0.25209f,-0.45523f,-0.02767f,0.35163f,0.58195f,0.37199f,0.1036f,-0.10346f,-0.01935f,-0.07507f,0.25794f,0.22263f,-0.11057f,-0.14694f,-0.33697f,-0.04092f,0.63395f,0.26375f,-0.24193f,0.29369f,-0.3067f,-0.91225f,0.04131f,0.39003f,-0.30312f,-0.7331f,-0.14389f,0.46935f,0.39966f,-0.32242f,-0.61644f,-0.12213f,0.49939f,0.67163f,-0.4605f,-0.46974f,0.10147f,0.24f,0.11531f,-0.09301f,0.33257f,0.34428f,-0.04455f,-0.30916f,0.06399f,0.2385f,-0.15845f,-0.12198f,0.34897f,-0.09074f,-0.38457f,0.32783f,0.60154f,0.34674f,0.31869f,0.17688f,-0.24275f,-0.27879f,-0.13427f,0.24275f,0.12053f,0.04619f,0.21456f,-0.59419f,-0.43821f,0.51434f,0.82737f,-0.07748f,-1.11873f,-0.73334f,-0.09185f,0.37402f,0.34113f,0.03081f,-0.44628f,-0.29262f,0.24406f,0.36653f,0.15386f,-0.36164f,-0.32566f,-0.08092f,0.07618f,0.05335f,-0.11671f,-0.11313f,0.78442f,0.8295f,-0.50892f,-0.8117f,-0.48691f,0.07874f,0.30621f,0.09949f,-0.26974f,-0.47985f,0.30036f,0.53049f,-0.28227f,-0.46935f,0.10689f,0.01093f,-0.34878f,-0.31888f,0.18123f,0.37683f,0.0979f,0.28483f,-0.00131f,0.00996f,0.42094f,0.23613f,-0.28416f,-0.27869f,0.20648f,0.39371f,0.03371f,-0.55859f,0.03889f,0.46495f,0.49818f,0.46007f,-0.24483f,-0.34157f,-0.18694f,0.2163f,0.62099f,-0.40154f,-1.04618f,-0.24541f,0.40894f,0.54481f,-0.10051f,-0.31932f,0.21136f,0.37131f,-0.0016f,-0.55279f,-0.19995f,0.4044f,0.13107f,-0.6132f,-0.30466f,0.46263f,0.07405f,-0.14375f,0.11753f,-0.00097f,-0.32063f,-0.65397f,-0.26752f,0.11279f,0.03773f,0.05209f,0.30776f,0.07908f,-0.40981f,0.04077f,0.16159f,0.29209f,0.31356f,-0.04193f,-0.35535f,0.37417f,0.73755f,-0.14123f,-0.20309f,-0.01132f,0.17141f,0.10636f,0.14196f,-0.48411f,-0.37209f,0.28701f,0.35647f,0.32624f,-0.36358f,-0.45905f,-0.05678f,0.36861f,0.40967f,-0.05243f,-0.67124f,-0.76014f,-0.08184f,0.2518f,0.29794f,0.33765f,0.10733f,-0.17606f,-0.60188f,-0.25639f,0.47816f,0.14002f,-0.35414f,0.14568f,0.00667f,-0.36285f,-0.29872f,0.34636f,1.13735f,-0.26466f,-0.92115f,-0.23008f,-0.12189f,-0.00295f,0.20155f,0.25064f,0.3405f,-0.12382f,-0.48237f,0.33296f,0.67603f,0.25344f,-0.21069f,-0.24938f,-0.15255f,0.07565f,0.61924f,0.54326f,0.09654f,-0.39656f,-0.27139f,-0.08774f,0.17349f,0.34374f,-0.45731f,-0.4029f,0.22902f,-0.02418f,-0.38302f,0.14558f,0.10056f,-0.10771f,-0.12387f,0.21219f,0.08237f,-0.63632f,-0.13059f,0.43859f,0.41513f,-0.15835f,-0.65363f,-0.23685f,0.39129f,-0.00044f,-0.18321f,0.268f,0.31197f,0.04329f,-0.31284f,-0.48706f,-0.02515f,0.10196f,0.07158f,0.0105f,0.00392f,0.24803f,-0.22868f,0.04532f,0.55346f,-0.01403f,-0.57494f,-0.15482f,0.42002f,0.10554f,0.13606f,0.80962f,0.53296f,-0.35342f,-0.60609f,-0.03758f,0.4975f,0.4651f,0.00967f,-0.4294f,-0.22597f,0.3954f,0.21896f,-0.01156f,0.16329f,-0.23864f,-0.54326f,-0.14162f,0.02849f,-0.52087f,-0.07918f,0.60706f,0.4753f,-0.35487f,-0.5539f,-0.09083f,-0.09988f,0.1482f,-0.01809f,-0.25494f,-0.46984f,-0.20682f,0.59521f,0.57397f,-0.04832f,-0.42752f,-0.14128f,-0.00547f,0.16653f,0.09693f,-0.01722f,-0.09557f,-0.09693f,0.12551f,-0.2223f,-0.19245f,0.19419f,0.65528f,-0.13552f,-0.52053f,0.2711f,0.24609f,-0.27144f,-0.45987f,0.23951f,0.34031f,-0.04111f,-0.25074f,-0.15192f,-0.08561f,0.47066f,0.49039f,-0.47211f,-0.28048f,-0.13848f,0.19961f,0.51941f,0.04537f,-0.17262f,-0.10307f,0.24769f,0.41301f,0.04435f,-0.20169f,0.14278f,0.3733f,0.10118f,-0.42529f,-0.44551f,-0.00208f,0.01886f,0.43434f,0.28343f,-0.56459f,-0.68575f,-0.14089f,0.29997f,-0.13412f,-0.10031f,0.02312f,-0.2178f,-0.18602f,-0.33107f,0.10825f,0.36711f,0.1288f,-0.15632f,-0.81276f,0.19308f,0.91719f,0.21838f,0.36125f,0.24653f,-0.16711f,-0.14438f,0.32401f,0.85581f,0.08111f,-0.36744f,0.20571f,-0.01509f,-0.06259f,0.03342f,0.11671f,0.33344f,0.11526f,-0.11942f,-0.40996f,-0.51666f,-0.11826f,-0.05407f,-0.01006f,-0.13649f,-0.77542f,-0.14172f,0.60009f,0.13528f,-0.10733f,-0.31381f,-0.41509f,-0.13035f,-0.29925f,-0.21635f,-0.06302f,0.00527f,0.28353f,-0.06515f,0.26829f,0.57044f,-0.47129f,-0.54805f,0.21282f,0.51554f,0.00667f,-0.52614f,0.01761f,0.52154f,0.3763f,-0.05364f,-0.08401f,0.34476f,-0.09935f,-0.25281f,0.64556f,0.47893f,0.25074f,-0.40938f,-0.61533f,0.03889f,-0.0431f,0.43763f,0.50843f,0.32343f,-0.39192f,-1.0603f,-0.15811f,0.92333f,0.68918f,-0.21915f,-0.41538f,-0.31487f,-0.34394f,-0.43376f,0.12672f,0.90742f,0.44628f,-0.34631f,-0.99544f,-0.45301f,0.53973f,0.70437f,0.21103f,-0.38669f,-0.04034f,0.15342f,0.30321f,0.19163f,-0.08314f,-0.58616f,-0.19318f,0.31168f,0.25456f,-0.07507f,-0.02017f,0.54016f,0.13915f,-0.68773f,-0.72851f,0.25015f,0.32749f,-0.20392f,-0.62524f,0.06733f,-0.05732f,-0.30065f,0.10945f,-0.06883f,0.57286f,0.00276f,-0.48024f,0.18999f,-0.03971f,-0.20783f,-0.42244f,0.64285f,1.12502f,-0.25296f,-0.53804f,0.23008f,0.70838f,0.11066f,-0.32121f,-0.04619f,0.52923f,-0.03686f,-0.62069f,-0.07313f,0.54539f,0.21664f,-0.36348f,-0.17132f,-0.0636f,-0.69044f,-0.18718f,0.71859f,0.32537f,-0.55729f,-0.73237f,-0.10592f,0.55709f,0.63027f,-0.46974f,-0.4798f,-0.08929f,-0.1733f,-0.0253f,0.13886f,0.34626f,0.17843f,-0.59095f,-0.27995f,0.14389f,0.52338f,0.57184f,0.03521f,0.0964f,-0.30123f,-0.55322f,-0.06394f,0.42452f,-0.04111f,0.16454f,0.70423f,0.39608f,-0.2858f,-0.56967f,0.35414f,0.73373f,0.11255f,-0.37059f,-1.06819f,-0.31206f,0.63845f,0.20614f,-0.43792f,-0.20972f,0.51835f,-0.18868f,-0.26999f,0.16711f,-0.0237f,-0.3007f,-0.234f,-0.154f,-0.38437f,-0.10254f,0.31855f,0.17601f,0.0741f,0.0015f,-0.39661f,0.11985f,0.40033f,0.07884f,0.04455f,-0.11647f,-0.04329f,0.0874f,-0.18587f,-0.07734f,0.38094f,0.17277f,-0.23772f,-0.12648f,0.25267f,0.0548f,-0.00832f,0.30026f,0.29673f,0.40198f,-0.40261f,-0.76381f,-0.22645f,0.09567f,0.72705f,0.1569f,-0.34355f,-0.13731f,-0.1941f,-0.03584f,0.29214f,0.38268f,0.36653f,-0.03937f,-0.56174f,-0.65741f,-0.17306f,0.58703f,0.67593f,0.18762f,-0.41639f,-0.23008f,0.07753f,0.47501f,0.20493f,-0.03212f,-0.11584f,-0.46002f,-0.14655f,0.23013f,0.26544f,-0.33939f,-0.20885f,0.34795f,0.17127f,-0.26757f,-0.01185f,0.29702f,-0.48498f,-0.44474f,-0.39139f,-0.26404f,0.45262f,0.18868f,0.18974f,-0.45954f,-0.26771f,0.2354f,0.08943f,-0.11081f,-0.33596f,0.1436f,0.60053f,0.23352f,-0.75448f,-0.0816f,0.59255f,0.19023f,0.09359f,0.23385f,-0.05528f,0.0519f,-0.22467f,-0.41586f,0.33949f,0.66858f,0.25301f,-0.37722f,-0.01911f,-0.28483f,-0.31545f,0.65392f,0.68623f,-0.51042f,-1.2012f,0.05775f,0.66969f,0.6266f,0.03671f,-0.89474f,-0.2902f,0.41871f,0.27086f,-0.23864f,-0.08251f,-0.00198f,-0.22781f,-0.15516f,0.117f,0.01519f,-0.12542f,0.36154f,0.15773f,-0.59124f,-0.45006f,0.03729f,0.46321f,0.62998f,0.08861f,-0.63646f,-0.27047f,0.5096f,0.10815f,-0.18438f,0.10022f,0.44338f,0.04735f,-0.09127f,0.06515f,0.11792f,0.1421f,-0.03381f,0.11013f,0.33842f,0.18055f,-0.50853f,-0.24295f,-0.06568f,-0.11168f,-0.03758f,0.12798f,0.11831f,-0.02186f,-0.13746f,-0.19506f,0.74031f,0.12039f,-0.65136f,-0.8191f,-0.28198f,0.54452f,0.1377f,0.04271f,-0.25857f,-0.33122f,-0.13891f,0.5035f,0.69813f,0.23564f,-0.66282f,-0.79723f,-0.27323f,-0.16309f,0.54369f,-0.06181f,0.05572f,0.72996f,-0.31327f,-0.61644f,-0.02293f,0.18123f,-0.12561f,-0.19429f,0.35821f,0.15269f,-0.50882f,-0.14138f,0.38965f,0.61073f,0.28614f,-0.72696f,-0.22573f,0.38819f,0.19347f,-0.16406f,-0.10099f,0.23356f,-0.18868f,0.2963f,0.31695f,-0.1245f,-0.64227f,-0.25054f,0.39419f,0.37388f,0.12788f,-0.37992f,0.19739f,0.26152f,0.05364f,-0.5732f,-0.02888f,0.58781f,-0.39468f,-0.19032f,0.32962f,0.3288f,-0.13538f,-0.47922f,-0.17132f,0.58195f,0.3287f,-0.4177f,-0.2149f,0.1467f,0.09156f,0.11676f,0.45649f,-0.26229f,-0.70118f,-0.2667f,0.17978f,0.25765f,-0.323f,-0.1229f,0.67424f,0.67046f,-0.37881f,-0.95157f,-0.62887f,0.21141f,0.54645f,0.25325f,0.04106f,0.09543f,0.52517f,0.09166f,-0.3896f,0.13398f,0.65267f,0.05615f,-0.41789f,-0.44933f,-0.07637f,0.6044f,0.57852f,0.32706f,-0.24377f,-0.73083f,-0.04576f,0.47806f,0.00256f,-0.64681f,-0.61644f,0.45838f,0.64793f,0.01296f,-0.27071f,-0.14283f,0.07598f,-0.36928f,-0.79375f,-0.05746f,0.14699f,-0.02907f,-0.24498f,-0.394f,0.01161f,-0.02143f,-0.03323f,0.08334f,0.08435f,-0.33823f,-0.28367f,0.00276f,0.44575f,0.42239f,-0.36624f,-0.20982f,0.35264f,0.2384f,0.01809f,0.01238f,0.30563f,0.3422f,0.26665f,0.31768f,0.10776f,-0.04629f,-0.3065f,-0.32769f,0.26423f,0.73165f,0.25717f,0.06167f,-0.29601f,-0.23139f,0.19739f,0.19642f,-0.10147f,-0.91419f,-0.66466f,0.14321f,0.53213f,0.23226f,-0.3378f,-0.25402f,-0.03231f,-0.30031f,-0.36116f,0.13088f,0.07898f,-0.06868f,0.37939f,-0.13852f,-0.47409f,0.02346f,0.29857f,0.24271f,0.29552f,0.23937f,-0.69397f,-0.60918f,0.55148f,0.52304f,0.17562f,0.40565f,0.10041f,-0.21606f,-0.14897f,0.00943f,0.0623f,-0.04058f,-0.07139f,0.03178f,0.34002f,0.11023f,-0.42665f,-0.13635f,-0.0014f,-0.40135f,-0.28807f,0.28309f,0.26007f,-0.06104f,-0.14679f,-0.18597f,-0.50998f,-0.09528f,0.77092f,0.39231f,-0.12227f,-0.32657f,-0.05016f,-0.41455f,-0.72377f,0.42026f,0.91351f,0.31337f,-0.50577f,-0.63487f,0.19115f,0.25359f,-0.11768f,-0.0473f,0.18529f,0.04653f,-0.40633f,-0.07042f,0.25552f,0.17224f,-0.17122f,0.02728f,0.29417f,0.07792f,-0.14384f,-0.30316f,-0.14031f,-0.39376f,-0.012f,0.53857f,0.30075f,0.11908f,-0.35734f,-0.39816f,0.3806f,0.47985f,0.28807f,0.18554f,-0.21315f,-0.06999f,-0.20556f,-0.06355f,0.39903f,0.3213f,0.12387f,-0.2134f,0.01243f,0.4161f,-0.06438f,-0.32125f,0.1111f,-0.48149f,-0.05243f,0.36914f,-0.06114f,-0.12885f,-0.35864f,0.14418f,0.25891f,-0.01272f,-0.23874f,0.02675f,-0.04769f,-0.2843f,-0.36788f,-0.13359f,0.50698f,0.02268f,-0.42964f,-0.30452f,-0.02128f,0.5509f,0.39676f,-0.5642f,-0.81997f,0.49625f,1.05063f,-0.12924f,-0.20619f,0.20363f,0.08087f,-0.21325f,-0.3539f,-0.19729f,0.07744f,0.33291f,0.46099f,0.23318f,0.02878f,-0.29301f,-0.09751f,0.6458f,-0.33059f,-0.76773f,0.06839f,0.43763f,0.3479f,-0.31385f,-0.28754f,0.15695f,0.25954f,0.3302f,-0.20319f,-0.4901f,-0.17673f,-0.20827f,-0.32783f,0.12164f,0.1675f,0.14829f,0.1097f,-0.12382f,0.22897f,0.14104f,-0.31158f,-0.1438f,0.02767f,0.03569f,0.20483f,0.02665f,0.0327f,0.02215f,-0.25954f,-0.37678f,0.24479f,0.57151f,0.11806f,-0.33088f,-0.31477f,-0.51704f,-0.0384f,1.01194f,0.64739f,-0.34123f,-0.95989f,-0.24082f,0.54181f,0.7716f,0.03555f,-0.54258f,-0.56831f,-0.02273f,0.26931f,-0.17214f,0.30036f,0.25833f,0.07265f,-0.11555f,-0.11811f,0.16706f,-0.1837f,-0.23023f,0.29789f,0.41978f,-0.12193f,-0.54679f,0.10085f,0.6828f,0.22694f,-0.3763f,0.21552f,0.13296f,-0.4576f,0.13195f,0.25562f,0.13765f,-0.26931f,-0.4918f,0.01054f,0.65184f,0.52198f,-0.85189f,-0.7375f,0.1718f,-0.07719f,0.20256f,0.08164f,-0.42786f,-0.22549f,0.11671f,0.09959f,0.21151f,-0.43796f,-0.71627f,0.30911f,0.80952f,0.42123f,-0.59051f,-0.31961f,0.60135f,0.56072f,0.14684f,0.01915f,-0.2443f,-0.50074f,-0.26186f,0.42795f,0.35763f,0.00222f,-0.06723f,0.29518f,0.78906f,0.07942f,-0.49968f,-0.16595f,0.52265f,0.12938f,-0.45417f,-0.397f,-0.07153f,0.0976f,0.2013f,0.76294f,0.46355f,-0.16005f,-0.47511f,-0.38771f,-0.18771f,-0.31719f,0.0193f,-0.11095f,-0.21335f,-0.20991f,-0.07313f,0.25852f,0.41296f,0.03473f,-0.69528f,-0.33001f,-0.20812f,0.6634f,0.662f,-0.42906f,-0.59371f,-0.00788f,0.60038f,-0.24271f,-0.38602f,0.49422f,0.44058f,0.11618f,0.22704f,-0.0578f,-0.36295f,-0.33364f,0.44241f,0.40285f,-0.37248f,-0.35042f,-0.24759f,0.68386f,0.59056f,-0.60072f,-0.19826f,0.23792f,0.0549f,-0.17876f,-0.1881f,-0.04184f,-0.16179f,-0.34157f,-0.09446f,-0.10336f,-0.15328f,0.10704f,-0.45557f,-0.0104f,0.33122f,-0.08595f,0.1599f,0.27642f,0.1941f,0.41455f,0.16285f,-0.44507f,-0.6666f,-0.41862f,0.91254f,1.15114f,0.5375f,-0.1201f,-1.10896f,-0.10036f,0.83312f,0.42573f,0.18418f,-0.11013f,-0.33368f,-0.4649f,-0.04034f,0.63874f,0.49828f,-0.05069f,-0.4339f,-0.58708f,-0.18007f,-0.2591f,0.03279f,0.07942f,-0.25746f,-0.17461f,-0.19772f,0.13693f,-0.15419f,0.10046f,0.32295f,0.03163f,-0.07744f,-0.15918f,0.05727f,0.03661f,0.36778f,0.41059f,-0.08667f,-0.28421f,-0.05586f,0.49189f,0.23197f,-0.55733f,-0.14583f,0.1408f,-0.27008f,0.07095f,0.24875f,0.01035f,-0.24648f,0.03091f,-0.10249f,-0.30162f,0.29223f,-0.36145f,-0.5111f,0.38873f,0.31027f,-0.27294f,-0.97837f,-0.74079f,0.48633f,0.99912f,0.04813f,-0.98026f,-0.53775f,0.04566f,0.32914f,0.67617f,0.28522f,-0.80531f,-0.72546f,0.57823f,0.56967f,0.22225f,0.34863f,0.24479f,-0.12677f,-0.36217f,0.20682f,0.55395f,0.33262f,-0.15526f,0.08822f,0.69373f,0.60406f,-0.38539f,-0.83965f,0.15594f,0.91351f,0.48996f,-0.29736f,-0.38597f,-0.58781f,-0.41987f,0.63845f,0.43782f,-0.6665f,-0.69175f,-0.47448f,0.13427f,-0.08251f,-0.41185f,0.20508f,-0.20875f,0.19202f,0.47671f,-0.43661f,-0.55119f,0.32517f,0.63003f,-0.17238f,-0.50902f,-0.21098f,0.55501f,0.25543f,0.53107f,0.78722f,-0.29281f,-0.72019f,-0.20275f,1.23346f,0.69117f,-0.32865f,-0.25765f,0.01195f,0.0238f,0.37397f,-0.00155f,-0.19632f,0.05141f,-0.49359f,-0.15018f,-0.37141f,0.21983f,0.08334f,-0.66002f,-0.44474f,-0.5539f,-0.20295f,-0.08687f,0.29799f,0.06002f,-0.51496f,-0.37064f,0.02902f,0.28101f,-0.24909f,-0.33344f,-0.21572f,0.88386f,0.61359f,-0.74137f,-0.4103f,0.53325f,1.13039f,0.76057f,0.58979f,-0.82611f,-1.05532f,0.36111f,0.53668f,0.76768f,0.58215f,-0.47274f,-0.50215f,0.13456f,0.01591f,-0.41001f,-0.24382f,0.44038f,0.07395f,-0.52256f,-0.35105f,-0.56594f,0.16324f,0.47651f,-0.30786f,-0.25006f,-0.02317f,0.12169f,0.0978f,0.04585f,0.04764f,0.12837f,-0.32624f,0.2089f,0.39197f,-0.09548f,-0.11681f,0.01538f,0.37243f,-0.21373f,0.11574f,1.01266f,0.22355f,-0.66727f,-0.2666f];
struct DSP_DrumSIG0 {
nothrow:
@nogc:
	
  private:
	
	int fDSP_DrumSIG0Wave0_idx;
	
  public:
	
	int getNumInputsDSP_DrumSIG0() nothrow @nogc {
		return 0;
	}
	int getNumOutputsDSP_DrumSIG0() nothrow @nogc {
		return 1;
	}
	
	void instanceInitDSP_DrumSIG0(int sample_rate) nothrow @nogc {
		fDSP_DrumSIG0Wave0_idx = 0;
	}
	
	void fillDSP_DrumSIG0(int count, float[] table) nothrow @nogc {
		for (int i1 = 0; i1 < count; i1 = i1 + 1) {
			table[i1] = fDSP_DrumSIG0Wave0[fDSP_DrumSIG0Wave0_idx];
			fDSP_DrumSIG0Wave0_idx = (1 + fDSP_DrumSIG0Wave0_idx) % 8542;
		}
	}

};

DSP_DrumSIG0* newDSP_DrumSIG0() nothrow @nogc { return assumeNothrowNoGC(&mallocNew!(DSP_DrumSIG0))(); }
void deleteDSP_DrumSIG0(DSP_DrumSIG0* dsp) nothrow @nogc { destroyFree(dsp); }
alias FAUSTCLASS = DSP_Drum;
static float DSP_Drum_faustpower2_f(float value) nothrow @nogc {
	return value * value;
}
__gshared float[8542] ftbl0DSP_DrumSIG0;

class DSP_Drum : dsp
{
nothrow:
@nogc:
	
 private:
	
	int fSampleRate;
	float fConst0;
	float fConst2;
	FAUSTFLOAT fEntry0;
	float[2] fVec0;
	int[2] iRec1;
	float fConst3;
	float fConst4;
	FAUSTFLOAT fEntry1;
	int[2] iVec1;
	int[2] iRec4;
	float fConst5;
	FAUSTFLOAT fHslider0;
	float[2] fRec5;
	float[2] fRec6;
	float[2] fRec7;
	FAUSTFLOAT fHslider1;
	FAUSTFLOAT fHslider2;
	float[2] fRec8;
	float[2] fRec2;
	FAUSTFLOAT fHslider3;
	FAUSTFLOAT fCheckbox0;
	float[2] fVec2;
	FAUSTFLOAT fHslider4;
	float[2] fRec10;
	float fConst6;
	float[2] fRec11;
	float[2] fRec9;
	float[2] fVec3;
	float fConst7;
	float fConst8;
	float[2] fRec0;
	float[2] fRec14;
	FAUSTFLOAT fHslider5;
	float[2] fRec12;
	FAUSTFLOAT fHslider6;
	FAUSTFLOAT fHslider7;
	float fConst9;
	FAUSTFLOAT fHslider8;
	float fConst10;
	float[2] fRec16;
	float[3] fRec15;
	FAUSTFLOAT fHslider9;
	float[2] fRec17;
	float fConst13;
	float fConst14;
	float fConst16;
	float fConst17;
	int iConst18;
	int[2] iRec21;
	int[2] iRec22;
	float[2] fRec20;
	float[3] fRec19;
	float fConst19;
	float[3] fRec18;
	FAUSTFLOAT fHslider10;
	float[2] fRec25;
	FAUSTFLOAT fHslider11;
	float fConst20;
	float[2] fRec23;
	FAUSTFLOAT fHslider12;
	float[2] fRec28;
	float[2] fRec26;
	FAUSTFLOAT fHslider13;
	FAUSTFLOAT fHslider14;
	float[2] fRec29;
	FAUSTFLOAT fHslider15;
	float[2] fRec30;
	FAUSTFLOAT fHslider16;
	
 public:
	
	void metadata(Meta* m) nothrow @nogc { 
		m.declare("author", "Oren Kurtz");
		m.declare("basics.lib/downSample:author", "Romain Michon");
		m.declare("basics.lib/name", "Faust Basic Element Library");
		m.declare("basics.lib/sAndH:author", "Romain Michon");
		m.declare("basics.lib/version", "0.8");
		m.declare("compile_options", "-a arch.fastmath.d.template -lang dlang -cn DSP_Drum -pn processDrum -es 1 -mcd 16 -single -ftz 0");
		m.declare("copyright", "Punk Labs LLC 2022");
		m.declare("envelopes.lib/ar:author", "Yann Orlarey, Stéphane Letz");
		m.declare("envelopes.lib/author", "GRAME");
		m.declare("envelopes.lib/copyright", "GRAME");
		m.declare("envelopes.lib/license", "LGPL with exception");
		m.declare("envelopes.lib/name", "Faust Envelope Library");
		m.declare("envelopes.lib/version", "0.2");
		m.declare("filename", "Main.dsp");
		m.declare("filters.lib/fir:author", "Julius O. Smith III");
		m.declare("filters.lib/fir:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/fir:license", "MIT-style STK-4.3 license");
		m.declare("filters.lib/highpass:author", "Julius O. Smith III");
		m.declare("filters.lib/highpass:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/iir:author", "Julius O. Smith III");
		m.declare("filters.lib/iir:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/iir:license", "MIT-style STK-4.3 license");
		m.declare("filters.lib/lowpass0_highpass1", "MIT-style STK-4.3 license");
		m.declare("filters.lib/lowpass0_highpass1:author", "Julius O. Smith III");
		m.declare("filters.lib/lowpass:author", "Julius O. Smith III");
		m.declare("filters.lib/lowpass:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/lowpass:license", "MIT-style STK-4.3 license");
		m.declare("filters.lib/name", "Faust Filters Library");
		m.declare("filters.lib/pole:author", "Julius O. Smith III");
		m.declare("filters.lib/pole:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/pole:license", "MIT-style STK-4.3 license");
		m.declare("filters.lib/resonlp:author", "Julius O. Smith III");
		m.declare("filters.lib/resonlp:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/resonlp:license", "MIT-style STK-4.3 license");
		m.declare("filters.lib/tf1:author", "Julius O. Smith III");
		m.declare("filters.lib/tf1:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/tf1:license", "MIT-style STK-4.3 license");
		m.declare("filters.lib/tf1s:author", "Julius O. Smith III");
		m.declare("filters.lib/tf1s:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/tf1s:license", "MIT-style STK-4.3 license");
		m.declare("filters.lib/tf2:author", "Julius O. Smith III");
		m.declare("filters.lib/tf2:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/tf2:license", "MIT-style STK-4.3 license");
		m.declare("filters.lib/tf2s:author", "Julius O. Smith III");
		m.declare("filters.lib/tf2s:copyright", "Copyright (C) 2003-2019 by Julius O. Smith III <jos@ccrma.stanford.edu>");
		m.declare("filters.lib/tf2s:license", "MIT-style STK-4.3 license");
		m.declare("filters.lib/version", "0.3");
		m.declare("interpolators.lib/interpolate_linear:author", "Stéphane Letz");
		m.declare("interpolators.lib/interpolate_linear:licence", "MIT");
		m.declare("interpolators.lib/name", "Faust Interpolator Library");
		m.declare("interpolators.lib/version", "0.3");
		m.declare("license", "GPLv3-or-later");
		m.declare("maths.lib/author", "GRAME");
		m.declare("maths.lib/copyright", "GRAME");
		m.declare("maths.lib/license", "LGPL with exception");
		m.declare("maths.lib/name", "Faust Math Library");
		m.declare("maths.lib/version", "2.5");
		m.declare("name", "OneTrick SIMIAN");
		m.declare("noises.lib/name", "Faust Noise Generator Library");
		m.declare("noises.lib/version", "0.4");
		m.declare("options", "[midi:on]");
		m.declare("oscillators.lib/name", "Faust Oscillator Library");
		m.declare("oscillators.lib/polyblep:author", "Jacek Wieczorek");
		m.declare("oscillators.lib/polyblep_square:author", "Jacek Wieczorek");
		m.declare("oscillators.lib/polyblep_triangle:author", "Jacek Wieczorek");
		m.declare("oscillators.lib/version", "0.3");
		m.declare("platform.lib/name", "Generic Platform Library");
		m.declare("platform.lib/version", "0.2");
		m.declare("signals.lib/name", "Faust Signal Routing Library");
		m.declare("signals.lib/version", "0.3");
		m.declare("version", "1.00");
	}

	int getNumInputs() nothrow @nogc {
		return 0;
	}
	int getNumOutputs() nothrow @nogc {
		return 3;
	}
	
	static void classInit(int sample_rate) nothrow @nogc {
		DSP_DrumSIG0* sig0 = newDSP_DrumSIG0();
		sig0.instanceInitDSP_DrumSIG0(sample_rate);
		sig0.fillDSP_DrumSIG0(8542, ftbl0DSP_DrumSIG0);
		deleteDSP_DrumSIG0(sig0);
	}
	
	void instanceConstants(int sample_rate) nothrow @nogc {
		fSampleRate = sample_rate;
		fConst0 = fmin(1.92e+05, fmax(1.0f, cast(float)(fSampleRate)));
		float fConst1 = 1.0f / tan(1063.7433f / fConst0);
		fConst2 = 1.0f - fConst1;
		fConst3 = 1.0f / fmax(1.0f, 0.02f * fConst0);
		fConst4 = fmax(1.0f, 0.001f * fConst0);
		fConst5 = 1.0f / fConst4;
		fConst6 = 1.0f / fConst0;
		fConst7 = 4.0f / fConst0;
		fConst8 = 1.0f / (fConst1 + 1.0f);
		fConst9 = 3.1415927f / fConst0;
		fConst10 = 0.0045454544f / fConst0;
		float fConst11 = tan(628.31854f / fConst0);
		float fConst12 = DSP_Drum_faustpower2_f(fConst11);
		fConst13 = 1.0f / fConst12;
		fConst14 = 2.0f * (1.0f - fConst13);
		float fConst15 = 1.0f / fConst11;
		fConst16 = (fConst15 + -1.4142135f) / fConst11 + 1.0f;
		fConst17 = 1.0f / ((fConst15 + 1.4142135f) / fConst11 + 1.0f);
		iConst18 = cast(int)(fConst0 / fmin(fConst0, 4.8e+04));
		fConst19 = 0.0f - 2.0f / fConst12;
		fConst20 = 1.0f / fmax(1.0f, 0.019f * fConst0);
	}
	
	void instanceResetUserInterface() nothrow @nogc {
		fEntry0 = cast(FAUSTFLOAT)(0.0f);
		fEntry1 = cast(FAUSTFLOAT)(0.0f);
		fHslider0 = cast(FAUSTFLOAT)(0.0f);
		fHslider1 = cast(FAUSTFLOAT)(1e+02);
		fHslider2 = cast(FAUSTFLOAT)(2e+02);
		fHslider3 = cast(FAUSTFLOAT)(0.0f);
		fCheckbox0 = cast(FAUSTFLOAT)(0.0f);
		fHslider4 = cast(FAUSTFLOAT)(2.2e+02);
		fHslider5 = cast(FAUSTFLOAT)(1e+02);
		fHslider6 = cast(FAUSTFLOAT)(0.0f);
		fHslider7 = cast(FAUSTFLOAT)(1e+04);
		fHslider8 = cast(FAUSTFLOAT)(0.0f);
		fHslider9 = cast(FAUSTFLOAT)(0.0f);
		fHslider10 = cast(FAUSTFLOAT)(25.0f);
		fHslider11 = cast(FAUSTFLOAT)(1e+02);
		fHslider12 = cast(FAUSTFLOAT)(5e+01);
		fHslider13 = cast(FAUSTFLOAT)(0.0f);
		fHslider14 = cast(FAUSTFLOAT)(0.0f);
		fHslider15 = cast(FAUSTFLOAT)(0.0f);
		fHslider16 = cast(FAUSTFLOAT)(0.0f);
	}
	
	void instanceClear() nothrow @nogc {
		for (int l0 = 0; l0 < 2; l0 = l0 + 1) {
			fVec0[l0] = 0.0f;
		}
		for (int l1 = 0; l1 < 2; l1 = l1 + 1) {
			iRec1[l1] = 0;
		}
		for (int l2 = 0; l2 < 2; l2 = l2 + 1) {
			iVec1[l2] = 0;
		}
		for (int l3 = 0; l3 < 2; l3 = l3 + 1) {
			iRec4[l3] = 0;
		}
		for (int l4 = 0; l4 < 2; l4 = l4 + 1) {
			fRec5[l4] = 0.0f;
		}
		for (int l5 = 0; l5 < 2; l5 = l5 + 1) {
			fRec6[l5] = 0.0f;
		}
		for (int l6 = 0; l6 < 2; l6 = l6 + 1) {
			fRec7[l6] = 0.0f;
		}
		for (int l7 = 0; l7 < 2; l7 = l7 + 1) {
			fRec8[l7] = 0.0f;
		}
		for (int l8 = 0; l8 < 2; l8 = l8 + 1) {
			fRec2[l8] = 0.0f;
		}
		for (int l9 = 0; l9 < 2; l9 = l9 + 1) {
			fVec2[l9] = 0.0f;
		}
		for (int l10 = 0; l10 < 2; l10 = l10 + 1) {
			fRec10[l10] = 0.0f;
		}
		for (int l11 = 0; l11 < 2; l11 = l11 + 1) {
			fRec11[l11] = 0.0f;
		}
		for (int l12 = 0; l12 < 2; l12 = l12 + 1) {
			fRec9[l12] = 0.0f;
		}
		for (int l13 = 0; l13 < 2; l13 = l13 + 1) {
			fVec3[l13] = 0.0f;
		}
		for (int l14 = 0; l14 < 2; l14 = l14 + 1) {
			fRec0[l14] = 0.0f;
		}
		for (int l15 = 0; l15 < 2; l15 = l15 + 1) {
			fRec14[l15] = 0.0f;
		}
		for (int l16 = 0; l16 < 2; l16 = l16 + 1) {
			fRec12[l16] = 0.0f;
		}
		for (int l17 = 0; l17 < 2; l17 = l17 + 1) {
			fRec16[l17] = 0.0f;
		}
		for (int l18 = 0; l18 < 3; l18 = l18 + 1) {
			fRec15[l18] = 0.0f;
		}
		for (int l19 = 0; l19 < 2; l19 = l19 + 1) {
			fRec17[l19] = 0.0f;
		}
		for (int l20 = 0; l20 < 2; l20 = l20 + 1) {
			iRec21[l20] = 0;
		}
		for (int l21 = 0; l21 < 2; l21 = l21 + 1) {
			iRec22[l21] = 0;
		}
		for (int l22 = 0; l22 < 2; l22 = l22 + 1) {
			fRec20[l22] = 0.0f;
		}
		for (int l23 = 0; l23 < 3; l23 = l23 + 1) {
			fRec19[l23] = 0.0f;
		}
		for (int l24 = 0; l24 < 3; l24 = l24 + 1) {
			fRec18[l24] = 0.0f;
		}
		for (int l25 = 0; l25 < 2; l25 = l25 + 1) {
			fRec25[l25] = 0.0f;
		}
		for (int l26 = 0; l26 < 2; l26 = l26 + 1) {
			fRec23[l26] = 0.0f;
		}
		for (int l27 = 0; l27 < 2; l27 = l27 + 1) {
			fRec28[l27] = 0.0f;
		}
		for (int l28 = 0; l28 < 2; l28 = l28 + 1) {
			fRec26[l28] = 0.0f;
		}
		for (int l29 = 0; l29 < 2; l29 = l29 + 1) {
			fRec29[l29] = 0.0f;
		}
		for (int l30 = 0; l30 < 2; l30 = l30 + 1) {
			fRec30[l30] = 0.0f;
		}
	}
	
	void initialize(int sample_rate) nothrow @nogc {
		classInit(sample_rate);
		instanceInit(sample_rate);
	}
	void instanceInit(int sample_rate) nothrow @nogc {
		instanceConstants(sample_rate);
		instanceResetUserInterface();
		instanceClear();
	}
	
	DSP_Drum clone() {
		return (mallocNew!DSP_Drum());
	}
	
	int getSampleRate() nothrow @nogc {
		return fSampleRate;
	}
	
	void buildUserInterface(UI* uiInterface) nothrow @nogc {
		uiInterface.openVerticalBox("OneTrick SIMIAN");
		uiInterface.addNumEntry("Choke", &fEntry0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1.0, cast(FAUSTFLOAT)0.01);
		uiInterface.addNumEntry("Trigger", &fEntry1, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1.0, cast(FAUSTFLOAT)0.01);
		uiInterface.addCheckButton("WakeUp", &fCheckbox0);
		uiInterface.declare(&fHslider7, "10", "");
		uiInterface.declare(&fHslider7, "scale", "log");
		uiInterface.addHorizontalSlider("LowpassFreq", &fHslider7, cast(FAUSTFLOAT)1e+04, cast(FAUSTFLOAT)4.4e+02, cast(FAUSTFLOAT)1.5e+04, cast(FAUSTFLOAT)1.0);
		uiInterface.declare(&fHslider8, "11", "");
		uiInterface.declare(&fHslider8, "scale", "log");
		uiInterface.addHorizontalSlider("LowpassQ", &fHslider8, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider6, "12", "");
		uiInterface.declare(&fHslider6, "midi", "ctrl 72");
		uiInterface.addHorizontalSlider("LowpassEnvBend", &fHslider6, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)-36.0, cast(FAUSTFLOAT)36.0, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider5, "13", "");
		uiInterface.addHorizontalSlider("LowpassEnvBendDynamics", &fHslider5, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider4, "1", "");
		uiInterface.declare(&fHslider4, "midi", "ctrl 71");
		uiInterface.declare(&fHslider4, "scale", "log");
		uiInterface.addHorizontalSlider("OscPitch", &fHslider4, cast(FAUSTFLOAT)2.2e+02, cast(FAUSTFLOAT)3e+01, cast(FAUSTFLOAT)4.4e+03, cast(FAUSTFLOAT)0.001);
		uiInterface.declare(&fHslider9, "1", "");
		uiInterface.addHorizontalSlider("OscWaveform", &fHslider9, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.001);
		uiInterface.declare(&fHslider2, "20", "");
		uiInterface.declare(&fHslider2, "midi", "ctrl 74");
		uiInterface.addHorizontalSlider("EnvDecay", &fHslider2, cast(FAUSTFLOAT)2e+02, cast(FAUSTFLOAT)1e+01, cast(FAUSTFLOAT)2e+03, cast(FAUSTFLOAT)1.0);
		uiInterface.declare(&fHslider0, "21", "");
		uiInterface.declare(&fHslider0, "midi", "ctrl 74");
		uiInterface.addHorizontalSlider("EnvPunch", &fHslider0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider3, "2", "");
		uiInterface.declare(&fHslider3, "midi", "ctrl 72");
		uiInterface.addHorizontalSlider("OscEnvBend", &fHslider3, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)-36.0, cast(FAUSTFLOAT)36.0, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider10, "30", "");
		uiInterface.declare(&fHslider10, "midi", "ctrl 73");
		uiInterface.addHorizontalSlider("FadeToneNoise", &fHslider10, cast(FAUSTFLOAT)25.0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider12, "31", "");
		uiInterface.addHorizontalSlider("FadeClick", &fHslider12, cast(FAUSTFLOAT)5e+01, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider16, "32", "");
		uiInterface.addHorizontalSlider("SendReverb", &fHslider16, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider1, "3", "");
		uiInterface.addHorizontalSlider("OscEnvBendDynamics", &fHslider1, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider15, "40", "");
		uiInterface.addHorizontalSlider("Gain", &fHslider15, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)-1e+02, cast(FAUSTFLOAT)6.0, cast(FAUSTFLOAT)0.1);
		uiInterface.declare(&fHslider14, "41", "");
		uiInterface.addHorizontalSlider("Pan", &fHslider14, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)-1e+02, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider11, "42", "");
		uiInterface.addHorizontalSlider("GainDynamics", &fHslider11, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.declare(&fHslider13, "43", "");
		uiInterface.addHorizontalSlider("Saturation", &fHslider13, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)0.0, cast(FAUSTFLOAT)1e+02, cast(FAUSTFLOAT)0.01);
		uiInterface.closeBox();
	}
	
	void compute(int count, FAUSTFLOAT*[] inputs, FAUSTFLOAT*[] outputs) nothrow @nogc {
		FAUSTFLOAT* output0 = outputs[0];
		FAUSTFLOAT* output1 = outputs[1];
		FAUSTFLOAT* output2 = outputs[2];
		float fSlow0 = cast(float)(fEntry0);
		float fSlow1 = cast(float)(fEntry1);
		int iSlow2 = fSlow1 > 0.0f;
		float fSlow3 = 0.01f * cast(float)(fHslider0);
		float fSlow4 = 0.01f * cast(float)(fHslider1);
		float fSlow5 = 0.001f * cast(float)(fHslider2);
		float fSlow6 = 0.083333336f * cast(float)(fHslider3);
		float fSlow7 = cast(float)(fCheckbox0);
		float fSlow8 = cast(float)(fHslider4);
		float fSlow9 = 0.01f * cast(float)(fHslider5);
		float fSlow10 = 0.083333336f * cast(float)(fHslider6);
		float fSlow11 = cast(float)(fHslider7);
		float fSlow12 = 1.0f / fmax(0.01f, 0.01f * cast(float)(fHslider8) + 1.0f);
		float fSlow13 = 0.01f * cast(float)(fHslider9);
		float fSlow14 = cast(float)(fHslider10);
		float fSlow15 = 0.01f * fSlow14;
		float fSlow16 = 0.01f * cast(float)(fHslider11);
		float fSlow17 = cast(float)(fHslider12);
		float fSlow18 = cast(float)(fHslider13);
		float fSlow19 = 1.0f / fmin(fmax(0.004f * fSlow18 + 0.5f, 0.001f), 0.999f) + -2.0f;
		float fSlow20 = 0.01f * cast(float)(fHslider14);
		float fSlow21 = pow(1e+01, 0.05f * fmax(-1e+02, cast(float)(fHslider15)));
		float fSlow22 = 1.0f - 0.007f * fSlow18;
		float fSlow23 = 0.01f * cast(float)(fHslider16) * fSlow22;
		for (int i0 = 0; i0 < count; i0 = i0 + 1) {
			fVec0[0] = fSlow0;
			iRec1[0] = (fSlow0 > fVec0[1]) + (fSlow0 <= fVec0[1]) * (iRec1[1] + (iRec1[1] > 0));
			float fTemp0 = cast(float)(iRec1[0]);
			float fTemp1 = fSlow0 * (fmax(0.0f, fmin(fTemp0, fConst3 * (1.0f - fTemp0) + 1.0f)) + -1.0f) + 1.0f;
			iVec1[0] = iSlow2;
			int iTemp2 = iSlow2 > iVec1[1];
			iRec4[0] = (iRec4[1] + (iRec4[1] > 0)) * (iSlow2 <= iVec1[1]) + iTemp2;
			float fTemp3 = cast(float)(iRec4[0]);
			int iTemp4 = fTemp3 >= fConst4;
			fRec5[0] = ((iTemp2) ? fSlow3 : fRec5[1]);
			float fTemp5 = 1.0f / fmin(fmax(0.4f * fRec5[0] + 0.5f, 0.001f), 0.999f) + -2.0f;
			float fTemp6 = fTemp5 * (1.0f - fConst5 * fTemp3) + 1.0f;
			fRec6[0] = ((iTemp2) ? fRec2[1] : fRec6[1]);
			fRec7[0] = ((iSlow2) ? fSlow1 : fRec7[1]);
			float fTemp7 = fRec7[0] + -1.0f;
			float fTemp8 = fmax(fRec6[0], fSlow4 * fTemp7 + 1.0f);
			float fElse3 = fSlow5 * (0.7f * fTemp7 + 1.0f);
			fRec8[0] = ((iTemp2) ? fElse3 : fRec8[1]);
			float fTemp9 = fTemp3 - fConst4;
			float fTemp10 = fmin(1.0f, fTemp9 / fmax(1.0f, fConst0 * fRec8[0]));
			float fTemp11 = 1.0f - fTemp10 / (fTemp5 * (1.0f - fTemp10) + 1.0f);
			fRec2[0] = fmin(1.0f, fmax(0.0f, ((iTemp4) ? (fTemp8 * fTemp11) : (fRec6[0] + fConst5 * (fTemp3 * (fTemp8 - fRec6[0]) / fTemp6)))));
			float fRec3 = fRec2[1];
			float fTemp12 = pow(2.0f, fSlow6 * fRec3 * fTemp1);
			fVec2[0] = fSlow7;
			float fTemp13 = cast(float)(fSlow7 <= fVec2[1]);
			float fTemp14 = 1.0f - 0.9995f * fTemp13;
			fRec10[0] = fSlow8 * fTemp14 + 0.9995f * fTemp13 * fRec10[1];
			float fTemp15 = fRec10[0] * fTemp12;
			float fTemp16 = fConst6 * fTemp15;
			float fTemp17 = 1.0f - fTemp16;
			float fTemp18 = fRec11[1] + fTemp16;
			fRec11[0] = fTemp18 - floor(fTemp18);
			int iTemp19 = (fRec11[0] < fTemp16) + 2 * (fRec11[0] > fTemp17);
			float fTemp20 = fRec11[0] + -1.0f;
			float fThen6 = ((iTemp19 == 1) ? (fConst0 * (fRec11[0] * (2.0f - fConst0 * (fRec11[0] / fTemp15)) / fTemp15) + -1.0f) : (fConst0 * (fTemp20 * (fConst0 * (fTemp20 / fTemp15) + 2.0f) / fTemp15) + 1.0f));
			float fTemp21 = fmod(fmod(fRec11[0] + 0.5f, 1.0f) + 1.0f, 1.0f);
			int iTemp22 = (fTemp21 < fTemp16) + 2 * (fTemp21 > fTemp17);
			float fTemp23 = fTemp21 + -1.0f;
			float fThen8 = ((iTemp22 == 1) ? (fConst0 * (fTemp21 * (2.0f - fConst0 * (fTemp21 / fTemp15)) / fTemp15) + -1.0f) : (fConst0 * (fTemp23 * (fConst0 * (fTemp23 / fTemp15) + 2.0f) / fTemp15) + 1.0f));
			fRec9[0] = ((iTemp22 == 0) ? 0.0f : fThen8) + 0.999f * fRec9[1] + cast(float)(2 * cast(int)(2.0f * fRec11[0]) + -1) - ((iTemp19 == 0) ? 0.0f : fThen6);
			float fTemp24 = fRec10[0] * fRec9[0] * fTemp12;
			fVec3[0] = fTemp24;
			fRec0[0] = fConst8 * (fConst7 * (fTemp24 + fVec3[1]) - fConst2 * fRec0[1]);
			fRec14[0] = ((iTemp2) ? fRec12[1] : fRec14[1]);
			float fTemp25 = fmax(fRec14[0], fSlow9 * fTemp7 + 1.0f);
			fRec12[0] = fmin(1.0f, fmax(0.0f, ((iTemp4) ? (fTemp11 * fTemp25) : (fRec14[0] + fConst5 * (fTemp3 * (fTemp25 - fRec14[0]) / fTemp6)))));
			float fRec13 = fRec12[1];
			float fTemp26 = tan(fConst9 * fmin(fSlow11 * pow(2.0f, fSlow10 * fRec13 * fTemp1), 1.9e+04));
			float fTemp27 = 1.0f / fTemp26;
			float fTemp28 = (fSlow12 + fTemp27) / fTemp26 + 1.0f;
			float fTemp29 = 1.0f - 1.0f / DSP_Drum_faustpower2_f(fTemp26);
			float fTemp30 = (fTemp27 - fSlow12) / fTemp26 + 1.0f;
			float fTemp31 = fRec16[1] + fConst10 * fTemp15;
			fRec16[0] = fTemp31 - floor(fTemp31);
			float fTemp32 = 8542.0f * fRec16[0];
			int iTemp33 = cast(int)(fTemp32);
			float fTemp34 = ftbl0DSP_DrumSIG0[(iTemp33 + 3) % 8542];
			float fTemp35 = ftbl0DSP_DrumSIG0[(iTemp33 + 1) % 8542];
			float fTemp36 = ftbl0DSP_DrumSIG0[iTemp33];
			float fTemp37 = ftbl0DSP_DrumSIG0[(iTemp33 + 2) % 8542];
			float fTemp38 = fTemp32 - cast(float)(iTemp33);
			fRec15[0] = fTemp35 + fTemp38 * (fTemp37 - fTemp36 + fTemp38 * ((fTemp34 + fTemp35 - (fTemp37 + fTemp36)) * fTemp38 + (fTemp37 + 2.0f * (fTemp36 - fTemp35) - fTemp34))) - (fRec15[2] * fTemp30 + 2.0f * fRec15[1] * fTemp29) / fTemp28;
			fRec17[0] = fSlow13 * fTemp14 + 0.9995f * fTemp13 * fRec17[1];
			float fTemp39 = fRec0[0] + fRec17[0] * ((fRec15[2] + fRec15[0] + 2.0f * fRec15[1]) / fTemp28 - fRec0[0]);
			iRec21[0] = iRec21[1] + 1;
			iRec22[0] = 1103515245 * iRec22[1] + 12345;
			float fElse11 = 4.656613e-10f * cast(float)(iRec22[0]);
			fRec20[0] = ((((iRec21[0] + -1) % iConst18) == 0) ? fElse11 : fRec20[1]);
			fRec19[0] = fRec20[0] - fConst17 * (fConst16 * fRec19[2] + fConst14 * fRec19[1]);
			fRec18[0] = fConst17 * (fConst13 * fRec19[0] + fConst19 * fRec19[1] + fConst13 * fRec19[2]) - (fTemp30 * fRec18[2] + 2.0f * fTemp29 * fRec18[1]) / fTemp28;
			float fTemp40 = (fRec18[2] + fRec18[0] + 2.0f * fRec18[1]) / fTemp28;
			float fTemp41 = fTemp40 - fTemp39;
			fRec25[0] = ((iTemp2) ? fRec23[1] : fRec25[1]);
			float fTemp42 = sqrt(fSlow16 * fTemp7 + 1.0f);
			float fTemp43 = fmax(fRec25[0], fTemp42);
			fRec23[0] = fmin(1.0f, fmax(0.0f, ((iTemp4) ? (fTemp43 * (1.0f - fConst20 * fTemp9)) : (fRec25[0] + fConst5 * fTemp3 * (fTemp43 - fRec25[0])))));
			float fRec24 = fRec23[1];
			fRec28[0] = ((iTemp2) ? fRec26[1] : fRec28[1]);
			float fTemp44 = fmax(fRec28[0], fTemp42);
			fRec26[0] = fmin(1.0f, fmax(0.0f, ((iTemp4) ? (fTemp11 * fTemp44) : (fRec28[0] + fConst5 * (fTemp3 * (fTemp44 - fRec28[0]) / fTemp6)))));
			float fRec27 = fRec26[1];
			float fTemp45 = DSP_Drum_faustpower2_f(fRec27) * DSP_Drum_faustpower2_f(fTemp1) * (fTemp39 + 0.01f * (fSlow14 * fTemp41 + fSlow17 * DSP_Drum_faustpower2_f(fRec24) * (fTemp40 - (fTemp39 + fSlow15 * fTemp41))));
			float fTemp46 = fabs(fTemp45);
			float fTemp47 = fSlow19 * (1.0f - fTemp46) + 1.0f;
			float fTemp48 = 1.0f - cast(float)(2 * (fTemp45 < 0.0f));
			fRec29[0] = fSlow20 * fTemp14 + 0.9995f * fTemp13 * fRec29[1];
			fRec30[0] = fSlow21 * fTemp14 + 0.9995f * fTemp13 * fRec30[1];
			output0[i0] = cast(FAUSTFLOAT)(fSlow22 * (fRec30[0] * fmin(1.0f, 1.0f - fRec29[0]) * fTemp46 * fTemp48 / fTemp47));
			output1[i0] = cast(FAUSTFLOAT)(fSlow22 * (fRec30[0] * fmin(1.0f, fRec29[0] + 1.0f) * fTemp46 * fTemp48 / fTemp47));
			output2[i0] = cast(FAUSTFLOAT)(fSlow23 * (fRec30[0] * fTemp46 * fTemp48 / fTemp47));
			fVec0[1] = fVec0[0];
			iRec1[1] = iRec1[0];
			iVec1[1] = iVec1[0];
			iRec4[1] = iRec4[0];
			fRec5[1] = fRec5[0];
			fRec6[1] = fRec6[0];
			fRec7[1] = fRec7[0];
			fRec8[1] = fRec8[0];
			fRec2[1] = fRec2[0];
			fVec2[1] = fVec2[0];
			fRec10[1] = fRec10[0];
			fRec11[1] = fRec11[0];
			fRec9[1] = fRec9[0];
			fVec3[1] = fVec3[0];
			fRec0[1] = fRec0[0];
			fRec14[1] = fRec14[0];
			fRec12[1] = fRec12[0];
			fRec16[1] = fRec16[0];
			fRec15[2] = fRec15[1];
			fRec15[1] = fRec15[0];
			fRec17[1] = fRec17[0];
			iRec21[1] = iRec21[0];
			iRec22[1] = iRec22[0];
			fRec20[1] = fRec20[0];
			fRec19[2] = fRec19[1];
			fRec19[1] = fRec19[0];
			fRec18[2] = fRec18[1];
			fRec18[1] = fRec18[0];
			fRec25[1] = fRec25[0];
			fRec23[1] = fRec23[0];
			fRec28[1] = fRec28[0];
			fRec26[1] = fRec26[0];
			fRec29[1] = fRec29[0];
			fRec30[1] = fRec30[0];
		}
	}

}

/***************************END USER SECTION ***************************/

/*******************BEGIN ARCHITECTURE SECTION (part 2/2)***************/

/********************END ARCHITECTURE SECTION (part 2/2)****************/
